/* ====================================================================================================================

	The copyright in this software is being made available under the License included below. 
	This software may be subject to other third party and 	contributor rights, including patent rights, and no such 
	rights are granted under this license.
 
	Copyright (c) 2010, SAMSUNG ELECTRONICS CO., LTD. and BRITISH BROADCASTING CORPORATION
	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted only for 
	the purpose of developing standards within the Joint Collaborative Team on Video Coding and for testing and
	promoting such standards. The following conditions are required to be met:

		* Redistributions of source code must retain the above copyright notice, this list of conditions and 
		  the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and 
		  the following disclaimer in the documentation and/or other materials provided with the distribution.
		* Neither the name of SAMSUNG ELECTRONICS CO., LTD. nor the name of the BRITISH BROADCASTING CORPORATION 
		  may be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 * ====================================================================================================================
*/

/** \file			TComRom.cpp
    \brief		global variables & functions
*/

#include "TComRom.h"
#include <memory.h>
#include <stdlib.h>

// ACS
#include <algorithm>

// MPI
const UInt Optimal_MPI_index[7][9][9] =
/////////////// partition   INTRA_Pred_Mode  MPI_index
{
	{ // 2x2
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  { // 4x4
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  { //8x8
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  { //16x16
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  {//32x32
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  {//64x64
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
  {//128x128
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
    {0, 1,   2,  3,  4,  5,  6,  7,  8,},
  },
};

const UChar g_aucConvertTo4x4Idx[8]=
{  0, 2, 8, 10, 0, 2, 8, 10 };

const UChar g_aucConvertTo8x8Idx[16]=
{
    0 ,0 ,1 , 1,
    0 ,0 ,1 , 1,
    2 ,2 ,3 , 3,
    2 ,2 ,3 ,3
};

const UChar g_aucRelativeIndexIn8x8 [4] = { 0, 1, 4, 5 };

const UChar g_aucConvertBlockOrder[17] =
{
   0,  1,  4,  5,
   2,  3,  6,  7,
   8,  9, 12, 13,
  10, 11, 14, 15, 0xff,
};

///-- For Quantization Scalig Table

//const UChar g_aucIndexChromaDCScan[4] =
const UInt g_aucIndexChromaDCScan[4] =
{
  0, 16, 32, 48
};

//const UChar g_aucLumaFrameDCScan[16] =
const UInt g_aucLumaFrameDCScan[16] =
{
  0, 16, 64, 128, 80, 32, 48, 96, 144, 192, 208, 160, 112, 176, 224, 240
};

const UChar g_aucChroma2LumaIdx[8 ]  = { 0, 2, 8,10, 0, 2, 8,10 };

const UChar g_auc4x4Idx28x8Idx [16 ]  = { 0, 0, 1, 1,   0, 0, 1, 1,   2, 2, 3, 3,   2, 2, 3, 3 };

UInt g_aiQuantCoef4[6] =
{
  102, 93, 79, 73, 64, 57
};

Int g_aiDequantCoef4[6] =
{
  10, 11, 13, 14, 16,18
};

UInt g_aiQuantCoef[6][16] =
{
  { 13107, 8066,13107, 8066,
     8066, 5243, 8066, 5243,
    13107, 8066,13107, 8066,
     8066, 5243, 8066, 5243
  },
  { 11916, 7490,11916, 7490,
     7490, 4660, 7490, 4660,
    11916, 7490,11916, 7490,
     7490, 4660, 7490, 4660
  },
  { 10082, 6554,10082, 6554,
     6554, 4194, 6554, 4194,
    10082, 6554,10082, 6554,
     6554, 4194, 6554, 4194
  },
  {  9362, 5825, 9362, 5825,
     5825, 3647, 5825, 3647,
     9362, 5825, 9362, 5825,
     5825, 3647, 5825, 3647
  },
  {  8192, 5243, 8192, 5243,
     5243, 3355, 5243, 3355,
     8192, 5243, 8192, 5243,
     5243, 3355, 5243, 3355
  },
  {  7282, 4559, 7282, 4559,
     4559, 2893, 4559, 2893,
     7282, 4559, 7282, 4559,
     4559, 2893, 4559, 2893
  }
};


Int g_aiDequantCoef[6][16] =
{
  { 10, 13, 10, 13,
    13, 16, 13, 16,
    10, 13, 10, 13,
    13, 16, 13, 16
  },
  { 11, 14, 11, 14,
    14, 18, 14, 18,
    11, 14, 11, 14,
    14, 18, 14, 18
  },
  { 13, 16, 13, 16,
    16, 20, 16, 20,
    13, 16, 13, 16,
    16, 20, 16, 20
  },
  { 14, 18, 14, 18,
    18, 23, 18, 23,
    14, 18, 14, 18,
    18, 23, 18, 23
  },
  { 16, 20, 16, 20,
    20, 25, 20, 25,
    16, 20, 16, 20,
    20, 25, 20, 25
  },
  {
    18, 23, 18, 23,
    23, 29, 23, 29,
    18, 23, 18, 23,
    23, 29, 23, 29
  }
};

const UChar g_aucChromaScale[52]=
{
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
   12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,
   28,29,29,30,31,32,32,33,34,34,35,35,36,36,37,37,
   37,38,38,38,39,39,39,39
};

Int g_aiDequantCoef64[6][64] =
{
  {
    20,  19, 25, 19, 20, 19, 25, 19,
    19,  18, 24, 18, 19, 18, 24, 18,
    25,  24, 32, 24, 25, 24, 32, 24,
    19,  18, 24, 18, 19, 18, 24, 18,
    20,  19, 25, 19, 20, 19, 25, 19,
    19,  18, 24, 18, 19, 18, 24, 18,
    25,  24, 32, 24, 25, 24, 32, 24,
    19,  18, 24, 18, 19, 18, 24, 18
  },
  {
    22,  21, 28, 21, 22, 21, 28, 21,
    21,  19, 26, 19, 21, 19, 26, 19,
    28,  26, 35, 26, 28, 26, 35, 26,
    21,  19, 26, 19, 21, 19, 26, 19,
    22,  21, 28, 21, 22, 21, 28, 21,
    21,  19, 26, 19, 21, 19, 26, 19,
    28,  26, 35, 26, 28, 26, 35, 26,
    21,  19, 26, 19, 21, 19, 26, 19
  },
  {
    26,  24, 33, 24, 26, 24, 33, 24,
    24,  23, 31, 23, 24, 23, 31, 23,
    33,  31, 42, 31, 33, 31, 42, 31,
    24,  23, 31, 23, 24, 23, 31, 23,
    26,  24, 33, 24, 26, 24, 33, 24,
    24,  23, 31, 23, 24, 23, 31, 23,
    33,  31, 42, 31, 33, 31, 42, 31,
    24,  23, 31, 23, 24, 23, 31, 23
  },
  {
    28,  26, 35, 26, 28, 26, 35, 26,
    26,  25, 33, 25, 26, 25, 33, 25,
    35,  33, 45, 33, 35, 33, 45, 33,
    26,  25, 33, 25, 26, 25, 33, 25,
    28,  26, 35, 26, 28, 26, 35, 26,
    26,  25, 33, 25, 26, 25, 33, 25,
    35,  33, 45, 33, 35, 33, 45, 33,
    26,  25, 33, 25, 26, 25, 33, 25
  },
  {
    32,  30, 40, 30, 32, 30, 40, 30,
    30,  28, 38, 28, 30, 28, 38, 28,
    40,  38, 51, 38, 40, 38, 51, 38,
    30,  28, 38, 28, 30, 28, 38, 28,
    32,  30, 40, 30, 32, 30, 40, 30,
    30,  28, 38, 28, 30, 28, 38, 28,
    40,  38, 51, 38, 40, 38, 51, 38,
    30,  28, 38, 28, 30, 28, 38, 28
  },
  {
    36,  34, 46, 34, 36, 34, 46, 34,
    34,  32, 43, 32, 34, 32, 43, 32,
    46,  43, 58, 43, 46, 43, 58, 43,
    34,  32, 43, 32, 34, 32, 43, 32,
    36,  34, 46, 34, 36, 34, 46, 34,
    34,  32, 43, 32, 34, 32, 43, 32,
    46,  43, 58, 43, 46, 43, 58, 43,
    34,  32, 43, 32, 34, 32, 43, 32
  }

};

UInt g_aiQuantCoef64[6][64] =
{
  {
    0x3333, 0x2fbe, 0x4189, 0x2fbe, 0x3333, 0x2fbe, 0x4189, 0x2fbe,
    0x2fbe, 0x2ca4, 0x3c79, 0x2ca4, 0x2fbe, 0x2ca4, 0x3c79, 0x2ca4,
    0x4189, 0x3c79, 0x51ec, 0x3c79, 0x4189, 0x3c79, 0x51ec, 0x3c79,
    0x2fbe, 0x2ca4, 0x3c79, 0x2ca4, 0x2fbe, 0x2ca4, 0x3c79, 0x2ca4,
    0x3333, 0x2fbe, 0x4189, 0x2fbe, 0x3333, 0x2fbe, 0x4189, 0x2fbe,
    0x2fbe, 0x2ca4, 0x3c79, 0x2ca4, 0x2fbe, 0x2ca4, 0x3c79, 0x2ca4,
    0x4189, 0x3c79, 0x51ec, 0x3c79, 0x4189, 0x3c79, 0x51ec, 0x3c79,
    0x2fbe, 0x2ca4, 0x3c79, 0x2ca4, 0x2fbe, 0x2ca4, 0x3c79, 0x2ca4,
  },
  {
    0x2e8c, 0x2b32, 0x3a84, 0x2b32, 0x2e8c, 0x2b32, 0x3a84, 0x2b32,
    0x2b32, 0x2a4a, 0x37d2, 0x2a4a, 0x2b32, 0x2a4a, 0x37d2, 0x2a4a,
    0x3a84, 0x37d2, 0x4ae6, 0x37d2, 0x3a84, 0x37d2, 0x4ae6, 0x37d2,
    0x2b32, 0x2a4a, 0x37d2, 0x2a4a, 0x2b32, 0x2a4a, 0x37d2, 0x2a4a,
    0x2e8c, 0x2b32, 0x3a84, 0x2b32, 0x2e8c, 0x2b32, 0x3a84, 0x2b32,
    0x2b32, 0x2a4a, 0x37d2, 0x2a4a, 0x2b32, 0x2a4a, 0x37d2, 0x2a4a,
    0x3a84, 0x37d2, 0x4ae6, 0x37d2, 0x3a84, 0x37d2, 0x4ae6, 0x37d2,
    0x2b32, 0x2a4a, 0x37d2, 0x2a4a, 0x2b32, 0x2a4a, 0x37d2, 0x2a4a,
  },
  {
    0x2762, 0x25cb, 0x31a6, 0x25cb, 0x2762, 0x25cb, 0x31a6, 0x25cb,
    0x25cb, 0x22ef, 0x2ed1, 0x22ef, 0x25cb, 0x22ef, 0x2ed1, 0x22ef,
    0x31a6, 0x2ed1, 0x3e6a, 0x2ed1, 0x31a6, 0x2ed1, 0x3e6a, 0x2ed1,
    0x25cb, 0x22ef, 0x2ed1, 0x22ef, 0x25cb, 0x22ef, 0x2ed1, 0x22ef,
    0x2762, 0x25cb, 0x31a6, 0x25cb, 0x2762, 0x25cb, 0x31a6, 0x25cb,
    0x25cb, 0x22ef, 0x2ed1, 0x22ef, 0x25cb, 0x22ef, 0x2ed1, 0x22ef,
    0x31a6, 0x2ed1, 0x3e6a, 0x2ed1, 0x31a6, 0x2ed1, 0x3e6a, 0x2ed1,
    0x25cb, 0x22ef, 0x2ed1, 0x22ef, 0x25cb, 0x22ef, 0x2ed1, 0x22ef,
  },
  {
    0x2492, 0x22e3, 0x2ed0, 0x22e3, 0x2492, 0x22e3, 0x2ed0, 0x22e3,
    0x22e3, 0x2024, 0x2bfb, 0x2024, 0x22e3, 0x2024, 0x2bfb, 0x2024,
    0x2ed0, 0x2bfb, 0x3a41, 0x2bfb, 0x2ed0, 0x2bfb, 0x3a41, 0x2bfb,
    0x22e3, 0x2024, 0x2bfb, 0x2024, 0x22e3, 0x2024, 0x2bfb, 0x2024,
    0x2492, 0x22e3, 0x2ed0, 0x22e3, 0x2492, 0x22e3, 0x2ed0, 0x22e3,
    0x22e3, 0x2024, 0x2bfb, 0x2024, 0x22e3, 0x2024, 0x2bfb, 0x2024,
    0x2ed0, 0x2bfb, 0x3a41, 0x2bfb, 0x2ed0, 0x2bfb, 0x3a41, 0x2bfb,
    0x22e3, 0x2024, 0x2bfb, 0x2024, 0x22e3, 0x2024, 0x2bfb, 0x2024,
  },
  {
    0x2000, 0x1e3c, 0x28f6, 0x1e3c, 0x2000, 0x1e3c, 0x28f6, 0x1e3c,
    0x1e3c, 0x1cb2, 0x2631, 0x1cb2, 0x1e3c, 0x1cb2, 0x2631, 0x1cb2,
    0x28f6, 0x2631, 0x3367, 0x2631, 0x28f6, 0x2631, 0x3367, 0x2631,
    0x1e3c, 0x1cb2, 0x2631, 0x1cb2, 0x1e3c, 0x1cb2, 0x2631, 0x1cb2,
    0x2000, 0x1e3c, 0x28f6, 0x1e3c, 0x2000, 0x1e3c, 0x28f6, 0x1e3c,
    0x1e3c, 0x1cb2, 0x2631, 0x1cb2, 0x1e3c, 0x1cb2, 0x2631, 0x1cb2,
    0x28f6, 0x2631, 0x3367, 0x2631, 0x28f6, 0x2631, 0x3367, 0x2631,
    0x1e3c, 0x1cb2, 0x2631, 0x1cb2, 0x1e3c, 0x1cb2, 0x2631, 0x1cb2,
  },
  {
    0x1c72, 0x1aae, 0x239e, 0x1aae, 0x1c72, 0x1aae, 0x239e, 0x1aae,
    0x1aae, 0x191c, 0x21c0, 0x191c, 0x1aae, 0x191c, 0x21c0, 0x191c,
    0x239e, 0x21c0, 0x2d32, 0x21c0, 0x239e, 0x21c0, 0x2d32, 0x21c0,
    0x1aae, 0x191c, 0x21c0, 0x191c, 0x1aae, 0x191c, 0x21c0, 0x191c,
    0x1c72, 0x1aae, 0x239e, 0x1aae, 0x1c72, 0x1aae, 0x239e, 0x1aae,
    0x1aae, 0x191c, 0x21c0, 0x191c, 0x1aae, 0x191c, 0x21c0, 0x191c,
    0x239e, 0x21c0, 0x2d32, 0x21c0, 0x239e, 0x21c0, 0x2d32, 0x21c0,
    0x1aae, 0x191c, 0x21c0, 0x191c, 0x1aae, 0x191c, 0x21c0, 0x191c,
  }
};

const UChar g_aucScalingMatrixDefault4x4Intra[16] =
{
   6, 13, 20, 28,
  13, 20, 28, 32,
  20, 28, 32, 37,
  28, 32, 37, 42
};

const UChar g_aucScalingMatrixDefault4x4Inter[16] =
{
  10, 14, 20, 24,
  14, 20, 24, 27,
  20, 24, 27, 30,
  24, 27, 30, 34
};

const UChar g_aucScalingMatrixDefault8x8Intra[64] =
{
   6, 10, 13, 16, 18, 23, 25, 27,
  10, 11, 16, 18, 23, 25, 27, 29,
  13, 16, 18, 23, 25, 27, 29, 31,
  16, 18, 23, 25, 27, 29, 31, 33,
  18, 23, 25, 27, 29, 31, 33, 36,
  23, 25, 27, 29, 31, 33, 36, 38,
  25, 27, 29, 31, 33, 36, 38, 40,
  27, 29, 31, 33, 36, 38, 40, 42
};

const UChar g_aucScalingMatrixDefault8x8Inter[64] =
{
   9, 13, 15, 17, 19, 21, 22, 24,
  13, 13, 17, 19, 21, 22, 24, 25,
  15, 17, 19, 21, 22, 24, 25, 27,
  17, 19, 21, 22, 24, 25, 27, 28,
  19, 21, 22, 24, 25, 27, 28, 30,
  21, 22, 24, 25, 27, 28, 30, 32,
  22, 24, 25, 27, 28, 30, 32, 33,
  24, 25, 27, 28, 30, 32, 33, 35
};


//  Table for CAVLC
const UInt g_auiIncVlc[] = {0,3,6,12,24,48,32768};	// maximum vlc = 6

const UChar g_aucCodeTableTO16[3][4][17] =
{
  {
    { 1, 5, 7, 7, 7, 7,15,11, 8,15,11,15,11,15,11, 7,4},
    { 0, 1, 4, 6, 6, 6, 6,14,10,14,10,14,10, 1,14,10,6},
    { 0, 0, 1, 5, 5, 5, 5, 5,13, 9,13, 9,13, 9,13, 9,5},
    { 0, 0, 0, 3, 3, 4, 4, 4, 4, 4,12,12, 8,12, 8,12,8},
  },
  {
    { 3,11, 7, 7, 7, 4, 7,15,11,15,11, 8,15,11, 7, 9,7},
    { 0, 2, 7,10, 6, 6, 6, 6,14,10,14,10,14,10,11, 8,6},
    { 0, 0, 3, 9, 5, 5, 5, 5,13, 9,13, 9,13, 9, 6,10,5},
    { 0, 0, 0, 5, 4, 6, 8, 4, 4, 4,12, 8,12,12, 8, 1,4},
  },
  {
    {15,15,11, 8,15,11, 9, 8,15,11,15,11, 8,13, 9, 5,1},
    { 0,14,15,12,10, 8,14,10,14,14,10,14,10, 7,12, 8,4},
    { 0, 0,13,14,11, 9,13, 9,13,10,13, 9,13, 9,11, 7,3},
    { 0, 0, 0,12,11,10, 9, 8,13,12,12,12, 8,12,10, 6,2},
  },
};

const UChar g_aucLenTableTO16[3][4][17] =
{
  {   // 0702
    { 1, 6, 8, 9,10,11,13,13,13,14,14,15,15,16,16,16,16},
    { 0, 2, 6, 8, 9,10,11,13,13,14,14,15,15,15,16,16,16},
    { 0, 0, 3, 7, 8, 9,10,11,13,13,14,14,15,15,16,16,16},
    { 0, 0, 0, 5, 6, 7, 8, 9,10,11,13,14,14,15,15,16,16},
  },
  {
    { 2, 6, 6, 7, 8, 8, 9,11,11,12,12,12,13,13,13,14,14},
    { 0, 2, 5, 6, 6, 7, 8, 9,11,11,12,12,13,13,14,14,14},
    { 0, 0, 3, 6, 6, 7, 8, 9,11,11,12,12,13,13,13,14,14},
    { 0, 0, 0, 4, 4, 5, 6, 6, 7, 9,11,11,12,13,13,13,14},
  },
  {
    { 4, 6, 6, 6, 7, 7, 7, 7, 8, 8, 9, 9, 9,10,10,10,10},
    { 0, 4, 5, 5, 5, 5, 6, 6, 7, 8, 8, 9, 9, 9,10,10,10},
    { 0, 0, 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9,10,10,10},
    { 0, 0, 0, 4, 4, 4, 4, 4, 5, 6, 7, 8, 8, 9,10,10,10},
  },
};

const UChar g_aucCodeTable3[7][15] =
{
  {1,0},
  {1,1,0},
  {3,2,1,0},
  {3,2,1,1,0},
  {3,2,3,2,1,0},
  {3,0,1,3,2,5,4},
  {7,6,5,4,3,2,1,1,1,1,1,1,1,1,1},
};

const UChar g_aucLenTable3[7][15] =
{
  {1,1},
  {1,2,2},
  {2,2,2,2},
  {2,2,2,3,3},
  {2,2,3,3,3,3},
  {2,3,3,3,3,3,3},
  {3,3,3,3,3,3,3,4,5,6,7,8,9,10,11},
};

const UChar g_aucCodeTableTZ4[3][4] =
{
  { 1, 1, 1, 0,},
  { 1, 1, 0, 0,},
  { 1, 0, 0, 0,},
};

const UChar g_aucLenTableTZ4[3][4] =
{
  { 1, 2, 3, 3,},
  { 1, 2, 2, 0,},
  { 1, 1, 0, 0,},
};

const UChar g_aucCodeTableTZ16[15][16] =
{
  {1,3,2,3,2,3,2,3,2,3,2,3,2,3,2,1},
  {7,6,5,4,3,5,4,3,2,3,2,3,2,1,0},
  {5,7,6,5,4,3,4,3,2,3,2,1,1,0},
  {3,7,5,4,6,5,4,3,3,2,2,1,0},
  {5,4,3,7,6,5,4,3,2,1,1,0},
  {1,1,7,6,5,4,3,2,1,1,0},
  {1,1,5,4,3,3,2,1,1,0},
  {1,1,1,3,3,2,2,1,0},
  {1,0,1,3,2,1,1,1,},
  {1,0,1,3,2,1,1,},
  {0,1,1,2,1,3},
  {0,1,1,1,1},
  {0,1,1,1},
  {0,1,1},
  {0,1},
};

const UChar g_aucLenTableTZ16[15][16] =
{

  { 1,3,3,4,4,5,5,6,6,7,7,8,8,9,9,9},
  { 3,3,3,3,3,4,4,4,4,5,5,6,6,6,6},
  { 4,3,3,3,4,4,3,3,4,5,5,6,5,6},
  { 5,3,4,4,3,3,3,4,3,4,5,5,5},
  { 4,4,4,3,3,3,3,3,4,5,4,5},
  { 6,5,3,3,3,3,3,3,4,3,6},
  { 6,5,3,3,3,2,3,4,3,6},
  { 6,4,5,3,2,2,3,3,6},
  { 6,6,4,2,2,3,2,5},
  { 5,5,3,2,2,2,4},
  { 4,4,3,3,1,3},
  { 4,4,2,1,3},
  { 3,3,1,2},
  { 2,2,1},
  { 1,1},
};

const UChar g_aucCodeTableTO4[4][5] =
{
  {1,7,4,3,2},
  {0,1,6,3,3},
  {0,0,1,2,2},
  {0,0,0,5,0},
};

const UChar g_aucLenTableTO4[4][5] =
{
  { 2, 6, 6, 6, 6,},
  { 0, 1, 6, 7, 8,},
  { 0, 0, 3, 7, 8,},
  { 0, 0, 0, 6, 7,},
};

const UChar g_aucACTab[6] =
{
  0,1,2,0,1,2
};


//--> srlee : calc. refIdxLX bits
const UChar g_aucFrameBits[32] =
{
  0,
  1,
  3, 3,
  5, 5, 5, 5,
  7, 7, 7, 7, 7, 7, 7, 7,
  9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9
};

UInt g_aiQuantCoef4096[6]
 = {26, 23, 20, 18, 16, 14}; // 9
// = {51, 47, 39, 37, 32, 28}; // 10
// = {102, 93, 79, 73, 64, 57}; // 11
UInt g_aiDeQuantCoef4096[6]
 = {10, 11, 13, 14, 16, 18}; // 9
// = {20, 22, 26, 28, 32, 36}; // 10
// = {40, 44, 52, 56, 64, 72}; // 11

UInt g_aiQuantCoef1024_Loeffler_Lifting[6][1024] =
{
	{
		51, 102, 102, 71, 51, 74, 78, 67, 78, 78, 67, 87, 72, 60, 72, 73, 102, 72, 73, 88, 73, 60, 78, 67, 67, 77, 67, 73, 101, 74, 50, 53, 
			102, 204, 204, 142, 103, 149, 156, 133, 157, 158, 129, 182, 147, 119, 143, 151, 212, 143, 145, 174, 144, 120, 157, 133, 134, 157, 133, 141, 203, 147, 103, 101, 
			102, 204, 205, 141, 103, 144, 160, 128, 152, 159, 132, 175, 144, 120, 144, 145, 205, 144, 145, 182, 147, 119, 160, 129, 129, 159, 132, 143, 204, 148, 102, 102, 
			71, 142, 141, 99, 69, 99, 108, 93, 109, 109, 93, 121, 101, 85, 98, 101, 142, 100, 101, 121, 100, 84, 110, 100, 93, 108, 95, 98, 142, 102, 71, 71, 
			51, 103, 103, 69, 49, 74, 79, 67, 78, 79, 68, 85, 71, 60, 73, 72, 102, 72, 70, 85, 71, 60, 79, 67, 67, 85, 67, 69, 100, 74, 51, 51, 
			74, 149, 144, 99, 74, 106, 113, 96, 112, 134, 98, 125, 104, 87, 106, 110, 152, 103, 105, 125, 105, 87, 135, 96, 97, 114, 90, 105, 140, 105, 75, 74, 
			78, 156, 160, 108, 79, 113, 120, 105, 121, 120, 103, 134, 113, 94, 111, 111, 157, 108, 111, 132, 111, 92, 112, 102, 103, 120, 103, 129, 155, 114, 78, 79, 
			67, 133, 128, 93, 67, 96, 105, 88, 102, 102, 87, 117, 94, 79, 95, 98, 132, 95, 95, 111, 97, 79, 103, 87, 88, 104, 87, 92, 133, 96, 67, 67, 
			78, 157, 152, 109, 78, 112, 121, 102, 120, 121, 102, 133, 110, 91, 112, 111, 157, 117, 111, 133, 111, 93, 121, 102, 102, 122, 103, 109, 155, 113, 79, 78, 
			78, 158, 159, 109, 79, 134, 120, 102, 121, 119, 102, 133, 112, 92, 111, 108, 157, 111, 110, 134, 111, 92, 112, 103, 102, 123, 102, 109, 156, 113, 79, 78, 
			67, 129, 132, 93, 68, 98, 103, 87, 102, 102, 89, 115, 94, 78, 94, 95, 134, 92, 94, 114, 97, 79, 102, 86, 87, 102, 82, 94, 133, 90, 65, 67, 
			87, 182, 175, 121, 85, 125, 134, 117, 133, 133, 115, 148, 124, 102, 123, 122, 174, 123, 120, 148, 123, 102, 133, 114, 114, 133, 114, 122, 174, 124, 92, 87, 
			72, 147, 144, 101, 71, 104, 113, 94, 110, 112, 94, 124, 104, 85, 101, 102, 145, 105, 102, 126, 103, 85, 115, 94, 94, 113, 94, 98, 146, 105, 74, 72, 
			60, 119, 120, 85, 60, 87, 94, 79, 91, 92, 78, 102, 85, 79, 85, 85, 121, 85, 90, 102, 86, 71, 92, 78, 79, 90, 79, 82, 122, 87, 59, 60, 
			72, 143, 144, 98, 73, 106, 111, 95, 112, 111, 94, 123, 101, 85, 101, 102, 144, 102, 102, 118, 103, 92, 111, 96, 95, 110, 97, 101, 140, 104, 73, 72, 
			73, 151, 145, 101, 72, 110, 111, 98, 111, 108, 95, 122, 102, 85, 102, 102, 144, 103, 102, 123, 102, 85, 107, 94, 99, 111, 97, 100, 145, 104, 71, 73, 
			102, 212, 205, 142, 102, 152, 157, 132, 157, 157, 134, 174, 145, 121, 144, 144, 205, 145, 145, 174, 145, 121, 155, 134, 134, 157, 135, 142, 203, 148, 100, 103, 
			72, 143, 144, 100, 72, 103, 108, 95, 117, 111, 92, 123, 105, 85, 102, 103, 145, 103, 103, 122, 102, 90, 111, 100, 95, 107, 94, 101, 144, 104, 73, 72, 
			73, 145, 145, 101, 70, 105, 111, 95, 111, 110, 94, 120, 102, 90, 102, 102, 145, 103, 103, 123, 100, 85, 110, 95, 94, 107, 94, 96, 145, 105, 72, 73, 
			88, 174, 182, 121, 85, 125, 132, 111, 133, 134, 114, 148, 126, 102, 118, 123, 174, 122, 123, 152, 123, 102, 137, 114, 111, 133, 113, 116, 174, 132, 87, 88, 
			73, 144, 147, 100, 71, 105, 111, 97, 111, 111, 97, 123, 103, 86, 103, 102, 145, 102, 100, 123, 102, 85, 111, 95, 95, 111, 95, 101, 145, 104, 72, 73, 
			60, 120, 119, 84, 60, 87, 92, 79, 93, 92, 79, 102, 85, 71, 92, 85, 121, 90, 85, 102, 85, 71, 92, 79, 79, 91, 73, 83, 121, 81, 61, 60, 
			78, 157, 160, 110, 79, 135, 112, 103, 121, 112, 102, 133, 115, 92, 111, 107, 155, 111, 110, 137, 111, 92, 105, 103, 103, 115, 102, 109, 158, 113, 78, 78, 
			67, 133, 129, 100, 67, 96, 102, 87, 102, 103, 86, 114, 94, 78, 96, 94, 134, 100, 95, 114, 95, 79, 103, 88, 87, 103, 87, 93, 143, 96, 67, 67, 
			67, 134, 129, 93, 67, 97, 103, 88, 102, 102, 87, 114, 94, 79, 95, 99, 134, 95, 94, 111, 95, 79, 103, 87, 87, 102, 87, 92, 133, 97, 67, 67, 
			77, 157, 159, 108, 85, 114, 120, 104, 122, 123, 102, 133, 113, 90, 110, 111, 157, 107, 107, 133, 111, 91, 115, 103, 102, 120, 104, 139, 155, 113, 79, 77, 
			67, 133, 132, 95, 67, 90, 103, 87, 103, 102, 82, 114, 94, 79, 97, 97, 135, 94, 94, 113, 95, 73, 102, 87, 87, 104, 76, 94, 134, 96, 67, 67, 
			73, 141, 143, 98, 69, 105, 129, 92, 109, 109, 94, 122, 98, 82, 101, 100, 142, 101, 96, 116, 101, 83, 109, 93, 92, 139, 94, 98, 143, 103, 71, 73, 
			101, 203, 204, 142, 100, 140, 155, 133, 155, 156, 133, 174, 146, 122, 140, 145, 203, 144, 145, 174, 145, 121, 158, 143, 133, 155, 134, 143, 206, 148, 102, 101, 
			74, 147, 148, 102, 74, 105, 114, 96, 113, 113, 90, 124, 105, 87, 104, 104, 148, 104, 105, 132, 104, 81, 113, 96, 97, 113, 96, 103, 148, 107, 73, 74, 
			50, 103, 102, 71, 51, 75, 78, 67, 79, 79, 65, 92, 74, 59, 73, 71, 100, 73, 72, 87, 72, 61, 78, 67, 67, 79, 67, 71, 102, 73, 52, 49, 
			53, 101, 102, 71, 51, 74, 79, 67, 78, 78, 67, 87, 72, 60, 72, 73, 103, 72, 73, 88, 73, 60, 78, 67, 67, 77, 67, 73, 101, 74, 49, 54, 
	},
	{
		47, 93, 93, 65, 47, 67, 71, 61, 71, 71, 61, 79, 66, 55, 66, 66, 93, 66, 66, 80, 66, 54, 71, 61, 61, 70, 61, 66, 92, 67, 45, 48, 
			93, 186, 185, 129, 93, 135, 142, 121, 143, 144, 117, 166, 134, 108, 130, 137, 193, 130, 132, 158, 131, 109, 143, 121, 121, 143, 121, 128, 185, 134, 94, 92, 
			93, 185, 186, 128, 93, 130, 146, 116, 138, 144, 120, 159, 131, 109, 131, 132, 186, 131, 132, 165, 134, 108, 145, 117, 117, 145, 120, 130, 185, 134, 93, 93, 
			65, 129, 128, 90, 63, 90, 98, 85, 99, 99, 84, 110, 92, 77, 89, 92, 129, 91, 91, 110, 91, 76, 100, 91, 84, 98, 86, 89, 129, 93, 65, 64, 
			47, 93, 93, 63, 45, 67, 72, 61, 71, 72, 61, 78, 65, 55, 66, 66, 93, 66, 64, 77, 64, 55, 72, 61, 61, 77, 60, 63, 91, 67, 47, 47, 
			67, 135, 130, 90, 67, 96, 103, 88, 102, 122, 89, 114, 95, 80, 97, 100, 139, 94, 95, 113, 95, 79, 122, 87, 88, 104, 81, 95, 128, 95, 68, 67, 
			71, 142, 146, 98, 72, 103, 109, 95, 110, 109, 93, 122, 102, 85, 101, 101, 142, 98, 101, 120, 101, 84, 102, 93, 94, 109, 94, 117, 141, 103, 71, 72, 
			61, 121, 116, 85, 61, 88, 95, 80, 93, 93, 79, 107, 86, 72, 87, 89, 120, 86, 86, 101, 88, 72, 93, 79, 80, 94, 79, 84, 121, 87, 60, 61, 
			71, 143, 138, 99, 71, 102, 110, 93, 109, 110, 93, 121, 100, 83, 102, 101, 142, 106, 101, 121, 101, 85, 110, 93, 93, 111, 94, 99, 141, 103, 72, 71, 
			71, 144, 144, 99, 72, 122, 109, 93, 110, 108, 93, 121, 102, 84, 101, 98, 143, 101, 100, 122, 101, 84, 102, 93, 93, 112, 92, 99, 142, 102, 72, 71, 
			61, 117, 120, 84, 61, 89, 93, 79, 93, 93, 81, 105, 86, 71, 85, 86, 122, 83, 86, 103, 88, 72, 93, 78, 79, 93, 75, 85, 121, 82, 59, 61, 
			79, 166, 159, 110, 78, 114, 122, 107, 121, 121, 105, 135, 113, 93, 112, 111, 158, 112, 109, 135, 112, 93, 121, 103, 104, 121, 103, 111, 158, 113, 83, 79, 
			66, 134, 131, 92, 65, 95, 102, 86, 100, 102, 86, 113, 95, 77, 92, 93, 131, 95, 93, 114, 94, 77, 104, 86, 86, 103, 86, 89, 133, 95, 67, 66, 
			55, 108, 109, 77, 55, 80, 85, 72, 83, 84, 71, 93, 77, 72, 77, 78, 110, 78, 82, 93, 78, 64, 84, 71, 72, 82, 72, 75, 111, 79, 54, 55, 
			66, 130, 131, 89, 66, 97, 101, 87, 102, 101, 85, 112, 92, 77, 92, 93, 131, 93, 93, 108, 93, 84, 101, 87, 86, 100, 88, 92, 127, 95, 66, 66, 
			66, 137, 132, 92, 66, 100, 101, 89, 101, 98, 86, 111, 93, 78, 93, 92, 131, 93, 93, 111, 93, 78, 97, 86, 90, 101, 88, 91, 131, 95, 65, 66, 
			93, 193, 186, 129, 93, 139, 142, 120, 142, 143, 122, 158, 131, 110, 131, 131, 186, 132, 132, 158, 132, 110, 141, 121, 122, 143, 123, 129, 184, 134, 91, 93, 
			66, 130, 131, 91, 66, 94, 98, 86, 106, 101, 83, 112, 95, 78, 93, 93, 132, 94, 93, 111, 93, 82, 101, 91, 86, 97, 86, 92, 131, 95, 66, 66, 
			66, 132, 132, 91, 64, 95, 101, 86, 101, 100, 86, 109, 93, 82, 93, 93, 132, 93, 94, 112, 91, 77, 100, 86, 86, 97, 86, 88, 131, 95, 66, 66, 
			80, 158, 165, 110, 77, 113, 120, 101, 121, 122, 103, 135, 114, 93, 108, 111, 158, 111, 112, 139, 112, 93, 124, 103, 101, 121, 103, 105, 158, 120, 79, 80, 
			66, 131, 134, 91, 64, 95, 101, 88, 101, 101, 88, 112, 94, 78, 93, 93, 132, 93, 91, 112, 93, 78, 101, 86, 86, 101, 86, 92, 132, 95, 66, 66, 
			54, 109, 108, 76, 55, 79, 84, 72, 85, 84, 72, 93, 77, 64, 84, 78, 110, 82, 77, 93, 78, 64, 84, 72, 71, 83, 66, 76, 110, 74, 55, 54, 
			71, 143, 145, 100, 72, 122, 102, 93, 110, 102, 93, 121, 104, 84, 101, 97, 141, 101, 100, 124, 101, 84, 96, 94, 93, 105, 93, 99, 144, 103, 71, 71, 
			61, 121, 117, 91, 61, 87, 93, 79, 93, 93, 78, 103, 86, 71, 87, 86, 121, 91, 86, 103, 86, 72, 94, 80, 79, 93, 80, 85, 130, 87, 61, 61, 
			61, 121, 117, 84, 61, 88, 94, 80, 93, 93, 79, 104, 86, 72, 86, 90, 122, 86, 86, 101, 86, 71, 93, 79, 80, 93, 79, 84, 121, 88, 61, 61, 
			70, 143, 145, 98, 77, 104, 109, 94, 111, 112, 93, 121, 103, 82, 100, 101, 143, 97, 97, 121, 101, 83, 105, 93, 93, 109, 95, 126, 141, 103, 72, 70, 
			61, 121, 120, 86, 60, 81, 94, 79, 94, 92, 75, 103, 86, 72, 88, 88, 123, 86, 86, 103, 86, 66, 93, 80, 79, 95, 69, 86, 122, 88, 61, 61, 
			66, 128, 130, 89, 63, 95, 117, 84, 99, 99, 85, 111, 89, 75, 92, 91, 129, 92, 88, 105, 92, 76, 99, 85, 84, 126, 86, 89, 130, 93, 64, 66, 
			92, 185, 185, 129, 91, 128, 141, 121, 141, 142, 121, 158, 133, 111, 127, 131, 184, 131, 131, 158, 132, 110, 144, 130, 121, 141, 122, 130, 187, 134, 93, 92, 
			67, 134, 134, 93, 67, 95, 103, 87, 103, 102, 82, 113, 95, 79, 95, 95, 134, 95, 95, 120, 95, 74, 103, 87, 88, 103, 88, 93, 134, 97, 67, 67, 
			45, 94, 93, 65, 47, 68, 71, 60, 72, 72, 59, 83, 67, 54, 66, 65, 91, 66, 66, 79, 66, 55, 71, 61, 61, 72, 61, 64, 93, 67, 47, 45, 
			48, 92, 93, 64, 47, 67, 72, 61, 71, 71, 61, 79, 66, 55, 66, 66, 93, 66, 66, 80, 66, 54, 71, 61, 61, 70, 61, 66, 92, 67, 45, 49, 
		},
		{
			39, 79, 79, 55, 39, 57, 60, 52, 60, 60, 52, 67, 56, 46, 56, 56, 79, 55, 56, 67, 56, 46, 60, 52, 51, 59, 52, 56, 78, 57, 38, 40, 
				79, 157, 157, 109, 79, 114, 120, 102, 121, 122, 99, 140, 113, 92, 110, 116, 163, 110, 112, 134, 111, 93, 121, 103, 103, 121, 103, 108, 157, 113, 79, 78, 
				79, 157, 158, 108, 79, 110, 123, 98, 117, 122, 101, 135, 111, 93, 111, 111, 157, 111, 112, 140, 113, 91, 123, 99, 99, 122, 102, 110, 157, 114, 78, 78, 
				55, 109, 108, 76, 53, 76, 83, 72, 84, 84, 71, 93, 78, 65, 75, 78, 109, 77, 77, 93, 77, 65, 85, 77, 71, 83, 73, 75, 110, 78, 55, 55, 
				39, 79, 79, 53, 38, 57, 61, 51, 60, 61, 52, 66, 55, 46, 56, 56, 79, 56, 54, 65, 54, 46, 61, 52, 51, 65, 51, 53, 77, 57, 39, 39, 
				57, 114, 110, 76, 57, 81, 87, 74, 86, 103, 76, 96, 80, 67, 82, 84, 117, 80, 80, 96, 81, 67, 104, 73, 74, 88, 69, 81, 108, 81, 57, 57, 
				60, 120, 123, 83, 61, 87, 92, 81, 93, 92, 79, 103, 87, 72, 86, 85, 120, 83, 85, 102, 85, 71, 86, 79, 79, 92, 79, 99, 120, 87, 60, 61, 
				52, 102, 98, 72, 51, 74, 81, 68, 79, 79, 67, 90, 72, 61, 73, 75, 102, 73, 73, 85, 75, 61, 79, 67, 68, 80, 67, 71, 102, 74, 51, 52, 
				60, 121, 117, 84, 60, 86, 93, 79, 92, 93, 79, 102, 85, 70, 86, 85, 120, 90, 86, 102, 85, 72, 93, 79, 79, 94, 79, 84, 119, 87, 61, 60, 
				60, 122, 122, 84, 61, 103, 92, 79, 93, 92, 79, 102, 86, 71, 86, 83, 121, 85, 85, 103, 85, 71, 86, 79, 79, 95, 78, 84, 120, 87, 61, 60, 
				52, 99, 101, 71, 52, 76, 79, 67, 79, 79, 69, 89, 72, 60, 72, 73, 103, 71, 73, 87, 74, 61, 78, 66, 67, 79, 63, 72, 103, 69, 50, 52, 
				67, 140, 135, 93, 66, 96, 103, 90, 102, 102, 89, 114, 96, 79, 95, 94, 134, 95, 92, 114, 95, 78, 102, 87, 88, 103, 87, 94, 134, 95, 70, 67, 
				56, 113, 111, 78, 55, 80, 87, 72, 85, 86, 72, 96, 80, 65, 78, 79, 111, 81, 79, 97, 79, 65, 88, 73, 73, 87, 73, 75, 112, 80, 57, 56, 
				46, 92, 93, 65, 46, 67, 72, 61, 70, 71, 60, 79, 65, 61, 65, 66, 93, 66, 69, 79, 66, 54, 71, 60, 61, 69, 61, 63, 94, 67, 46, 46, 
				56, 110, 111, 75, 56, 82, 86, 73, 86, 86, 72, 95, 78, 65, 78, 78, 111, 78, 79, 91, 79, 71, 86, 74, 73, 85, 74, 78, 108, 80, 56, 56, 
				56, 116, 111, 78, 56, 84, 85, 75, 85, 83, 73, 94, 79, 66, 78, 78, 111, 79, 79, 94, 79, 66, 82, 73, 76, 85, 75, 77, 111, 80, 55, 56, 
				79, 163, 157, 109, 79, 117, 120, 102, 120, 121, 103, 134, 111, 93, 111, 111, 158, 112, 112, 133, 112, 93, 119, 103, 103, 121, 104, 109, 156, 114, 77, 79, 
				55, 110, 111, 77, 56, 80, 83, 73, 90, 85, 71, 95, 81, 66, 78, 79, 112, 79, 79, 94, 79, 69, 85, 77, 73, 82, 72, 77, 111, 80, 56, 56, 
				56, 112, 112, 77, 54, 80, 85, 73, 86, 85, 73, 92, 79, 69, 79, 79, 112, 79, 79, 95, 77, 65, 85, 73, 73, 82, 73, 74, 111, 81, 56, 56, 
				67, 134, 140, 93, 65, 96, 102, 85, 102, 103, 87, 114, 97, 79, 91, 94, 133, 94, 95, 117, 95, 79, 105, 88, 85, 103, 87, 89, 134, 102, 67, 68, 
				56, 111, 113, 77, 54, 81, 85, 75, 85, 85, 74, 95, 79, 66, 79, 79, 112, 79, 77, 95, 78, 66, 85, 73, 73, 85, 73, 78, 111, 80, 56, 56, 
				46, 93, 91, 65, 46, 67, 71, 61, 72, 71, 61, 78, 65, 54, 71, 66, 93, 69, 65, 79, 66, 54, 71, 61, 60, 70, 56, 64, 93, 63, 47, 46, 
				60, 121, 123, 85, 61, 104, 86, 79, 93, 86, 78, 102, 88, 71, 86, 82, 119, 85, 85, 105, 85, 71, 81, 79, 79, 89, 79, 84, 122, 87, 60, 60, 
				52, 103, 99, 77, 52, 73, 79, 67, 79, 79, 66, 87, 73, 60, 74, 73, 103, 77, 73, 88, 73, 61, 79, 67, 67, 79, 67, 72, 110, 74, 51, 52, 
				51, 103, 99, 71, 51, 74, 79, 68, 79, 79, 67, 88, 73, 61, 73, 76, 103, 73, 73, 85, 73, 60, 79, 67, 67, 79, 67, 71, 102, 74, 52, 52, 
				59, 121, 122, 83, 65, 88, 92, 80, 94, 95, 79, 103, 87, 69, 85, 85, 121, 82, 82, 103, 85, 70, 89, 79, 79, 92, 80, 107, 119, 87, 61, 60, 
				52, 103, 102, 73, 51, 69, 79, 67, 79, 78, 63, 87, 73, 61, 74, 75, 104, 72, 73, 87, 73, 56, 79, 67, 67, 80, 58, 72, 103, 74, 52, 52, 
				56, 108, 110, 75, 53, 81, 99, 71, 84, 84, 72, 94, 75, 63, 78, 77, 109, 77, 74, 89, 78, 64, 84, 72, 71, 107, 72, 75, 110, 79, 54, 56, 
				78, 157, 157, 110, 77, 108, 120, 102, 119, 120, 103, 134, 112, 94, 108, 111, 156, 111, 111, 134, 111, 93, 122, 110, 102, 119, 103, 110, 158, 114, 78, 78, 
				57, 113, 114, 78, 57, 81, 87, 74, 87, 87, 69, 95, 80, 67, 80, 80, 114, 80, 81, 102, 80, 63, 87, 74, 74, 87, 74, 79, 114, 82, 57, 57, 
				38, 79, 78, 55, 39, 57, 60, 51, 61, 61, 50, 70, 57, 46, 56, 55, 77, 56, 56, 67, 56, 47, 60, 51, 52, 61, 52, 54, 78, 57, 40, 38, 
				40, 78, 78, 55, 39, 57, 61, 52, 60, 60, 52, 67, 56, 46, 56, 56, 79, 56, 56, 68, 56, 46, 60, 52, 52, 60, 52, 56, 78, 57, 38, 42, 
		},
		{
			37, 73, 73, 51, 37, 53, 56, 48, 56, 56, 48, 62, 52, 43, 52, 52, 73, 52, 52, 63, 52, 43, 55, 48, 48, 55, 48, 52, 72, 53, 36, 38, 
				73, 146, 146, 101, 73, 106, 112, 95, 112, 113, 92, 130, 105, 85, 102, 108, 151, 102, 104, 124, 103, 86, 112, 95, 95, 112, 95, 101, 145, 105, 74, 72, 
				73, 146, 146, 101, 73, 103, 115, 91, 108, 113, 94, 125, 103, 86, 103, 103, 146, 103, 104, 130, 105, 85, 114, 92, 92, 114, 94, 102, 146, 106, 73, 73, 
				51, 101, 101, 70, 49, 71, 77, 66, 78, 78, 66, 86, 72, 61, 70, 72, 101, 72, 72, 86, 72, 60, 79, 71, 66, 77, 68, 70, 102, 73, 51, 51, 
				37, 73, 73, 49, 35, 53, 56, 48, 56, 56, 48, 61, 51, 43, 52, 52, 73, 52, 50, 61, 51, 43, 56, 48, 48, 61, 48, 50, 72, 53, 37, 37, 
				53, 106, 103, 71, 53, 75, 81, 69, 80, 96, 70, 89, 75, 62, 76, 78, 109, 74, 75, 89, 75, 62, 96, 68, 69, 81, 64, 75, 100, 75, 53, 53, 
				56, 112, 115, 77, 56, 81, 86, 75, 87, 86, 73, 96, 80, 67, 79, 79, 112, 77, 79, 95, 79, 66, 80, 73, 74, 85, 74, 92, 111, 81, 56, 56, 
				48, 95, 91, 66, 48, 69, 75, 63, 73, 73, 62, 84, 67, 56, 68, 70, 95, 68, 68, 79, 69, 56, 73, 62, 63, 74, 62, 66, 95, 68, 48, 48, 
				56, 112, 108, 78, 56, 80, 87, 73, 86, 86, 73, 95, 79, 65, 80, 79, 112, 83, 79, 95, 79, 67, 87, 73, 73, 87, 74, 78, 111, 81, 56, 56, 
				56, 113, 113, 78, 56, 96, 86, 73, 86, 85, 73, 95, 80, 66, 79, 77, 112, 79, 79, 96, 79, 66, 80, 73, 73, 88, 73, 78, 111, 80, 56, 56, 
				48, 92, 94, 66, 48, 70, 73, 62, 73, 73, 64, 82, 67, 56, 67, 68, 96, 66, 67, 81, 69, 57, 73, 62, 62, 73, 59, 67, 95, 64, 46, 48, 
				62, 130, 125, 86, 61, 89, 96, 84, 95, 95, 82, 106, 89, 73, 88, 87, 124, 88, 86, 106, 88, 73, 95, 81, 82, 95, 81, 87, 124, 88, 65, 62, 
				52, 105, 103, 72, 51, 75, 80, 67, 79, 80, 67, 89, 74, 61, 72, 73, 103, 75, 73, 90, 74, 61, 82, 67, 67, 81, 67, 70, 104, 75, 53, 52, 
				43, 85, 86, 61, 43, 62, 67, 56, 65, 66, 56, 73, 61, 56, 61, 61, 86, 61, 64, 73, 61, 51, 66, 56, 56, 64, 56, 59, 87, 62, 42, 43, 
				52, 102, 103, 70, 52, 76, 79, 68, 80, 79, 67, 88, 72, 61, 72, 73, 103, 73, 73, 84, 73, 66, 79, 69, 68, 79, 69, 72, 100, 74, 52, 52, 
				52, 108, 103, 72, 52, 78, 79, 70, 79, 77, 68, 87, 73, 61, 73, 73, 103, 73, 73, 88, 73, 61, 76, 67, 71, 79, 69, 72, 103, 74, 51, 52, 
				73, 151, 146, 101, 73, 109, 112, 95, 112, 112, 96, 124, 103, 86, 103, 103, 146, 104, 104, 124, 104, 86, 111, 95, 96, 112, 96, 101, 145, 105, 71, 73, 
				52, 102, 103, 72, 52, 74, 77, 68, 83, 79, 66, 88, 75, 61, 73, 73, 104, 74, 73, 87, 73, 64, 79, 71, 68, 76, 67, 72, 103, 74, 52, 52, 
				52, 104, 104, 72, 50, 75, 79, 68, 79, 79, 67, 86, 73, 64, 73, 73, 104, 73, 74, 88, 72, 61, 79, 68, 67, 76, 67, 69, 103, 75, 52, 52, 
				63, 124, 130, 86, 61, 89, 95, 79, 95, 96, 81, 106, 90, 73, 84, 88, 124, 87, 88, 109, 88, 73, 98, 81, 79, 95, 81, 83, 125, 94, 62, 63, 
				52, 103, 105, 72, 51, 75, 79, 69, 79, 79, 69, 88, 74, 61, 73, 73, 104, 73, 72, 88, 73, 61, 79, 68, 68, 79, 68, 72, 103, 74, 52, 52, 
				43, 86, 85, 60, 43, 62, 66, 56, 67, 66, 57, 73, 61, 51, 66, 61, 86, 64, 61, 73, 61, 51, 66, 56, 56, 65, 52, 60, 87, 58, 43, 43, 
				55, 112, 114, 79, 56, 96, 80, 73, 87, 80, 73, 95, 82, 66, 79, 76, 111, 79, 79, 98, 79, 66, 75, 74, 73, 82, 73, 78, 113, 81, 56, 55, 
				48, 95, 92, 71, 48, 68, 73, 62, 73, 73, 62, 81, 67, 56, 69, 67, 95, 71, 68, 81, 68, 56, 74, 63, 62, 73, 62, 67, 102, 69, 48, 48, 
				48, 95, 92, 66, 48, 69, 74, 63, 73, 73, 62, 82, 67, 56, 68, 71, 96, 68, 67, 79, 68, 56, 73, 62, 62, 73, 62, 66, 95, 69, 48, 48, 
				55, 112, 114, 77, 61, 81, 85, 74, 87, 88, 73, 95, 81, 64, 79, 79, 112, 76, 76, 95, 79, 65, 82, 73, 73, 85, 74, 99, 111, 81, 56, 55, 
				48, 95, 94, 68, 48, 64, 74, 62, 74, 73, 59, 81, 67, 56, 69, 69, 96, 67, 67, 81, 68, 52, 73, 62, 62, 74, 54, 67, 96, 69, 48, 48, 
				52, 101, 102, 70, 50, 75, 92, 66, 78, 78, 67, 87, 70, 59, 72, 72, 101, 72, 69, 83, 72, 60, 78, 67, 66, 99, 67, 70, 102, 73, 50, 52, 
				72, 145, 146, 102, 72, 100, 111, 95, 111, 111, 95, 124, 104, 87, 100, 103, 145, 103, 103, 125, 103, 87, 113, 102, 95, 111, 96, 102, 147, 106, 73, 72, 
				53, 105, 106, 73, 53, 75, 81, 68, 81, 80, 64, 88, 75, 62, 74, 74, 105, 74, 75, 94, 74, 58, 81, 69, 69, 81, 69, 73, 106, 76, 52, 53, 
				36, 74, 73, 51, 37, 53, 56, 48, 56, 56, 46, 65, 53, 42, 52, 51, 71, 52, 52, 62, 52, 43, 56, 48, 48, 56, 48, 50, 73, 52, 37, 35, 
				38, 72, 73, 51, 37, 53, 56, 48, 56, 56, 48, 62, 52, 43, 52, 52, 73, 52, 52, 63, 52, 43, 55, 48, 48, 55, 48, 52, 72, 53, 35, 39, 
			},
			{
				32, 64, 64, 44, 32, 46, 49, 42, 49, 49, 42, 54, 45, 38, 45, 45, 64, 45, 45, 55, 45, 37, 49, 42, 42, 48, 42, 46, 63, 46, 31, 33, 
					64, 128, 127, 89, 64, 93, 98, 83, 98, 99, 81, 114, 92, 75, 89, 94, 132, 90, 91, 109, 90, 75, 98, 83, 83, 98, 83, 88, 127, 92, 65, 63, 
					64, 127, 128, 88, 64, 90, 100, 80, 95, 99, 82, 109, 90, 75, 90, 90, 128, 90, 91, 114, 92, 74, 100, 80, 80, 99, 83, 89, 128, 92, 64, 64, 
					44, 89, 88, 62, 43, 62, 68, 58, 68, 68, 58, 75, 63, 53, 61, 63, 89, 63, 63, 75, 63, 52, 69, 62, 58, 67, 59, 61, 89, 64, 44, 44, 
					32, 64, 64, 43, 31, 46, 49, 42, 49, 49, 42, 53, 45, 38, 46, 45, 64, 45, 44, 53, 44, 38, 49, 42, 42, 53, 42, 43, 63, 46, 32, 32, 
					46, 93, 90, 62, 46, 66, 71, 60, 70, 84, 61, 78, 65, 55, 66, 69, 95, 65, 65, 78, 65, 54, 84, 60, 60, 71, 56, 66, 88, 65, 47, 46, 
					49, 98, 100, 68, 49, 71, 75, 65, 76, 75, 64, 84, 70, 59, 70, 69, 98, 67, 69, 83, 69, 57, 70, 64, 64, 75, 64, 80, 97, 71, 49, 49, 
					42, 83, 80, 58, 42, 60, 65, 55, 64, 64, 54, 73, 59, 49, 60, 61, 83, 59, 59, 69, 61, 49, 64, 54, 55, 65, 54, 58, 83, 60, 42, 42, 
					49, 98, 95, 68, 49, 70, 76, 64, 75, 75, 64, 83, 69, 57, 70, 69, 98, 73, 69, 83, 69, 58, 76, 64, 64, 76, 64, 68, 97, 71, 49, 49, 
					49, 99, 99, 68, 49, 84, 75, 64, 75, 75, 64, 83, 70, 58, 69, 68, 98, 69, 69, 84, 69, 58, 70, 64, 64, 77, 64, 68, 97, 70, 49, 49, 
					42, 81, 82, 58, 42, 61, 64, 54, 64, 64, 56, 72, 59, 49, 59, 59, 84, 57, 59, 71, 60, 50, 64, 54, 54, 64, 51, 59, 83, 56, 41, 42, 
					54, 114, 109, 75, 53, 78, 84, 73, 83, 83, 72, 93, 78, 64, 77, 76, 108, 77, 75, 93, 77, 64, 83, 71, 71, 83, 71, 76, 109, 77, 57, 55, 
					45, 92, 90, 63, 45, 65, 70, 59, 69, 70, 59, 78, 65, 53, 63, 64, 90, 66, 64, 79, 64, 53, 72, 59, 59, 71, 59, 61, 91, 65, 46, 45, 
					38, 75, 75, 53, 38, 55, 59, 49, 57, 58, 49, 64, 53, 49, 53, 53, 76, 53, 56, 64, 54, 44, 58, 49, 49, 56, 49, 52, 76, 54, 37, 38, 
					45, 89, 90, 61, 46, 66, 70, 60, 70, 69, 59, 77, 63, 53, 63, 64, 90, 64, 64, 74, 64, 58, 69, 60, 59, 69, 60, 63, 88, 65, 45, 45, 
					45, 94, 90, 63, 45, 69, 69, 61, 69, 68, 59, 76, 64, 53, 64, 64, 90, 64, 64, 77, 64, 53, 67, 59, 62, 69, 61, 63, 90, 65, 44, 45, 
					64, 132, 128, 89, 64, 95, 98, 83, 98, 98, 84, 108, 90, 76, 90, 90, 128, 91, 91, 108, 91, 75, 97, 84, 84, 98, 84, 89, 127, 92, 62, 64, 
					45, 90, 90, 63, 45, 65, 67, 59, 73, 69, 57, 77, 66, 53, 64, 64, 91, 64, 64, 76, 64, 56, 69, 63, 59, 67, 59, 63, 90, 65, 46, 45, 
					45, 91, 91, 63, 44, 65, 69, 59, 69, 69, 59, 75, 64, 56, 64, 64, 91, 64, 65, 77, 63, 53, 69, 59, 59, 67, 59, 60, 90, 65, 45, 46, 
					55, 109, 114, 75, 53, 78, 83, 69, 83, 84, 71, 93, 79, 64, 74, 77, 108, 76, 77, 95, 77, 64, 85, 71, 69, 83, 71, 72, 109, 83, 55, 55, 
					45, 90, 92, 63, 44, 65, 69, 61, 69, 69, 60, 77, 64, 54, 64, 64, 91, 64, 63, 77, 64, 53, 69, 59, 59, 69, 59, 63, 90, 65, 45, 45, 
					37, 75, 74, 52, 38, 54, 57, 49, 58, 58, 50, 64, 53, 44, 58, 53, 75, 56, 53, 64, 53, 44, 57, 49, 49, 57, 45, 52, 76, 51, 38, 37, 
					49, 98, 100, 69, 49, 84, 70, 64, 76, 70, 64, 83, 72, 58, 69, 67, 97, 69, 69, 85, 69, 57, 66, 64, 64, 72, 64, 68, 99, 71, 49, 49, 
					42, 83, 80, 62, 42, 60, 64, 54, 64, 64, 54, 71, 59, 49, 60, 59, 84, 63, 59, 71, 59, 49, 64, 55, 54, 64, 55, 58, 89, 60, 42, 42, 
					42, 83, 80, 58, 42, 60, 64, 55, 64, 64, 54, 71, 59, 49, 59, 62, 84, 59, 59, 69, 59, 49, 64, 54, 55, 64, 55, 58, 83, 60, 42, 42, 
					48, 98, 99, 67, 53, 71, 75, 65, 76, 77, 64, 83, 71, 56, 69, 69, 98, 67, 67, 83, 69, 57, 72, 64, 64, 75, 65, 87, 97, 70, 49, 48, 
					42, 83, 83, 59, 42, 56, 64, 54, 64, 64, 51, 71, 59, 49, 60, 61, 84, 59, 59, 71, 59, 45, 64, 55, 55, 65, 48, 59, 84, 60, 42, 42, 
					46, 88, 89, 61, 43, 66, 80, 58, 68, 68, 59, 76, 61, 52, 63, 63, 89, 63, 60, 72, 63, 52, 68, 58, 58, 87, 59, 61, 89, 64, 44, 46, 
					63, 127, 128, 89, 63, 88, 97, 83, 97, 97, 83, 109, 91, 76, 88, 90, 127, 90, 90, 109, 90, 76, 99, 89, 83, 97, 84, 89, 128, 92, 64, 63, 
					46, 92, 92, 64, 46, 65, 71, 60, 71, 70, 56, 77, 65, 54, 65, 65, 92, 65, 65, 83, 65, 51, 71, 60, 60, 70, 60, 64, 92, 67, 46, 46, 
					31, 65, 64, 44, 32, 47, 49, 42, 49, 49, 41, 57, 46, 37, 45, 44, 62, 46, 45, 55, 45, 38, 49, 42, 42, 49, 42, 44, 64, 46, 33, 31, 
					33, 63, 64, 44, 32, 46, 49, 42, 49, 49, 42, 55, 45, 38, 45, 45, 64, 45, 46, 55, 45, 37, 49, 42, 42, 48, 42, 46, 63, 46, 31, 34, 
			},
			{
				28, 57, 57, 39, 28, 41, 44, 37, 43, 44, 37, 48, 40, 33, 40, 40, 57, 40, 40, 49, 40, 33, 43, 37, 37, 43, 37, 41, 56, 41, 28, 29, 
					57, 113, 113, 79, 57, 83, 87, 74, 87, 88, 72, 101, 82, 66, 79, 84, 118, 80, 81, 97, 80, 67, 87, 74, 74, 87, 74, 78, 113, 82, 57, 56, 
					57, 113, 114, 78, 57, 80, 89, 71, 84, 88, 73, 97, 80, 67, 80, 80, 114, 80, 81, 101, 82, 66, 89, 71, 72, 88, 73, 80, 113, 82, 57, 57, 
					39, 79, 78, 55, 38, 55, 60, 52, 60, 60, 51, 67, 56, 47, 54, 56, 79, 56, 56, 67, 56, 47, 61, 56, 52, 60, 53, 54, 79, 57, 40, 39, 
					28, 57, 57, 38, 27, 41, 44, 37, 43, 44, 38, 47, 40, 33, 40, 40, 57, 40, 39, 47, 39, 33, 44, 37, 37, 47, 37, 39, 56, 41, 29, 28, 
					41, 83, 80, 55, 41, 59, 63, 54, 62, 74, 55, 70, 58, 49, 59, 61, 85, 57, 58, 69, 58, 48, 75, 53, 54, 63, 50, 58, 78, 58, 42, 41, 
					44, 87, 89, 60, 44, 63, 67, 58, 67, 67, 57, 74, 63, 52, 62, 62, 87, 60, 62, 74, 62, 51, 62, 57, 57, 66, 57, 71, 86, 63, 44, 44, 
					37, 74, 71, 52, 37, 54, 58, 49, 57, 57, 48, 65, 52, 44, 53, 54, 74, 53, 53, 61, 54, 44, 57, 48, 49, 58, 48, 51, 74, 53, 37, 37, 
					43, 87, 84, 60, 43, 62, 67, 57, 67, 67, 57, 74, 61, 51, 62, 62, 87, 65, 62, 74, 62, 52, 67, 57, 57, 68, 57, 61, 86, 63, 44, 44, 
					44, 88, 88, 60, 44, 74, 67, 57, 67, 66, 57, 74, 62, 51, 62, 60, 87, 62, 61, 75, 61, 51, 62, 57, 57, 68, 57, 61, 87, 63, 44, 44, 
					37, 72, 73, 51, 38, 55, 57, 48, 57, 57, 50, 64, 52, 44, 52, 53, 74, 51, 52, 63, 54, 44, 57, 48, 48, 57, 46, 52, 74, 50, 36, 37, 
					48, 101, 97, 67, 47, 70, 74, 65, 74, 74, 64, 82, 69, 57, 69, 68, 96, 68, 67, 82, 68, 57, 74, 63, 63, 74, 63, 68, 97, 69, 51, 49, 
					40, 82, 80, 56, 40, 58, 63, 52, 61, 62, 52, 69, 58, 47, 56, 57, 80, 58, 57, 70, 57, 47, 64, 52, 52, 63, 52, 54, 81, 58, 41, 40, 
					33, 66, 67, 47, 33, 49, 52, 44, 51, 51, 44, 57, 47, 44, 47, 47, 67, 47, 50, 57, 48, 39, 51, 43, 44, 50, 44, 46, 68, 48, 33, 33, 
					40, 79, 80, 54, 40, 59, 62, 53, 62, 62, 52, 69, 56, 47, 56, 57, 80, 57, 57, 66, 57, 51, 62, 53, 53, 61, 54, 56, 78, 58, 40, 40, 
					40, 84, 80, 56, 40, 61, 62, 54, 62, 60, 53, 68, 57, 47, 57, 56, 80, 57, 57, 68, 57, 47, 59, 52, 55, 62, 54, 56, 80, 58, 39, 40, 
					57, 118, 114, 79, 57, 85, 87, 74, 87, 87, 74, 96, 80, 67, 80, 80, 114, 81, 81, 96, 81, 67, 86, 74, 74, 87, 75, 79, 113, 82, 55, 57, 
					40, 80, 80, 56, 40, 57, 60, 53, 65, 62, 51, 68, 58, 47, 57, 57, 81, 57, 57, 68, 57, 50, 62, 56, 53, 59, 52, 56, 80, 58, 41, 40, 
					40, 81, 81, 56, 39, 58, 62, 53, 62, 61, 52, 67, 57, 50, 57, 57, 81, 57, 57, 68, 56, 47, 61, 53, 52, 59, 52, 53, 80, 58, 40, 40, 
					49, 97, 101, 67, 47, 69, 74, 61, 74, 75, 63, 82, 70, 57, 66, 68, 96, 68, 68, 85, 68, 57, 76, 63, 62, 74, 63, 64, 97, 73, 49, 49, 
					40, 80, 82, 56, 39, 58, 62, 54, 62, 61, 54, 68, 57, 48, 57, 57, 81, 57, 56, 68, 57, 47, 62, 53, 53, 62, 53, 56, 80, 58, 40, 40, 
					33, 67, 66, 47, 33, 48, 51, 44, 52, 51, 44, 57, 47, 39, 51, 47, 67, 50, 47, 57, 47, 39, 51, 44, 44, 51, 40, 46, 67, 45, 34, 33, 
					43, 87, 89, 61, 44, 75, 62, 57, 67, 62, 57, 74, 64, 51, 62, 59, 86, 62, 61, 76, 62, 51, 59, 57, 57, 64, 57, 60, 88, 63, 44, 43, 
					37, 74, 71, 56, 37, 53, 57, 48, 57, 57, 48, 63, 52, 43, 53, 52, 74, 56, 53, 63, 53, 44, 57, 49, 48, 57, 49, 52, 79, 53, 37, 37, 
					37, 74, 72, 52, 37, 54, 57, 49, 57, 57, 48, 63, 52, 44, 53, 55, 74, 53, 52, 62, 53, 44, 57, 48, 49, 57, 49, 51, 74, 54, 37, 37, 
					43, 87, 88, 60, 47, 63, 66, 58, 68, 68, 57, 74, 63, 50, 61, 62, 87, 59, 59, 74, 62, 51, 64, 57, 57, 66, 58, 77, 86, 63, 44, 43, 
					37, 74, 73, 53, 37, 50, 57, 48, 57, 57, 46, 63, 52, 44, 54, 54, 75, 52, 52, 63, 53, 40, 57, 49, 49, 58, 42, 52, 74, 54, 37, 37, 
					41, 78, 80, 54, 39, 58, 71, 51, 61, 61, 52, 68, 54, 46, 56, 56, 79, 56, 53, 64, 56, 46, 60, 52, 51, 77, 52, 54, 79, 57, 39, 41, 
					56, 113, 113, 79, 56, 78, 86, 74, 86, 87, 74, 97, 81, 68, 78, 80, 113, 80, 80, 97, 80, 67, 88, 79, 74, 86, 74, 79, 114, 82, 57, 56, 
					41, 82, 82, 57, 41, 58, 63, 53, 63, 63, 50, 69, 58, 48, 58, 58, 82, 58, 58, 73, 58, 45, 63, 53, 54, 63, 54, 57, 82, 59, 41, 41, 
					28, 57, 57, 40, 29, 42, 44, 37, 44, 44, 36, 51, 41, 33, 40, 39, 55, 41, 40, 49, 40, 34, 44, 37, 37, 44, 37, 39, 57, 41, 29, 27, 
					29, 56, 57, 39, 28, 41, 44, 37, 44, 44, 37, 49, 40, 33, 40, 40, 57, 40, 40, 49, 40, 33, 43, 37, 37, 43, 37, 41, 56, 41, 27, 30, 
				}
};

UInt g_aiDeQuantCoef1024_Loeffler_Lifting[6][1024] =
{
	{
		80, 40, 40, 58, 80, 56, 52, 61, 52, 52, 61, 47, 57, 68, 57, 56, 40, 57, 56, 47, 56, 68, 53, 61, 61, 53, 61, 56, 40, 55, 82, 78, 
			40, 20, 20, 29, 40, 28, 26, 31, 26, 26, 32, 22, 28, 34, 29, 27, 19, 29, 28, 24, 28, 34, 26, 31, 31, 26, 31, 29, 20, 28, 40, 41, 
			40, 20, 20, 29, 40, 29, 26, 32, 27, 26, 31, 23, 28, 34, 28, 28, 20, 28, 28, 23, 28, 34, 26, 32, 32, 26, 31, 29, 20, 28, 40, 40, 
			58, 29, 29, 42, 59, 41, 38, 44, 38, 38, 44, 34, 40, 48, 42, 40, 29, 41, 41, 34, 41, 49, 37, 41, 44, 38, 43, 42, 29, 40, 58, 58, 
			80, 40, 40, 59, 83, 55, 52, 61, 52, 52, 61, 48, 57, 68, 56, 57, 40, 57, 58, 48, 58, 68, 52, 61, 61, 48, 62, 59, 41, 56, 80, 80, 
			56, 28, 29, 41, 55, 39, 36, 42, 36, 31, 42, 33, 39, 47, 39, 37, 27, 40, 39, 33, 39, 47, 30, 43, 42, 36, 46, 39, 29, 39, 55, 56, 
			52, 26, 26, 38, 52, 36, 34, 39, 34, 34, 40, 31, 36, 44, 37, 37, 26, 38, 37, 31, 37, 45, 36, 40, 40, 34, 40, 32, 26, 36, 52, 52, 
			61, 31, 32, 44, 61, 42, 39, 46, 40, 40, 47, 35, 44, 52, 43, 42, 31, 43, 43, 37, 42, 52, 40, 47, 47, 39, 47, 45, 31, 43, 62, 61, 
			52, 26, 27, 38, 52, 36, 34, 40, 34, 34, 40, 31, 37, 45, 37, 37, 26, 35, 37, 31, 37, 44, 34, 40, 40, 34, 40, 38, 26, 36, 52, 52, 
			52, 26, 26, 38, 52, 31, 34, 40, 34, 34, 40, 31, 36, 44, 37, 38, 26, 37, 37, 31, 37, 44, 37, 40, 40, 33, 40, 38, 26, 36, 52, 52, 
			61, 32, 31, 44, 61, 42, 40, 47, 40, 40, 46, 35, 43, 52, 44, 43, 31, 45, 43, 36, 42, 52, 40, 47, 47, 40, 50, 44, 31, 45, 63, 61, 
			47, 22, 23, 34, 48, 33, 31, 35, 31, 31, 35, 28, 33, 40, 33, 33, 24, 33, 34, 28, 33, 40, 31, 36, 36, 31, 36, 34, 24, 33, 45, 47, 
			57, 28, 28, 40, 57, 39, 36, 44, 37, 36, 43, 33, 39, 48, 40, 40, 28, 39, 40, 33, 40, 48, 36, 43, 43, 36, 43, 42, 28, 39, 55, 57, 
			68, 34, 34, 48, 68, 47, 44, 52, 45, 44, 52, 40, 48, 52, 48, 48, 34, 48, 45, 40, 48, 58, 44, 53, 52, 45, 52, 50, 33, 47, 69, 68, 
			57, 29, 28, 42, 56, 39, 37, 43, 37, 37, 44, 33, 40, 48, 41, 40, 28, 40, 40, 35, 40, 44, 37, 43, 43, 37, 42, 40, 29, 39, 56, 57, 
			56, 27, 28, 40, 57, 37, 37, 42, 37, 38, 43, 33, 40, 48, 40, 40, 28, 40, 40, 33, 40, 48, 38, 43, 41, 37, 42, 41, 28, 39, 58, 56, 
			40, 19, 20, 29, 40, 27, 26, 31, 26, 26, 31, 24, 28, 34, 28, 28, 20, 28, 28, 24, 28, 34, 26, 31, 31, 26, 30, 29, 20, 28, 41, 40, 
			57, 29, 28, 41, 57, 40, 38, 43, 35, 37, 45, 33, 39, 48, 40, 40, 28, 40, 40, 34, 40, 45, 37, 41, 43, 38, 43, 41, 28, 39, 56, 57, 
			56, 28, 28, 41, 58, 39, 37, 43, 37, 37, 43, 34, 40, 45, 40, 40, 28, 40, 40, 33, 41, 48, 37, 43, 43, 38, 43, 43, 28, 39, 57, 56, 
			47, 24, 23, 34, 48, 33, 31, 37, 31, 31, 36, 28, 33, 40, 35, 33, 24, 34, 33, 27, 33, 40, 30, 36, 37, 31, 36, 35, 23, 31, 47, 47, 
			56, 28, 28, 41, 58, 39, 37, 42, 37, 37, 42, 33, 40, 48, 40, 40, 28, 40, 41, 33, 40, 48, 37, 43, 43, 37, 43, 41, 28, 39, 57, 56, 
			68, 34, 34, 49, 68, 47, 45, 52, 44, 44, 52, 40, 48, 58, 44, 48, 34, 45, 48, 40, 48, 58, 45, 52, 52, 45, 56, 49, 34, 50, 68, 68, 
			53, 26, 26, 37, 52, 30, 36, 40, 34, 37, 40, 31, 36, 44, 37, 38, 26, 37, 37, 30, 37, 45, 39, 40, 40, 35, 40, 38, 26, 36, 52, 53, 
			61, 31, 32, 41, 61, 43, 40, 47, 40, 40, 47, 36, 43, 53, 43, 43, 31, 41, 43, 36, 43, 52, 40, 47, 47, 40, 47, 44, 29, 43, 61, 61, 
			61, 31, 32, 44, 61, 42, 40, 47, 40, 40, 47, 36, 43, 52, 43, 41, 31, 43, 43, 37, 43, 52, 40, 47, 47, 40, 47, 45, 31, 42, 61, 61, 
			53, 26, 26, 38, 48, 36, 34, 39, 34, 33, 40, 31, 36, 45, 37, 37, 26, 38, 38, 31, 37, 45, 35, 40, 40, 34, 39, 30, 26, 36, 52, 53, 
			61, 31, 31, 43, 62, 46, 40, 47, 40, 40, 50, 36, 43, 52, 42, 42, 30, 43, 43, 36, 43, 56, 40, 47, 47, 39, 54, 43, 31, 43, 61, 61, 
			56, 29, 29, 42, 59, 39, 32, 45, 38, 38, 44, 34, 42, 50, 40, 41, 29, 41, 43, 35, 41, 49, 38, 44, 45, 30, 43, 42, 29, 40, 58, 56, 
			40, 20, 20, 29, 41, 29, 26, 31, 26, 26, 31, 24, 28, 33, 29, 28, 20, 28, 28, 23, 28, 34, 26, 29, 31, 26, 31, 29, 20, 28, 40, 40, 
			55, 28, 28, 40, 56, 39, 36, 43, 36, 36, 45, 33, 39, 47, 39, 39, 28, 39, 39, 31, 39, 50, 36, 43, 42, 36, 43, 40, 28, 38, 56, 55, 
			82, 40, 40, 58, 80, 55, 52, 62, 52, 52, 63, 45, 55, 69, 56, 58, 41, 56, 57, 47, 57, 68, 52, 61, 61, 52, 61, 58, 40, 56, 78, 83, 
			78, 41, 40, 58, 80, 56, 52, 61, 52, 52, 61, 47, 57, 68, 57, 56, 40, 57, 56, 47, 56, 68, 53, 61, 61, 53, 61, 56, 40, 55, 83, 76, 
	},
	{
		88, 44, 44, 63, 88, 61, 58, 67, 58, 58, 67, 52, 62, 75, 62, 62, 44, 62, 62, 51, 62, 75, 58, 67, 67, 58, 67, 62, 44, 61, 90, 86, 
			44, 22, 22, 32, 44, 30, 29, 34, 29, 29, 35, 25, 31, 38, 32, 30, 21, 31, 31, 26, 31, 37, 29, 34, 34, 29, 34, 32, 22, 31, 44, 45, 
			44, 22, 22, 32, 44, 31, 28, 35, 30, 28, 34, 26, 31, 37, 31, 31, 22, 31, 31, 25, 31, 38, 28, 35, 35, 28, 34, 31, 22, 30, 44, 44, 
			63, 32, 32, 46, 65, 45, 42, 48, 41, 41, 49, 37, 45, 53, 46, 45, 32, 45, 45, 37, 45, 54, 41, 45, 49, 42, 48, 46, 32, 44, 63, 64, 
			88, 44, 44, 65, 92, 61, 57, 67, 58, 57, 67, 53, 63, 75, 62, 62, 44, 62, 64, 53, 64, 75, 57, 67, 67, 53, 68, 65, 45, 61, 88, 88, 
			61, 30, 31, 45, 61, 43, 40, 47, 40, 34, 46, 36, 43, 52, 42, 41, 30, 44, 43, 36, 43, 52, 33, 47, 47, 40, 50, 43, 32, 43, 60, 61, 
			58, 29, 28, 42, 57, 40, 38, 43, 37, 38, 44, 34, 40, 48, 41, 41, 29, 42, 41, 34, 41, 49, 40, 44, 44, 38, 44, 35, 29, 40, 58, 57, 
			67, 34, 35, 48, 67, 47, 43, 51, 44, 44, 52, 38, 48, 57, 47, 46, 34, 48, 48, 41, 46, 57, 44, 52, 51, 43, 52, 49, 34, 47, 68, 67, 
			58, 29, 30, 41, 58, 40, 37, 44, 38, 37, 44, 34, 41, 49, 40, 41, 29, 39, 41, 34, 41, 48, 37, 44, 44, 37, 44, 41, 29, 40, 57, 57, 
			58, 29, 28, 41, 57, 34, 38, 44, 37, 38, 44, 34, 40, 49, 41, 42, 29, 41, 41, 34, 41, 49, 40, 44, 44, 37, 44, 41, 29, 40, 57, 57, 
			67, 35, 34, 49, 67, 46, 44, 52, 44, 44, 51, 39, 48, 57, 48, 48, 34, 49, 48, 40, 47, 57, 44, 52, 52, 44, 55, 48, 34, 50, 69, 67, 
			52, 25, 26, 37, 53, 36, 34, 38, 34, 34, 39, 30, 36, 44, 36, 37, 26, 37, 38, 30, 37, 44, 34, 40, 39, 34, 40, 37, 26, 36, 49, 52, 
			62, 31, 31, 45, 63, 43, 40, 48, 41, 40, 48, 36, 43, 53, 44, 44, 31, 43, 44, 36, 44, 53, 39, 48, 48, 40, 48, 46, 31, 43, 61, 62, 
			75, 38, 37, 53, 75, 52, 48, 57, 49, 49, 57, 44, 53, 57, 53, 53, 37, 53, 50, 44, 53, 64, 49, 58, 57, 50, 57, 55, 37, 52, 76, 75, 
			62, 32, 31, 46, 62, 42, 41, 47, 40, 41, 48, 36, 44, 53, 45, 44, 31, 44, 44, 38, 44, 49, 41, 47, 47, 41, 47, 44, 32, 43, 62, 62, 
			62, 30, 31, 45, 62, 41, 41, 46, 41, 42, 48, 37, 44, 53, 44, 44, 31, 44, 44, 37, 44, 53, 42, 48, 45, 41, 46, 45, 31, 43, 63, 62, 
			44, 21, 22, 32, 44, 30, 29, 34, 29, 29, 34, 26, 31, 37, 31, 31, 22, 31, 31, 26, 31, 37, 29, 34, 34, 29, 33, 32, 22, 31, 45, 44, 
			62, 31, 31, 45, 62, 44, 42, 48, 39, 41, 49, 37, 43, 53, 44, 44, 31, 44, 44, 37, 44, 50, 41, 45, 48, 42, 48, 45, 31, 43, 62, 62, 
			62, 31, 31, 45, 64, 43, 41, 48, 41, 41, 48, 38, 44, 50, 44, 44, 31, 44, 44, 37, 45, 53, 41, 48, 48, 42, 48, 47, 31, 43, 62, 62, 
			51, 26, 25, 37, 53, 36, 34, 41, 34, 34, 40, 30, 36, 44, 38, 37, 26, 37, 37, 30, 37, 44, 33, 40, 41, 34, 40, 39, 26, 34, 52, 51, 
			62, 31, 31, 45, 64, 43, 41, 46, 41, 41, 47, 37, 44, 53, 44, 44, 31, 44, 45, 37, 44, 53, 41, 48, 48, 41, 47, 45, 31, 43, 62, 62, 
			75, 37, 38, 54, 75, 52, 49, 57, 48, 49, 57, 44, 53, 64, 49, 53, 37, 50, 53, 44, 53, 64, 49, 57, 57, 49, 62, 54, 37, 55, 74, 75, 
			58, 29, 28, 41, 57, 33, 40, 44, 37, 40, 44, 34, 39, 49, 41, 42, 29, 41, 41, 33, 41, 49, 43, 44, 44, 39, 44, 41, 29, 40, 58, 58, 
			67, 34, 35, 45, 67, 47, 44, 52, 44, 44, 52, 40, 48, 58, 47, 48, 34, 45, 48, 40, 48, 57, 44, 51, 52, 44, 52, 48, 32, 47, 67, 67, 
			67, 34, 35, 49, 67, 47, 44, 51, 44, 44, 52, 39, 48, 57, 47, 45, 34, 48, 48, 41, 48, 57, 44, 52, 52, 44, 52, 49, 34, 47, 67, 67, 
			58, 29, 28, 42, 53, 40, 38, 43, 37, 37, 44, 34, 40, 50, 41, 41, 29, 42, 42, 34, 41, 49, 39, 44, 44, 38, 43, 33, 29, 40, 57, 58, 
			67, 34, 34, 48, 68, 50, 44, 52, 44, 44, 55, 40, 48, 57, 47, 46, 33, 48, 48, 40, 47, 62, 44, 52, 52, 43, 59, 48, 34, 47, 67, 67, 
			62, 32, 31, 46, 65, 43, 35, 49, 41, 41, 48, 37, 46, 55, 44, 45, 32, 45, 47, 39, 45, 54, 41, 48, 49, 33, 48, 46, 32, 44, 64, 62, 
			44, 22, 22, 32, 45, 32, 29, 34, 29, 29, 34, 26, 31, 37, 32, 31, 22, 31, 31, 26, 31, 37, 29, 32, 34, 29, 34, 32, 22, 31, 44, 45, 
			61, 31, 30, 44, 61, 43, 40, 47, 40, 40, 50, 36, 43, 52, 43, 43, 31, 43, 43, 34, 43, 55, 40, 47, 47, 40, 47, 44, 31, 42, 61, 61, 
			90, 44, 44, 63, 88, 60, 58, 68, 57, 57, 69, 49, 61, 76, 62, 63, 45, 62, 62, 52, 62, 74, 58, 67, 67, 57, 67, 64, 44, 61, 86, 91, 
			86, 45, 44, 64, 88, 61, 57, 67, 57, 57, 67, 52, 62, 75, 62, 62, 44, 62, 62, 51, 62, 75, 58, 67, 67, 58, 67, 62, 45, 61, 91, 83, 
		},
		{
			104, 52, 52, 75, 104, 72, 68, 80, 68, 68, 80, 61, 74, 89, 74, 73, 52, 74, 73, 61, 73, 89, 69, 79, 80, 69, 79, 73, 52, 72, 107, 101, 
				52, 26, 26, 37, 52, 36, 34, 40, 34, 34, 41, 29, 36, 45, 37, 35, 25, 37, 37, 31, 37, 44, 34, 40, 40, 34, 40, 38, 26, 36, 52, 53, 
				52, 26, 26, 38, 52, 37, 33, 42, 35, 34, 40, 30, 37, 44, 37, 37, 26, 37, 37, 29, 36, 45, 33, 41, 41, 33, 40, 37, 26, 36, 52, 52, 
				75, 37, 38, 54, 77, 54, 49, 57, 49, 49, 58, 44, 53, 63, 55, 53, 38, 53, 53, 44, 53, 63, 48, 53, 57, 49, 56, 54, 37, 52, 75, 75, 
				104, 52, 52, 77, 108, 72, 68, 80, 68, 68, 79, 62, 75, 89, 73, 74, 52, 74, 76, 63, 75, 89, 68, 80, 80, 63, 80, 77, 53, 72, 104, 104, 
				72, 36, 37, 54, 72, 50, 47, 55, 47, 40, 54, 43, 51, 61, 50, 49, 35, 51, 51, 43, 51, 61, 40, 56, 55, 47, 59, 51, 38, 51, 71, 72, 
				68, 34, 33, 49, 68, 47, 44, 51, 44, 44, 52, 40, 47, 57, 48, 48, 34, 49, 48, 40, 48, 58, 47, 52, 52, 44, 52, 41, 34, 47, 68, 68, 
				80, 40, 42, 57, 80, 55, 51, 60, 52, 52, 61, 45, 57, 68, 56, 54, 40, 56, 56, 48, 55, 68, 52, 61, 61, 51, 61, 58, 40, 56, 80, 79, 
				68, 34, 35, 49, 68, 47, 44, 52, 44, 44, 52, 40, 48, 58, 48, 48, 34, 46, 48, 40, 48, 57, 44, 52, 52, 44, 52, 49, 34, 47, 67, 68, 
				68, 34, 34, 49, 68, 40, 44, 52, 44, 45, 52, 40, 47, 58, 48, 49, 34, 48, 48, 40, 48, 58, 47, 52, 52, 43, 52, 49, 34, 47, 68, 68, 
				80, 41, 40, 58, 79, 54, 52, 61, 52, 52, 60, 46, 57, 68, 57, 56, 40, 58, 56, 47, 55, 67, 52, 62, 61, 52, 65, 57, 40, 59, 82, 79, 
				61, 29, 30, 44, 62, 43, 40, 45, 40, 40, 46, 36, 43, 52, 43, 44, 31, 43, 44, 36, 43, 52, 40, 47, 47, 40, 47, 44, 31, 43, 58, 61, 
				74, 36, 37, 53, 75, 51, 47, 57, 48, 47, 57, 43, 51, 63, 52, 52, 37, 51, 52, 42, 52, 63, 46, 56, 56, 47, 56, 55, 36, 51, 72, 74, 
				89, 45, 44, 63, 89, 61, 57, 68, 58, 58, 68, 52, 63, 68, 63, 62, 44, 62, 59, 52, 62, 75, 58, 68, 68, 59, 67, 65, 44, 61, 90, 88, 
				74, 37, 37, 55, 73, 50, 48, 56, 48, 48, 57, 43, 52, 63, 53, 52, 37, 52, 52, 45, 52, 58, 48, 55, 56, 48, 55, 53, 38, 51, 73, 74, 
				73, 35, 37, 53, 74, 49, 48, 54, 48, 49, 56, 44, 52, 62, 52, 52, 37, 52, 52, 43, 52, 62, 50, 56, 54, 48, 55, 53, 37, 51, 75, 73, 
				52, 25, 26, 38, 52, 35, 34, 40, 34, 34, 40, 31, 37, 44, 37, 37, 26, 37, 37, 31, 37, 44, 34, 40, 40, 34, 40, 38, 26, 36, 53, 52, 
				74, 37, 37, 53, 74, 51, 49, 56, 46, 48, 58, 43, 51, 62, 52, 52, 37, 52, 52, 44, 52, 59, 48, 53, 56, 50, 57, 53, 37, 51, 73, 74, 
				73, 37, 37, 53, 76, 51, 48, 56, 48, 48, 56, 44, 52, 59, 52, 52, 37, 52, 52, 43, 53, 63, 48, 56, 56, 50, 56, 55, 37, 51, 74, 73, 
				61, 31, 29, 44, 63, 43, 40, 48, 40, 40, 47, 36, 42, 52, 45, 43, 31, 44, 43, 35, 43, 52, 39, 47, 48, 40, 47, 46, 31, 40, 61, 61, 
				73, 37, 36, 53, 75, 51, 48, 55, 48, 48, 55, 43, 52, 62, 52, 52, 37, 52, 53, 43, 52, 62, 48, 56, 56, 48, 56, 53, 37, 51, 74, 73, 
				89, 44, 45, 63, 89, 61, 58, 68, 57, 58, 67, 52, 63, 75, 58, 62, 44, 59, 63, 52, 62, 75, 58, 68, 68, 58, 73, 64, 44, 65, 88, 89, 
				69, 34, 33, 48, 68, 40, 47, 52, 44, 47, 52, 40, 46, 58, 48, 50, 34, 48, 48, 39, 48, 58, 51, 52, 52, 46, 52, 49, 34, 47, 68, 69, 
				79, 40, 41, 53, 80, 56, 52, 61, 52, 52, 62, 47, 56, 68, 55, 56, 40, 53, 56, 47, 56, 68, 52, 61, 61, 52, 61, 57, 37, 55, 80, 79, 
				80, 40, 41, 57, 80, 55, 52, 61, 52, 52, 61, 47, 56, 68, 56, 54, 40, 56, 56, 48, 56, 68, 52, 61, 61, 52, 61, 58, 40, 55, 80, 79, 
				69, 34, 33, 49, 63, 47, 44, 51, 44, 43, 52, 40, 47, 59, 48, 48, 34, 50, 50, 40, 48, 58, 46, 52, 52, 45, 51, 38, 34, 47, 68, 69, 
				79, 40, 40, 56, 80, 59, 52, 61, 52, 52, 65, 47, 56, 67, 55, 55, 40, 57, 56, 47, 56, 73, 52, 61, 61, 51, 70, 57, 40, 55, 79, 79, 
				73, 38, 37, 54, 77, 51, 41, 58, 49, 49, 57, 44, 55, 65, 53, 53, 38, 53, 55, 46, 53, 64, 49, 57, 58, 38, 57, 54, 37, 52, 75, 73, 
				52, 26, 26, 37, 53, 38, 34, 40, 34, 34, 40, 31, 36, 44, 38, 37, 26, 37, 37, 31, 37, 44, 34, 37, 40, 34, 40, 37, 26, 36, 52, 53, 
				72, 36, 36, 52, 72, 51, 47, 56, 47, 47, 59, 43, 51, 61, 51, 51, 36, 51, 51, 40, 51, 65, 47, 55, 55, 47, 55, 52, 36, 50, 72, 72, 
				107, 52, 52, 75, 104, 71, 68, 80, 67, 68, 82, 58, 72, 90, 73, 75, 53, 73, 74, 61, 74, 88, 68, 80, 80, 68, 79, 75, 52, 72, 102, 108, 
				101, 53, 52, 75, 104, 72, 68, 79, 68, 68, 79, 61, 74, 88, 74, 73, 52, 74, 73, 61, 73, 89, 69, 79, 79, 69, 79, 73, 53, 72, 108, 98, 
		},
		{
			112, 56, 56, 81, 112, 78, 73, 86, 73, 73, 86, 66, 79, 95, 79, 79, 56, 79, 79, 65, 79, 96, 74, 85, 86, 74, 85, 79, 57, 77, 115, 109, 
				56, 28, 28, 40, 56, 39, 37, 43, 36, 36, 44, 31, 39, 48, 40, 38, 27, 40, 40, 33, 40, 48, 37, 43, 43, 37, 43, 41, 28, 39, 56, 57, 
				56, 28, 28, 41, 56, 40, 36, 45, 38, 36, 44, 33, 40, 48, 40, 40, 28, 40, 39, 32, 39, 48, 36, 45, 45, 36, 43, 40, 28, 39, 56, 56, 
				81, 40, 41, 58, 83, 58, 53, 62, 53, 53, 62, 47, 57, 67, 59, 57, 40, 57, 57, 48, 57, 68, 52, 57, 62, 53, 61, 59, 40, 56, 81, 81, 
				112, 56, 56, 83, 117, 77, 73, 86, 73, 73, 85, 67, 80, 95, 79, 79, 56, 79, 82, 68, 81, 95, 73, 86, 86, 67, 86, 83, 57, 78, 112, 112, 
				78, 39, 40, 58, 77, 54, 51, 59, 51, 43, 58, 46, 55, 66, 54, 52, 38, 55, 55, 46, 55, 66, 43, 60, 59, 50, 64, 55, 41, 55, 77, 78, 
				73, 37, 36, 53, 73, 51, 48, 55, 47, 48, 56, 43, 51, 61, 52, 52, 37, 53, 52, 43, 52, 62, 51, 56, 56, 48, 56, 45, 37, 50, 73, 73, 
				86, 43, 45, 62, 86, 59, 55, 65, 56, 56, 66, 49, 61, 73, 60, 58, 43, 61, 61, 52, 59, 73, 56, 66, 65, 55, 66, 62, 43, 60, 86, 85, 
				73, 36, 38, 53, 73, 51, 47, 56, 48, 48, 56, 43, 52, 63, 51, 52, 37, 49, 52, 43, 52, 61, 47, 56, 56, 47, 56, 53, 37, 51, 73, 73, 
				73, 36, 36, 53, 73, 43, 48, 56, 48, 48, 56, 43, 51, 62, 52, 53, 36, 52, 52, 43, 52, 62, 51, 56, 56, 47, 56, 53, 37, 51, 73, 73, 
				86, 44, 44, 62, 85, 58, 56, 66, 56, 56, 64, 50, 61, 73, 61, 61, 43, 63, 61, 50, 59, 72, 56, 66, 66, 56, 70, 61, 43, 64, 88, 86, 
				66, 31, 33, 47, 67, 46, 43, 49, 43, 43, 50, 39, 46, 56, 46, 47, 33, 47, 48, 39, 47, 56, 43, 50, 50, 43, 50, 47, 33, 46, 63, 66, 
				79, 39, 40, 57, 80, 55, 51, 61, 52, 51, 61, 46, 55, 67, 57, 56, 40, 55, 56, 46, 56, 67, 50, 61, 61, 51, 61, 59, 39, 55, 78, 79, 
				95, 48, 48, 67, 95, 66, 61, 73, 63, 62, 73, 56, 67, 73, 67, 67, 47, 67, 64, 56, 67, 81, 62, 74, 73, 64, 73, 70, 47, 66, 97, 95, 
				79, 40, 40, 59, 79, 54, 52, 60, 51, 52, 61, 46, 57, 67, 57, 56, 40, 56, 56, 48, 56, 62, 52, 60, 60, 52, 59, 57, 41, 55, 79, 79, 
				79, 38, 40, 57, 79, 52, 52, 58, 52, 53, 61, 47, 56, 67, 56, 56, 40, 56, 56, 47, 56, 67, 54, 61, 58, 52, 59, 57, 40, 55, 81, 79, 
				56, 27, 28, 40, 56, 38, 37, 43, 37, 36, 43, 33, 40, 47, 40, 40, 28, 39, 39, 33, 40, 48, 37, 43, 43, 36, 43, 40, 28, 39, 57, 56, 
				79, 40, 40, 57, 79, 55, 53, 61, 49, 52, 63, 47, 55, 67, 56, 56, 39, 56, 56, 47, 56, 64, 52, 57, 61, 54, 61, 57, 40, 55, 79, 79, 
				79, 40, 39, 57, 82, 55, 52, 61, 52, 52, 61, 48, 56, 64, 56, 56, 39, 56, 56, 47, 57, 67, 52, 61, 61, 54, 61, 60, 40, 55, 79, 79, 
				65, 33, 32, 48, 68, 46, 43, 52, 43, 43, 50, 39, 46, 56, 48, 47, 33, 47, 47, 38, 47, 56, 42, 50, 52, 43, 51, 49, 33, 43, 66, 65, 
				79, 40, 39, 57, 81, 55, 52, 59, 52, 52, 59, 47, 56, 67, 56, 56, 40, 56, 57, 47, 56, 67, 52, 61, 60, 52, 60, 57, 40, 55, 79, 79, 
				96, 48, 48, 68, 95, 66, 62, 73, 61, 62, 72, 56, 67, 81, 62, 67, 48, 64, 67, 56, 67, 81, 62, 73, 73, 63, 79, 69, 47, 70, 95, 96, 
				74, 37, 36, 52, 73, 43, 51, 56, 47, 51, 56, 43, 50, 62, 52, 54, 37, 52, 52, 42, 52, 62, 54, 56, 56, 50, 56, 53, 36, 51, 73, 74, 
				85, 43, 45, 57, 86, 60, 56, 66, 56, 56, 66, 50, 61, 74, 60, 61, 43, 57, 61, 50, 61, 73, 56, 66, 66, 56, 66, 61, 40, 60, 86, 85, 
				86, 43, 45, 62, 86, 59, 56, 65, 56, 56, 66, 50, 61, 73, 60, 58, 43, 61, 61, 52, 60, 73, 56, 66, 66, 56, 66, 62, 43, 59, 86, 86, 
				74, 37, 36, 53, 67, 50, 48, 55, 47, 47, 56, 43, 51, 64, 52, 52, 36, 54, 54, 43, 52, 63, 50, 56, 56, 48, 55, 41, 37, 51, 73, 74, 
				85, 43, 43, 61, 86, 64, 56, 66, 56, 56, 70, 50, 61, 73, 59, 59, 43, 61, 61, 51, 60, 79, 56, 66, 66, 55, 75, 61, 43, 60, 86, 85, 
				79, 41, 40, 59, 83, 55, 45, 62, 53, 53, 61, 47, 59, 70, 57, 57, 40, 57, 60, 49, 57, 69, 53, 61, 62, 41, 61, 59, 40, 56, 81, 78, 
				57, 28, 28, 40, 57, 41, 37, 43, 37, 37, 43, 33, 39, 47, 41, 40, 28, 40, 40, 33, 40, 47, 36, 40, 43, 37, 43, 40, 28, 39, 56, 57, 
				77, 39, 39, 56, 78, 55, 50, 60, 51, 51, 64, 46, 55, 66, 55, 55, 39, 55, 55, 43, 55, 70, 51, 60, 59, 51, 60, 56, 39, 54, 78, 78, 
				115, 56, 56, 81, 112, 77, 73, 86, 73, 73, 88, 63, 78, 97, 79, 81, 57, 79, 79, 66, 79, 95, 73, 86, 86, 73, 86, 81, 56, 78, 110, 116, 
				109, 57, 56, 81, 112, 78, 73, 85, 73, 73, 86, 66, 79, 95, 79, 79, 56, 79, 79, 65, 79, 96, 74, 85, 86, 74, 85, 78, 57, 78, 116, 106, 
			},
			{
				128, 64, 64, 92, 128, 89, 84, 98, 84, 84, 98, 75, 91, 109, 91, 90, 64, 91, 90, 75, 90, 109, 84, 97, 98, 85, 97, 90, 65, 88, 131, 124, 
					64, 32, 32, 46, 64, 44, 42, 49, 42, 41, 51, 36, 45, 55, 46, 43, 31, 46, 45, 38, 45, 54, 42, 49, 49, 42, 49, 47, 32, 45, 63, 65, 
					64, 32, 32, 46, 64, 46, 41, 51, 43, 41, 50, 37, 45, 54, 45, 45, 32, 45, 45, 36, 45, 55, 41, 51, 51, 41, 50, 46, 32, 44, 64, 64, 
					92, 46, 46, 66, 95, 66, 61, 70, 60, 60, 71, 54, 65, 77, 67, 65, 46, 65, 65, 54, 65, 78, 59, 66, 71, 61, 69, 67, 46, 64, 92, 92, 
					128, 64, 64, 95, 134, 88, 83, 98, 84, 83, 97, 77, 92, 109, 90, 90, 64, 91, 93, 77, 93, 109, 83, 98, 98, 77, 99, 94, 65, 89, 128, 128, 
					89, 44, 46, 66, 88, 62, 58, 68, 58, 49, 67, 52, 63, 75, 62, 60, 43, 63, 63, 53, 63, 76, 49, 69, 68, 58, 73, 62, 47, 63, 88, 89, 
					84, 42, 41, 61, 83, 58, 55, 63, 54, 55, 64, 49, 58, 70, 59, 59, 42, 61, 59, 49, 59, 71, 58, 64, 64, 55, 64, 51, 42, 58, 84, 83, 
					98, 49, 51, 70, 98, 68, 63, 74, 64, 64, 75, 56, 70, 83, 69, 67, 49, 69, 69, 59, 67, 83, 64, 75, 75, 63, 75, 71, 49, 68, 98, 98, 
					84, 42, 43, 60, 84, 58, 54, 64, 55, 54, 64, 49, 59, 72, 58, 59, 42, 56, 59, 49, 59, 70, 54, 64, 64, 54, 64, 60, 42, 58, 83, 84, 
					84, 41, 41, 60, 83, 49, 55, 64, 54, 55, 64, 49, 58, 71, 59, 61, 42, 59, 59, 49, 59, 71, 58, 64, 64, 53, 64, 60, 42, 58, 83, 84, 
					98, 51, 50, 71, 97, 67, 64, 75, 64, 64, 73, 57, 70, 84, 70, 69, 49, 71, 70, 58, 68, 82, 64, 76, 76, 64, 80, 70, 49, 73, 101, 98, 
					75, 36, 37, 54, 77, 52, 49, 56, 49, 49, 57, 44, 53, 64, 53, 54, 38, 53, 55, 44, 53, 64, 49, 58, 57, 49, 58, 54, 38, 53, 72, 75, 
					91, 45, 45, 65, 92, 63, 58, 70, 59, 58, 70, 53, 63, 77, 65, 64, 45, 63, 64, 52, 64, 77, 57, 69, 69, 58, 69, 67, 45, 63, 89, 91, 
					109, 55, 54, 77, 109, 75, 70, 83, 72, 71, 84, 64, 77, 83, 77, 77, 54, 77, 73, 64, 77, 93, 71, 84, 83, 73, 83, 79, 54, 76, 110, 109, 
					91, 46, 45, 67, 90, 62, 59, 69, 58, 59, 70, 53, 65, 77, 65, 64, 45, 64, 64, 55, 64, 71, 59, 68, 69, 59, 68, 65, 47, 63, 90, 91, 
					90, 43, 45, 65, 90, 60, 59, 67, 59, 61, 69, 54, 64, 77, 64, 64, 45, 64, 64, 53, 64, 77, 61, 69, 66, 59, 68, 65, 45, 63, 92, 90, 
					64, 31, 32, 46, 64, 43, 42, 49, 42, 42, 49, 38, 45, 54, 45, 45, 32, 45, 45, 38, 45, 54, 42, 49, 49, 42, 49, 46, 32, 44, 66, 64, 
					91, 46, 45, 65, 91, 63, 61, 69, 56, 59, 71, 53, 63, 77, 64, 64, 45, 64, 64, 54, 64, 73, 59, 66, 69, 61, 70, 65, 46, 63, 90, 91, 
					90, 45, 45, 65, 93, 63, 59, 69, 59, 59, 70, 55, 64, 73, 64, 64, 45, 64, 63, 53, 65, 77, 59, 69, 69, 61, 69, 68, 45, 63, 91, 90, 
					75, 38, 36, 54, 77, 53, 49, 59, 49, 49, 58, 44, 52, 64, 55, 53, 38, 54, 53, 43, 53, 64, 48, 58, 59, 49, 58, 57, 38, 50, 75, 75, 
					90, 45, 45, 65, 93, 63, 59, 67, 59, 59, 68, 53, 64, 77, 64, 64, 45, 64, 65, 53, 64, 77, 59, 69, 69, 59, 69, 65, 45, 63, 91, 90, 
					109, 54, 55, 78, 109, 76, 71, 83, 70, 71, 82, 64, 77, 93, 71, 77, 54, 73, 77, 64, 77, 93, 71, 83, 83, 72, 90, 79, 54, 80, 108, 109, 
					84, 42, 41, 59, 83, 49, 58, 64, 54, 58, 64, 49, 57, 71, 59, 61, 42, 59, 59, 48, 59, 71, 62, 64, 64, 57, 64, 60, 41, 58, 84, 84, 
					97, 49, 51, 66, 98, 69, 64, 75, 64, 64, 76, 58, 69, 84, 68, 69, 49, 66, 69, 58, 69, 83, 64, 75, 75, 64, 75, 70, 46, 68, 98, 97, 
					98, 49, 51, 71, 98, 68, 64, 75, 64, 64, 76, 57, 69, 83, 69, 66, 49, 69, 69, 59, 69, 83, 64, 75, 75, 64, 75, 71, 49, 68, 98, 98, 
					85, 42, 41, 61, 77, 58, 55, 63, 54, 53, 64, 49, 58, 73, 59, 59, 42, 61, 61, 49, 59, 72, 57, 64, 64, 55, 63, 47, 42, 58, 83, 85, 
					97, 49, 50, 69, 99, 73, 64, 75, 64, 64, 80, 58, 69, 83, 68, 68, 49, 70, 69, 58, 69, 90, 64, 75, 75, 63, 86, 70, 49, 68, 98, 97, 
					90, 47, 46, 67, 94, 62, 51, 71, 60, 60, 70, 54, 67, 79, 65, 65, 46, 65, 68, 57, 65, 79, 60, 70, 71, 47, 70, 67, 46, 64, 93, 90, 
					65, 32, 32, 46, 65, 47, 42, 49, 42, 42, 49, 38, 45, 54, 47, 45, 32, 46, 45, 38, 45, 54, 41, 46, 49, 42, 49, 46, 32, 44, 64, 65, 
					88, 45, 44, 64, 89, 63, 58, 68, 58, 58, 73, 53, 63, 76, 63, 63, 44, 63, 63, 50, 63, 80, 58, 68, 68, 58, 68, 64, 44, 61, 89, 89, 
					131, 63, 64, 92, 128, 88, 84, 98, 83, 83, 101, 72, 89, 110, 90, 92, 66, 90, 91, 75, 91, 108, 84, 98, 98, 83, 98, 93, 64, 89, 126, 133, 
					124, 65, 64, 92, 128, 89, 83, 98, 84, 84, 98, 75, 91, 109, 91, 90, 64, 91, 90, 75, 90, 109, 84, 97, 98, 85, 97, 90, 65, 89, 133, 121, 
			},
			{
				144, 72, 72, 104, 144, 100, 94, 110, 94, 94, 110, 85, 102, 123, 102, 102, 72, 102, 102, 84, 102, 123, 95, 110, 110, 96, 109, 101, 73, 99, 148, 140, 
					72, 36, 36, 52, 72, 50, 47, 56, 47, 47, 57, 40, 50, 62, 52, 49, 35, 51, 51, 42, 51, 61, 47, 55, 55, 47, 55, 52, 36, 50, 71, 73, 
					72, 36, 36, 52, 72, 51, 46, 58, 49, 46, 56, 42, 51, 61, 51, 51, 36, 51, 51, 41, 50, 62, 46, 57, 57, 46, 56, 52, 36, 50, 72, 72, 
					104, 52, 52, 75, 107, 74, 68, 79, 68, 68, 80, 61, 73, 87, 76, 73, 52, 74, 73, 61, 73, 88, 67, 74, 80, 68, 78, 75, 52, 72, 104, 104, 
					144, 72, 72, 107, 150, 99, 94, 110, 94, 94, 109, 86, 103, 123, 101, 102, 72, 102, 105, 87, 104, 123, 94, 110, 110, 87, 111, 106, 73, 100, 144, 144, 
					100, 50, 51, 74, 99, 70, 65, 76, 66, 55, 75, 59, 71, 84, 69, 67, 48, 71, 71, 59, 70, 85, 55, 77, 76, 65, 82, 70, 52, 70, 99, 100, 
					94, 47, 46, 68, 94, 65, 62, 70, 61, 61, 72, 55, 65, 79, 66, 67, 47, 68, 66, 56, 66, 80, 66, 72, 72, 62, 71, 57, 47, 65, 94, 94, 
					110, 56, 58, 79, 110, 76, 70, 84, 72, 72, 85, 63, 78, 94, 77, 75, 56, 78, 78, 67, 76, 94, 72, 85, 84, 71, 85, 80, 55, 77, 111, 110, 
					94, 47, 49, 68, 94, 66, 61, 72, 62, 61, 72, 55, 67, 81, 66, 67, 47, 63, 66, 55, 66, 79, 61, 72, 72, 61, 72, 68, 48, 65, 93, 94, 
					94, 47, 46, 68, 94, 55, 61, 72, 61, 62, 72, 56, 66, 80, 66, 68, 47, 66, 67, 55, 67, 80, 66, 72, 72, 60, 72, 68, 47, 65, 93, 94, 
					110, 57, 56, 80, 109, 75, 72, 85, 72, 72, 83, 64, 78, 94, 79, 78, 55, 80, 78, 65, 76, 93, 72, 85, 85, 72, 90, 79, 55, 82, 113, 110, 
					85, 40, 42, 61, 86, 59, 55, 63, 55, 56, 64, 50, 59, 72, 60, 60, 42, 60, 61, 50, 60, 72, 55, 65, 65, 55, 65, 61, 42, 60, 81, 84, 
					102, 50, 51, 73, 103, 71, 65, 78, 67, 66, 78, 59, 71, 87, 73, 72, 51, 70, 72, 59, 72, 87, 64, 78, 78, 65, 78, 76, 50, 70, 100, 102, 
					123, 62, 61, 87, 123, 84, 79, 94, 81, 80, 94, 72, 87, 94, 87, 86, 61, 86, 82, 72, 86, 104, 80, 95, 94, 82, 93, 89, 60, 85, 124, 122, 
					102, 52, 51, 76, 101, 69, 66, 77, 66, 66, 79, 60, 73, 87, 73, 72, 51, 72, 72, 62, 72, 80, 66, 77, 78, 67, 76, 73, 53, 71, 102, 102, 
					102, 49, 51, 73, 102, 67, 67, 75, 67, 68, 78, 60, 72, 86, 72, 73, 51, 72, 72, 60, 72, 86, 69, 78, 74, 66, 76, 73, 51, 71, 104, 101, 
					72, 35, 36, 52, 72, 48, 47, 56, 47, 47, 55, 42, 51, 61, 51, 51, 36, 51, 51, 42, 51, 61, 47, 55, 55, 47, 55, 52, 36, 50, 74, 72, 
					102, 51, 51, 74, 102, 71, 68, 78, 63, 66, 80, 60, 70, 86, 72, 72, 51, 72, 72, 60, 72, 82, 66, 74, 78, 69, 78, 73, 51, 71, 101, 102, 
					102, 51, 51, 73, 105, 71, 66, 78, 66, 67, 78, 61, 72, 82, 72, 72, 51, 72, 71, 60, 73, 87, 67, 78, 78, 69, 78, 77, 51, 70, 102, 101, 
					84, 42, 41, 61, 87, 59, 56, 67, 55, 55, 65, 50, 59, 72, 62, 60, 42, 60, 60, 48, 60, 72, 54, 65, 66, 55, 65, 64, 42, 56, 84, 84, 
					102, 51, 50, 73, 104, 70, 66, 76, 66, 67, 76, 60, 72, 86, 72, 72, 51, 72, 73, 60, 72, 86, 66, 78, 78, 67, 78, 73, 51, 71, 102, 101, 
					123, 61, 62, 88, 123, 85, 80, 94, 79, 80, 93, 72, 87, 104, 80, 86, 61, 82, 87, 72, 86, 104, 80, 93, 94, 81, 102, 88, 61, 91, 122, 123, 
					95, 47, 46, 67, 94, 55, 66, 72, 61, 66, 72, 55, 64, 80, 66, 69, 47, 66, 67, 54, 66, 80, 70, 72, 72, 64, 72, 68, 47, 65, 94, 95, 
					110, 55, 57, 74, 110, 77, 72, 85, 72, 72, 85, 65, 78, 95, 77, 78, 55, 74, 78, 65, 78, 93, 72, 84, 85, 72, 84, 79, 52, 77, 110, 109, 
					110, 55, 57, 80, 110, 76, 72, 84, 72, 72, 85, 65, 78, 94, 78, 74, 55, 78, 78, 66, 78, 94, 72, 85, 84, 72, 84, 80, 56, 76, 110, 110, 
					96, 47, 46, 68, 87, 65, 62, 71, 61, 60, 72, 55, 65, 82, 67, 66, 47, 69, 69, 55, 67, 81, 64, 72, 72, 62, 71, 53, 48, 65, 94, 95, 
					109, 55, 56, 78, 111, 82, 71, 85, 72, 72, 90, 65, 78, 93, 76, 76, 55, 78, 78, 65, 78, 102, 72, 84, 84, 71, 97, 78, 55, 77, 110, 110, 
					101, 52, 52, 75, 106, 70, 57, 80, 68, 68, 79, 61, 76, 89, 73, 73, 52, 73, 77, 64, 73, 88, 68, 79, 80, 53, 78, 75, 52, 72, 105, 101, 
					73, 36, 36, 52, 73, 52, 47, 55, 48, 47, 55, 42, 50, 60, 53, 51, 36, 51, 51, 42, 51, 61, 47, 52, 56, 48, 55, 52, 36, 50, 72, 73, 
					99, 50, 50, 72, 100, 70, 65, 77, 65, 65, 82, 60, 70, 85, 71, 71, 50, 71, 70, 56, 71, 91, 65, 77, 76, 65, 77, 72, 50, 69, 100, 100, 
					148, 71, 72, 104, 144, 99, 94, 111, 93, 93, 113, 81, 100, 124, 102, 104, 74, 101, 102, 84, 102, 122, 94, 110, 110, 94, 110, 105, 72, 100, 141, 149, 
					140, 73, 72, 104, 144, 100, 94, 110, 94, 94, 110, 84, 102, 122, 102, 101, 72, 102, 101, 84, 101, 123, 95, 109, 110, 95, 110, 101, 73, 100, 149, 136, 
				}
};

UInt g_aiQuantCoef1024_Chen[6][1024] =
{
  {
    102, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 102, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    103, 103, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    102, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 102, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    103, 103, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 104, 104, 104, 103, 104, 104, 103, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 103, 104, 104, 103, 104, 104, 104,
    103, 103, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 104, 103, 103, 103, 103, 104, 103, 103, 103, 104, 104, 103, 104, 104, 104, 103, 104, 104, 104, 103, 104, 104, 103, 103, 103, 104, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103,
    103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 104, 103, 103, 103, 103, 103, 103, 103, 103, 103, 103
  },
  {
    93, 94, 94, 94, 93, 94, 94, 93, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 93, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 93, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 93, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 93, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 93, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    93, 94, 94, 94, 93, 94, 94, 93, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 93, 94, 94, 94, 94, 93, 94, 94, 93, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94,
    94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94, 94
  },
  {
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 79, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 79, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 79, 79, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 79, 79, 79, 79, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 79, 79, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 79, 79, 79, 79, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 79, 79, 79, 80, 79, 80, 79, 79, 79, 80, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 79, 80, 80,
    79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79,
    79, 80, 80, 79, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 79, 79, 79, 79, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 79, 79, 79, 80, 80, 79, 79, 80, 80,
    79, 80, 80, 80, 79, 80, 80, 79, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 80, 80, 80, 79, 79, 80, 80, 79, 80, 80, 80
  },
  {
    73, 74, 74, 73, 73, 74, 74, 73, 73, 73, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 73, 73, 73, 74, 73, 73, 73, 73, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 73, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 73, 74, 74, 73, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 73, 73, 73, 74, 74, 73, 73, 73, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 73, 73, 73, 74, 73, 73, 73, 73, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 73, 74, 74, 73, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 74, 74, 74, 73, 73, 74, 74, 73, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 73, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74,
    74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74, 74
  },
  {
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 64, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 64, 64, 64, 64, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 64, 64, 64, 64, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65
  },
  {
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 58, 58, 58, 57, 58, 58, 57, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 57, 58, 58, 57, 58, 58, 58,
    57, 57, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 58, 57, 57, 57, 57, 58, 57, 57, 57, 58, 58, 57, 58, 58, 58, 57, 58, 58, 58, 57, 58, 58, 57, 57, 57, 58, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57,
    57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 58, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57
  }
};
UInt g_aiDeQuantCoef1024_Chen[6][1024] =
{
  {
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 41, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
    40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40
  },
  {
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 44, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 44, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 44, 44, 44, 44, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 44, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 44, 44, 44, 44, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 45, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 44, 44, 44, 44, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 45, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 45, 44, 44, 45, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 44, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 44, 44, 44, 44, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 45, 45, 44, 44, 45, 45, 44, 44, 44, 45, 45, 44, 45, 45, 45, 44, 45, 45, 45, 44, 45, 45, 44, 44, 44, 45, 44, 44, 44, 44, 45,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 45, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44,
    44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 45, 44, 44, 44, 44, 44, 44, 44, 44, 44, 44
  },
  {
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 53, 52, 52, 53, 53,
    52, 53, 52, 52, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 53,
    52, 53, 52, 52, 52, 52, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 52, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 53, 52, 52, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 53, 53, 53, 53, 53, 53, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 52, 52, 52, 52, 52, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 53, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 53, 53, 53, 52, 53, 53, 52, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 52, 53, 53, 52, 53, 53, 53,
    52, 53, 52, 52, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 53,
    52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52, 52,
    52, 52, 52, 52, 52, 52, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 52,
    52, 53, 52, 52, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 52, 52, 52, 52, 53,
    52, 53, 53, 52, 52, 53, 53, 52, 52, 52, 53, 53, 52, 53, 53, 53, 52, 53, 53, 53, 52, 53, 53, 52, 52, 52, 53, 53, 52, 52, 53, 53
  },
  {
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 57, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 56, 56, 56, 56, 56, 57, 56, 56, 56, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 57, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 56, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 56, 56, 56, 57, 57, 56, 56, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 57, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 57, 56, 56, 56, 57, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 57, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 56, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 56, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 57, 56, 56, 56, 57, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 57, 56, 56, 56, 56, 56,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 57, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 57, 57, 57, 57, 57, 57, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56, 56,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 56, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57,
    56, 57, 57, 57, 56, 57, 57, 56, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 57, 57, 57, 56, 56, 57, 57, 56, 57, 57, 57
  },
  {
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 64, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 64, 64, 64, 64, 65, 64, 64, 64, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 64, 64, 64, 65, 64, 64, 64, 64, 65,
    64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 64, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65,
    64, 65, 65, 65, 64, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 65, 64, 65, 65, 65
  },
  {
    72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 73, 72, 73, 72, 73, 72, 73, 72, 73, 72, 73, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 72, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 73, 73, 73, 72, 72, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72, 73, 72, 73, 72, 73, 72, 73, 72, 73, 72, 73, 72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73,
    72, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 73, 72, 73, 73, 73
  }
};

UInt g_aiQuantCoef256_Loeffler_Lifting[6][256] =
{
  {
  102,	204,	102,	157,	156,	134,	145,	145,	205,	144,	145,	157,	134,	134,	205,	103,	
  204,	407,	204,	312,	312,	267,	289,	290,	408,	288,	289,	312,	268,	267,	409,	205,	
  102,	204,	102,	156,	156,	134,	145,	145,	205,	144,	145,	157,	134,	134,	205,	103,	
  157,	312,	156,	239,	239,	205,	222,	222,	313,	221,	221,	239,	205,	205,	313,	157,	
  156,	312,	156,	239,	238,	204,	221,	222,	312,	220,	221,	239,	205,	204,	313,	157,	
  134,	267,	134,	205,	204,	175,	189,	190,	268,	189,	189,	205,	175,	175,	268,	134,	
  145,	289,	145,	222,	221,	189,	205,	206,	290,	204,	205,	222,	190,	190,	290,	145,	
  145,	290,	145,	222,	222,	190,	206,	206,	291,	205,	206,	222,	190,	190,	291,	146,	
  205,	408,	205,	313,	312,	268,	290,	291,	409,	288,	290,	313,	268,	268,	410,	205,	
  144,	288,	144,	221,	220,	189,	204,	205,	288,	203,	204,	221,	189,	189,	289,	145,	
  145,	289,	145,	221,	221,	189,	205,	206,	290,	204,	205,	222,	190,	189,	290,	145,	
  157,	312,	157,	239,	239,	205,	222,	222,	313,	221,	222,	240,	205,	205,	313,	157,	
  134,	268,	134,	205,	205,	175,	190,	190,	268,	189,	190,	205,	176,	175,	268,	134,	
  134,	267,	134,	205,	204,	175,	190,	190,	268,	189,	189,	205,	175,	175,	268,	134,	
  205,	409,	205,	313,	313,	268,	290,	291,	410,	289,	290,	313,	268,	268,	410,	205,	
  103,	205,	103,	157,	157,	134,	145,	146,	205,	145,	145,	157,	134,	134,	205,	103,	
  },
  {
   93,	186,	 93,	142,	142,	122,	132,	132,	186,	131,	132,	142,	122,	122,	186,	 93,	
  186,	370,	186,	284,	283,	243,	263,	264,	371,	262,	263,	284,	243,	243,	372,	186,	
   93,	186,	 93,	142,	142,	122,	132,	132,	186,	131,	132,	142,	122,	122,	186,	 93,	
  142,	284,	142,	218,	217,	186,	201,	202,	285,	200,	201,	218,	186,	186,	285,	143,	
  142,	283,	142,	217,	217,	186,	201,	202,	284,	200,	201,	217,	186,	186,	284,	142,	
  122,	243,	122,	186,	186,	159,	172,	173,	243,	171,	172,	186,	159,	159,	243,	122,	
  132,	263,	132,	201,	201,	172,	186,	187,	263,	186,	186,	201,	173,	172,	264,	132,	
  132,	264,	132,	202,	202,	173,	187,	188,	264,	186,	187,	202,	173,	173,	264,	132,	
  186,	371,	186,	285,	284,	243,	263,	264,	372,	262,	263,	285,	244,	243,	372,	187,	
  131,	262,	131,	200,	200,	171,	186,	186,	262,	185,	185,	201,	172,	172,	262,	131,	
  132,	263,	132,	201,	201,	172,	186,	187,	263,	185,	186,	201,	173,	172,	263,	132,	
  142,	284,	142,	218,	217,	186,	201,	202,	285,	201,	201,	218,	187,	186,	285,	143,	
  122,	243,	122,	186,	186,	159,	173,	173,	244,	172,	173,	187,	160,	160,	244,	122,	
  122,	243,	122,	186,	186,	159,	172,	173,	243,	172,	172,	186,	160,	159,	244,	122,	
  186,	372,	186,	285,	284,	243,	264,	264,	372,	262,	263,	285,	244,	244,	373,	187,	
   93,	186,	 93,	143,	142,	122,	132,	132,	187,	131,	132,	143,	122,	122,	187,	 93,	
  },
  {
   79,	157,	 79,	120,	120,	103,	111,	112,	157,	111,	111,	120,	103,	103,	158,	 79,	
  157,	313,	157,	240,	240,	205,	222,	223,	314,	221,	222,	240,	206,	206,	314,	157,	
   79,	157,	 79,	120,	120,	103,	111,	112,	157,	111,	111,	120,	103,	103,	158,	 79,	
  120,	240,	120,	184,	184,	157,	170,	171,	241,	170,	170,	184,	158,	158,	241,	121,	
  120,	240,	120,	184,	183,	157,	170,	171,	240,	169,	170,	184,	157,	157,	241,	120,	
  103,	205,	103,	157,	157,	135,	146,	146,	206,	145,	146,	157,	135,	135,	206,	103,	
  111,	222,	111,	170,	170,	146,	158,	158,	223,	157,	158,	170,	146,	146,	223,	112,	
  112,	223,	112,	171,	171,	146,	158,	159,	224,	157,	158,	171,	146,	146,	224,	112,	
  157,	314,	157,	241,	240,	206,	223,	224,	315,	222,	223,	241,	206,	206,	315,	158,	
  111,	221,	111,	170,	169,	145,	157,	157,	222,	156,	157,	170,	145,	145,	222,	111,	
  111,	222,	111,	170,	170,	146,	158,	158,	223,	157,	158,	170,	146,	146,	223,	112,	
  120,	240,	120,	184,	184,	157,	170,	171,	241,	170,	170,	184,	158,	158,	241,	121,	
  103,	206,	103,	158,	157,	135,	146,	146,	206,	145,	146,	158,	135,	135,	206,	103,	
  103,	206,	103,	158,	157,	135,	146,	146,	206,	145,	146,	158,	135,	135,	206,	103,	
  158,	314,	158,	241,	241,	206,	223,	224,	315,	222,	223,	241,	206,	206,	315,	158,	
   79,	157,	 79,	121,	120,	103,	112,	112,	158,	111,	112,	121,	103,	103,	158,	 79,	
  },
  {
   73,	146,	 73,	112,	112,	 96,	103,	104,	146,	103,	103,	112,	 96,	 96,	146,	 73,	
  146,	291,	146,	223,	223,	191,	206,	207,	292,	206,	206,	223,	191,	191,	292,	146,	
   73,	146,	 73,	112,	112,	 96,	103,	104,	146,	103,	103,	112,	 96,	 96,	146,	 73,	
  112,	223,	112,	171,	171,	146,	158,	159,	224,	158,	158,	171,	147,	146,	224,	112,	
  112,	223,	112,	171,	170,	146,	158,	158,	223,	157,	158,	171,	146,	146,	223,	112,	
   96,	191,	 96,	146,	146,	125,	135,	136,	191,	135,	135,	146,	125,	125,	191,	 96,	
  103,	206,	103,	158,	158,	135,	146,	147,	207,	146,	146,	158,	136,	135,	207,	104,	
  104,	207,	104,	159,	158,	136,	147,	147,	208,	146,	147,	159,	136,	136,	208,	104,	
  146,	292,	146,	224,	223,	191,	207,	208,	292,	206,	207,	224,	192,	191,	293,	147,	
  103,	206,	103,	158,	157,	135,	146,	146,	206,	145,	146,	158,	135,	135,	206,	103,	
  103,	206,	103,	158,	158,	135,	146,	147,	207,	146,	146,	158,	136,	135,	207,	104,	
  112,	223,	112,	171,	171,	146,	158,	159,	224,	158,	158,	171,	147,	146,	224,	112,	
   96,	191,	 96,	147,	146,	125,	136,	136,	192,	135,	136,	147,	126,	125,	192,	 96,	
   96,	191,	 96,	146,	146,	125,	135,	136,	191,	135,	135,	146,	125,	125,	191,	 96,	
  146,	292,	146,	224,	223,	191,	207,	208,	293,	206,	207,	224,	192,	191,	293,	147,	
   73,	146,	 73,	112,	112,	 96,	104,	104,	147,	103,	104,	112,	 96,	 96,	147,	 73,	
  },
  {
   64,	128,	 64,	 98,	 98,	 84,	 91,	 91,	128,	 90,	 91,	 98,	 84,	 84,	128,	 64,	
  128,	255,	128,	195,	195,	167,	181,	181,	255,	180,	181,	195,	167,	167,	255,	128,	
   64,	128,	 64,	 98,	 98,	 84,	 91,	 91,	128,	 90,	 90,	 98,	 84,	 84,	128,	 64,	
   98,	195,	 98,	150,	149,	128,	138,	139,	196,	138,	138,	150,	128,	128,	196,	 98,	
   98,	195,	 98,	149,	149,	128,	138,	139,	195,	138,	138,	149,	128,	128,	195,	 98,	
   84,	167,	 84,	128,	128,	109,	118,	119,	167,	118,	118,	128,	110,	109,	167,	 84,	
   91,	181,	 91,	138,	138,	118,	128,	129,	181,	128,	128,	139,	119,	118,	181,	 91,	
   91,	181,	 91,	139,	139,	119,	129,	129,	182,	128,	128,	139,	119,	119,	182,	 91,	
  128,	255,	128,	196,	195,	167,	181,	182,	256,	180,	181,	196,	168,	167,	256,	128,	
   90,	180,	 90,	138,	138,	118,	128,	128,	180,	127,	128,	138,	118,	118,	180,	 90,	
   91,	181,	 90,	138,	138,	118,	128,	128,	181,	128,	128,	138,	119,	118,	181,	 91,	
   98,	195,	 98,	150,	149,	128,	139,	139,	196,	138,	138,	150,	128,	128,	196,	 98,	
   84,	167,	 84,	128,	128,	110,	119,	119,	168,	118,	119,	128,	110,	110,	168,	 84,	
   84,	167,	 84,	128,	128,	109,	118,	119,	167,	118,	118,	128,	110,	109,	168,	 84,	
  128,	255,	128,	196,	195,	167,	181,	182,	256,	180,	181,	196,	168,	168,	256,	128,	
   64,	128,	 64,	 98,	 98,	 84,	 91,	 91,	128,	 90,	 91,	 98,	 84,	 84,	128,	 64,	
  },
  {
   57,	113,	 57,	 87,	 87,	 74,	 80,	 81,	114,	 80,	 80,	 87,	 75,	 74,	114,	 57,	
  113,	226,	113,	173,	173,	148,	161,	161,	227,	160,	160,	174,	149,	148,	227,	114,	
   57,	113,	 57,	 87,	 87,	 74,	 80,	 81,	114,	 80,	 80,	 87,	 74,	 74,	114,	 57,	
   87,	173,	 87,	133,	133,	114,	123,	123,	174,	123,	123,	133,	114,	114,	174,	 87,	
   87,	173,	 87,	133,	132,	113,	123,	123,	174,	122,	123,	133,	114,	114,	174,	 87,	
   74,	148,	 74,	114,	113,	 97,	105,	106,	149,	105,	105,	114,	 97,	 97,	149,	 75,	
   80,	161,	 80,	123,	123,	105,	114,	114,	161,	113,	114,	123,	105,	105,	161,	 81,	
   81,	161,	 81,	123,	123,	106,	114,	115,	161,	114,	114,	124,	106,	106,	162,	 81,	
  114,	227,	114,	174,	174,	149,	161,	161,	227,	160,	161,	174,	149,	149,	228,	114,	
   80,	160,	 80,	123,	122,	105,	113,	114,	160,	113,	113,	123,	105,	105,	160,	 80,	
   80,	160,	 80,	123,	123,	105,	114,	114,	161,	113,	114,	123,	105,	105,	161,	 81,	
   87,	174,	 87,	133,	133,	114,	123,	124,	174,	123,	123,	133,	114,	114,	174,	 87,	
   75,	149,	 74,	114,	114,	 97,	105,	106,	149,	105,	105,	114,	 98,	 97,	149,	 75,	
   74,	148,	 74,	114,	114,	 97,	105,	106,	149,	105,	105,	114,	 97,	 97,	149,	 75,	
  114,	227,	114,	174,	174,	149,	161,	162,	228,	160,	161,	174,	149,	149,	228,	114,	
   57,	114,	 57,	 87,	 87,	 75,	 81,	 81,	114,	 80,	 81,	 87,	 75,	 75,	114,	 57,	
  }
};
UInt g_aiDeQuantCoef256_Loeffler_Lifting[6][256] =
{
  {
  160,	 80,	160,	105,	105,	122,	113,	113,	 80,	114,	113,	105,	122,	122,	 80,	160,	
   80,	 40,	 80,	 52,	 53,	 61,	 57,	 57,	 40,	 57,	 57,	 52,	 61,	 61,	 40,	 80,	
  160,	 80,	160,	105,	105,	122,	113,	113,	 80,	114,	113,	105,	122,	122,	 80,	160,	
  105,	 52,	105,	 68,	 69,	 80,	 74,	 74,	 52,	 74,	 74,	 68,	 80,	 80,	 52,	104,	
  105,	 53,	105,	 69,	 69,	 80,	 74,	 74,	 52,	 74,	 74,	 69,	 80,	 80,	 52,	105,	
  122,	 61,	122,	 80,	 80,	 94,	 87,	 86,	 61,	 87,	 87,	 80,	 93,	 94,	 61,	122,	
  113,	 57,	113,	 74,	 74,	 87,	 80,	 80,	 57,	 80,	 80,	 74,	 86,	 86,	 57,	113,	
  113,	 57,	113,	 74,	 74,	 86,	 80,	 79,	 56,	 80,	 80,	 74,	 86,	 86,	 56,	112,	
   80,	 40,	 80,	 52,	 52,	 61,	 57,	 56,	 40,	 57,	 57,	 52,	 61,	 61,	 40,	 80,	
  114,	 57,	114,	 74,	 74,	 87,	 80,	 80,	 57,	 81,	 80,	 74,	 87,	 87,	 57,	113,	
  113,	 57,	113,	 74,	 74,	 87,	 80,	 80,	 57,	 80,	 80,	 74,	 86,	 86,	 57,	113,	
  105,	 52,	105,	 68,	 69,	 80,	 74,	 74,	 52,	 74,	 74,	 68,	 80,	 80,	 52,	104,	
  122,	 61,	122,	 80,	 80,	 93,	 86,	 86,	 61,	 87,	 86,	 80,	 93,	 93,	 61,	122,	
  122,	 61,	122,	 80,	 80,	 94,	 86,	 86,	 61,	 87,	 86,	 80,	 93,	 94,	 61,	122,	
   80,	 40,	 80,	 52,	 52,	 61,	 57,	 56,	 40,	 57,	 57,	 52,	 61,	 61,	 40,	 80,	
  160,	 80,	160,	104,	105,	122,	113,	112,	 80,	113,	113,	104,	122,	122,	 80,	159,	
  },
  {
  176,	 88,	176,	115,	115,	135,	124,	124,	 88,	125,	124,	115,	134,	135,	 88,	176,	
   88,	 44,	 88,	 58,	 58,	 68,	 62,	 62,	 44,	 63,	 62,	 58,	 67,	 67,	 44,	 88,	
  176,	 88,	176,	115,	115,	135,	124,	124,	 88,	125,	125,	115,	134,	135,	 88,	176,	
  115,	 58,	115,	 75,	 75,	 88,	 81,	 81,	 58,	 82,	 81,	 75,	 88,	 88,	 58,	115,	
  115,	 58,	115,	 75,	 76,	 88,	 82,	 81,	 58,	 82,	 82,	 75,	 88,	 88,	 58,	115,	
  135,	 68,	135,	 88,	 88,	103,	 95,	 95,	 67,	 96,	 95,	 88,	103,	103,	 67,	134,	
  124,	 62,	124,	 81,	 82,	 95,	 88,	 88,	 62,	 88,	 88,	 81,	 95,	 95,	 62,	124,	
  124,	 62,	124,	 81,	 81,	 95,	 88,	 87,	 62,	 88,	 88,	 81,	 95,	 95,	 62,	124,	
   88,	 44,	 88,	 58,	 58,	 67,	 62,	 62,	 44,	 62,	 62,	 58,	 67,	 67,	 44,	 88,	
  125,	 63,	125,	 82,	 82,	 96,	 88,	 88,	 62,	 89,	 88,	 82,	 95,	 96,	 62,	125,	
  124,	 62,	125,	 81,	 82,	 95,	 88,	 88,	 62,	 88,	 88,	 81,	 95,	 95,	 62,	124,	
  115,	 58,	115,	 75,	 75,	 88,	 81,	 81,	 58,	 82,	 81,	 75,	 88,	 88,	 58,	115,	
  134,	 67,	134,	 88,	 88,	103,	 95,	 95,	 67,	 95,	 95,	 88,	103,	103,	 67,	134,	
  135,	 67,	135,	 88,	 88,	103,	 95,	 95,	 67,	 96,	 95,	 88,	103,	103,	 67,	134,	
   88,	 44,	 88,	 58,	 58,	 67,	 62,	 62,	 44,	 62,	 62,	 58,	 67,	 67,	 44,	 88,	
  176,	 88,	176,	115,	115,	134,	124,	124,	 88,	125,	124,	115,	134,	134,	 88,	175,	
  },
  {
  208,	104,	208,	136,	136,	159,	147,	147,	104,	148,	147,	136,	159,	159,	104,	208,	
  104,	 52,	104,	 68,	 68,	 80,	 74,	 73,	 52,	 74,	 74,	 68,	 80,	 80,	 52,	104,	
  208,	104,	208,	136,	136,	159,	147,	147,	104,	148,	147,	136,	159,	159,	104,	208,	
  136,	 68,	136,	 89,	 89,	104,	 96,	 96,	 68,	 97,	 96,	 89,	104,	104,	 68,	136,	
  136,	 68,	136,	 89,	 89,	104,	 96,	 96,	 68,	 97,	 96,	 89,	104,	104,	 68,	136,	
  159,	 80,	159,	104,	104,	122,	112,	112,	 80,	113,	113,	104,	121,	122,	 80,	159,	
  147,	 74,	147,	 96,	 96,	112,	104,	104,	 74,	104,	104,	 96,	112,	112,	 73,	147,	
  147,	 73,	147,	 96,	 96,	112,	104,	103,	 73,	104,	104,	 96,	112,	112,	 73,	146,	
  104,	 52,	104,	 68,	 68,	 80,	 74,	 73,	 52,	 74,	 74,	 68,	 79,	 80,	 52,	104,	
  148,	 74,	148,	 97,	 97,	113,	104,	104,	 74,	105,	104,	 97,	113,	113,	 74,	147,	
  147,	 74,	147,	 96,	 96,	113,	104,	104,	 74,	104,	104,	 96,	112,	112,	 73,	147,	
  136,	 68,	136,	 89,	 89,	104,	 96,	 96,	 68,	 97,	 96,	 89,	104,	104,	 68,	136,	
  159,	 80,	159,	104,	104,	121,	112,	112,	 79,	113,	112,	104,	121,	121,	 79,	158,	
  159,	 80,	159,	104,	104,	122,	112,	112,	 80,	113,	112,	104,	121,	122,	 79,	159,	
  104,	 52,	104,	 68,	 68,	 80,	 73,	 73,	 52,	 74,	 73,	 68,	 79,	 79,	 52,	104,	
  208,	104,	208,	136,	136,	159,	147,	146,	104,	147,	147,	136,	158,	159,	104,	207,	
  },
  {
  224,	112,	224,	147,	147,	171,	158,	158,	112,	159,	158,	146,	171,	171,	112,	224,	
  112,	 56,	112,	 73,	 74,	 86,	 79,	 79,	 56,	 80,	 79,	 73,	 86,	 86,	 56,	112,	
  224,	112,	224,	147,	147,	171,	158,	158,	112,	159,	158,	147,	171,	171,	112,	224,	
  147,	 73,	147,	 96,	 96,	112,	104,	103,	 73,	104,	104,	 96,	112,	112,	 73,	146,	
  147,	 74,	147,	 96,	 96,	112,	104,	103,	 73,	104,	104,	 96,	112,	112,	 73,	146,	
  171,	 86,	171,	112,	112,	131,	121,	121,	 86,	122,	121,	112,	131,	131,	 86,	171,	
  158,	 79,	158,	104,	104,	121,	112,	112,	 79,	112,	112,	103,	121,	121,	 79,	158,	
  158,	 79,	158,	103,	103,	121,	112,	111,	 79,	112,	112,	103,	120,	121,	 79,	157,	
  112,	 56,	112,	 73,	 73,	 86,	 79,	 79,	 56,	 80,	 79,	 73,	 86,	 86,	 56,	112,	
  159,	 80,	159,	104,	104,	122,	112,	112,	 80,	113,	112,	104,	121,	122,	 79,	159,	
  158,	 79,	158,	104,	104,	121,	112,	112,	 79,	112,	112,	104,	121,	121,	 79,	158,	
  146,	 73,	147,	 96,	 96,	112,	103,	103,	 73,	104,	104,	 96,	112,	112,	 73,	146,	
  171,	 86,	171,	112,	112,	131,	121,	120,	 86,	121,	121,	112,	131,	131,	 85,	171,	
  171,	 86,	171,	112,	112,	131,	121,	121,	 86,	122,	121,	112,	131,	131,	 86,	171,	
  112,	 56,	112,	 73,	 73,	 86,	 79,	 79,	 56,	 79,	 79,	 73,	 85,	 86,	 56,	112,	
  224,	112,	224,	146,	146,	171,	158,	157,	112,	159,	158,	146,	171,	171,	112,	223,	
  },
  {
  256,	128,	256,	167,	168,	196,	181,	180,	128,	182,	181,	167,	195,	196,	128,	256,	
  128,	 64,	128,	 84,	 84,	 98,	 91,	 90,	 64,	 91,	 91,	 84,	 98,	 98,	 64,	128,	
  256,	128,	256,	168,	168,	196,	181,	180,	128,	182,	181,	167,	196,	196,	128,	256,	
  167,	 84,	168,	110,	110,	128,	118,	118,	 84,	119,	118,	109,	128,	128,	 84,	167,	
  168,	 84,	168,	110,	110,	128,	119,	118,	 84,	119,	119,	110,	128,	128,	 84,	167,	
  196,	 98,	196,	128,	128,	150,	138,	138,	 98,	139,	138,	128,	149,	150,	 98,	195,	
  181,	 91,	181,	118,	119,	138,	128,	127,	 90,	128,	128,	118,	138,	138,	 90,	181,	
  180,	 90,	180,	118,	118,	138,	127,	127,	 90,	128,	128,	118,	138,	138,	 90,	180,	
  128,	 64,	128,	 84,	 84,	 98,	 90,	 90,	 64,	 91,	 91,	 84,	 98,	 98,	 64,	128,	
  182,	 91,	182,	119,	119,	139,	128,	128,	 91,	129,	128,	119,	139,	139,	 91,	181,	
  181,	 91,	181,	118,	119,	138,	128,	128,	 91,	128,	128,	118,	138,	138,	 90,	181,	
  167,	 84,	167,	109,	110,	128,	118,	118,	 84,	119,	118,	109,	128,	128,	 84,	167,	
  195,	 98,	196,	128,	128,	149,	138,	138,	 98,	139,	138,	128,	149,	149,	 98,	195,	
  196,	 98,	196,	128,	128,	150,	138,	138,	 98,	139,	138,	128,	149,	150,	 98,	195,	
  128,	 64,	128,	 84,	 84,	 98,	 90,	 90,	 64,	 91,	 90,	 84,	 98,	 98,	 64,	128,	
  256,	128,	256,	167,	167,	195,	181,	180,	128,	181,	181,	167,	195,	195,	128,	255,	
  },
  {
  288,	144,	288,	188,	189,	220,	204,	203,	144,	204,	204,	188,	220,	220,	144,	287,	
  144,	 72,	144,	 94,	 95,	110,	102,	102,	 72,	102,	102,	 94,	110,	110,	 72,	144,	
  288,	144,	288,	188,	189,	220,	204,	203,	144,	205,	204,	188,	220,	220,	144,	288,	
  188,	 94,	188,	123,	123,	144,	133,	133,	 94,	134,	133,	123,	144,	144,	 94,	188,	
  189,	 95,	189,	123,	124,	144,	133,	133,	 94,	134,	133,	123,	144,	144,	 94,	188,	
  220,	110,	220,	144,	144,	169,	156,	155,	110,	156,	156,	144,	168,	168,	110,	220,	
  204,	102,	204,	133,	133,	156,	144,	143,	102,	144,	144,	133,	155,	156,	102,	203,	
  203,	102,	203,	133,	133,	155,	143,	143,	101,	144,	143,	133,	155,	155,	101,	202,	
  144,	 72,	144,	 94,	 94,	110,	102,	101,	 72,	102,	102,	 94,	110,	110,	 72,	144,	
  204,	102,	205,	134,	134,	156,	144,	144,	102,	145,	145,	134,	156,	156,	102,	204,	
  204,	102,	204,	133,	133,	156,	144,	143,	102,	145,	144,	133,	155,	156,	102,	203,	
  188,	 94,	188,	123,	123,	144,	133,	133,	 94,	134,	133,	123,	144,	144,	 94,	188,	
  220,	110,	220,	144,	144,	168,	155,	155,	110,	156,	155,	144,	168,	168,	110,	219,	
  220,	110,	220,	144,	144,	168,	156,	155,	110,	156,	156,	144,	168,	168,	110,	220,	
  144,	 72,	144,	 94,	 94,	110,	102,	101,	 72,	102,	102,	 94,	110,	110,	 72,	144,	
  287,	144,	288,	188,	188,	220,	203,	202,	144,	204,	203,	188,	219,	220,	144,	287,	
  }
};

UInt g_aiQuantCoef256_Chen[6][256] =
{
  {
    207, 209, 209, 208, 207, 209, 208, 209, 207, 209, 208, 209, 207, 208, 209, 209,
    209, 212, 212, 211, 209, 211, 210, 212, 209, 212, 210, 211, 209, 211, 212, 212,
    209, 212, 212, 211, 209, 211, 210, 212, 209, 212, 210, 211, 209, 211, 212, 212,
    208, 211, 211, 210, 208, 210, 209, 211, 208, 211, 209, 210, 208, 210, 211, 211,
    207, 209, 209, 208, 207, 208, 208, 209, 207, 209, 208, 208, 207, 208, 209, 209,
    209, 211, 211, 210, 208, 210, 210, 211, 209, 211, 210, 210, 208, 210, 211, 211,
    208, 210, 210, 209, 208, 210, 209, 210, 208, 211, 209, 210, 208, 209, 210, 210,
    209, 212, 212, 211, 209, 211, 210, 212, 209, 212, 210, 211, 209, 211, 212, 212,
    207, 209, 209, 208, 207, 209, 208, 209, 207, 210, 208, 209, 207, 208, 209, 209,
    209, 212, 212, 211, 209, 211, 211, 212, 210, 212, 210, 211, 209, 211, 212, 212,
    208, 210, 210, 209, 208, 210, 209, 210, 208, 210, 209, 210, 208, 209, 210, 210,
    209, 211, 211, 210, 208, 210, 210, 211, 209, 211, 210, 210, 208, 210, 211, 211,
    207, 209, 209, 208, 207, 208, 208, 209, 207, 209, 208, 208, 207, 208, 209, 209,
    208, 211, 211, 210, 208, 210, 209, 211, 208, 211, 209, 210, 208, 210, 211, 211,
    209, 212, 212, 211, 209, 211, 210, 212, 209, 212, 210, 211, 209, 211, 212, 212,
    209, 212, 212, 211, 209, 211, 210, 212, 209, 212, 210, 211, 209, 211, 212, 212
  },
  {
    188, 190, 190, 189, 188, 190, 189, 190, 188, 190, 189, 190, 188, 189, 190, 190,
    190, 192, 192, 191, 190, 192, 191, 192, 190, 193, 191, 192, 190, 191, 192, 192,
    190, 192, 192, 191, 190, 192, 191, 192, 190, 192, 191, 192, 190, 191, 192, 192,
    189, 191, 191, 190, 189, 191, 190, 191, 189, 192, 190, 191, 189, 190, 191, 191,
    188, 190, 190, 189, 188, 189, 189, 190, 188, 190, 189, 189, 188, 189, 190, 190,
    190, 192, 192, 191, 189, 191, 191, 192, 190, 192, 191, 191, 189, 191, 192, 192,
    189, 191, 191, 190, 189, 191, 190, 191, 189, 191, 190, 191, 189, 190, 191, 191,
    190, 192, 192, 191, 190, 192, 191, 193, 190, 193, 191, 192, 190, 191, 192, 192,
    188, 190, 190, 189, 188, 190, 189, 190, 188, 190, 189, 190, 188, 189, 190, 190,
    190, 193, 192, 192, 190, 192, 191, 193, 190, 193, 191, 192, 190, 192, 192, 193,
    189, 191, 191, 190, 189, 191, 190, 191, 189, 191, 190, 191, 189, 190, 191, 191,
    190, 192, 192, 191, 189, 191, 191, 192, 190, 192, 191, 191, 189, 191, 192, 192,
    188, 190, 190, 189, 188, 189, 189, 190, 188, 190, 189, 189, 188, 189, 190, 190,
    189, 191, 191, 190, 189, 191, 190, 191, 189, 192, 190, 191, 189, 190, 191, 191,
    190, 192, 192, 191, 190, 192, 191, 192, 190, 192, 191, 192, 190, 191, 192, 192,
    190, 192, 192, 191, 190, 192, 191, 192, 190, 193, 191, 192, 190, 191, 192, 192
  },
  {
    159, 161, 161, 160, 159, 160, 160, 161, 159, 161, 160, 160, 159, 160, 161, 161,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163,
    160, 162, 162, 161, 160, 161, 161, 162, 160, 162, 161, 161, 160, 161, 162, 162,
    159, 161, 161, 160, 159, 160, 160, 161, 159, 161, 160, 160, 159, 160, 161, 161,
    160, 162, 162, 161, 160, 162, 161, 162, 160, 162, 161, 162, 160, 161, 162, 162,
    160, 162, 162, 161, 160, 161, 161, 162, 160, 162, 161, 161, 160, 161, 162, 162,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163,
    159, 161, 161, 160, 159, 160, 160, 161, 159, 161, 160, 160, 159, 160, 161, 161,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163,
    160, 162, 162, 161, 160, 161, 161, 162, 160, 162, 161, 161, 160, 161, 162, 162,
    160, 162, 162, 161, 160, 162, 161, 162, 160, 162, 161, 162, 160, 161, 162, 162,
    159, 161, 161, 160, 159, 160, 160, 161, 159, 161, 160, 160, 159, 160, 161, 161,
    160, 162, 162, 161, 160, 161, 161, 162, 160, 162, 161, 161, 160, 161, 162, 162,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163,
    161, 163, 163, 162, 161, 162, 162, 163, 161, 163, 162, 162, 161, 162, 163, 163
  },
  {
    148, 150, 150, 149, 148, 149, 149, 150, 148, 150, 149, 149, 148, 149, 150, 150,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151,
    149, 150, 150, 150, 149, 150, 150, 150, 149, 150, 150, 150, 149, 150, 150, 150,
    148, 149, 149, 149, 148, 149, 149, 149, 148, 149, 149, 149, 148, 149, 149, 149,
    149, 151, 151, 150, 149, 150, 150, 151, 149, 151, 150, 150, 149, 150, 151, 151,
    149, 150, 150, 150, 149, 150, 149, 150, 149, 150, 149, 150, 149, 150, 150, 150,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151,
    148, 150, 150, 149, 148, 149, 149, 150, 148, 150, 149, 149, 148, 149, 150, 150,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151,
    149, 150, 150, 150, 149, 150, 149, 150, 149, 150, 149, 150, 149, 150, 150, 150,
    149, 151, 151, 150, 149, 150, 150, 151, 149, 151, 150, 150, 149, 150, 151, 151,
    148, 149, 149, 149, 148, 149, 149, 149, 148, 149, 149, 149, 148, 149, 149, 149,
    149, 150, 150, 150, 149, 150, 150, 150, 149, 150, 150, 150, 149, 150, 150, 150,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151,
    150, 151, 151, 150, 149, 151, 150, 151, 150, 151, 150, 151, 149, 150, 151, 151
  },
  {
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    130, 131, 131, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 131, 131,
    131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132,
    130, 131, 131, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132
  },
  {
    115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118,
    116, 117, 117, 116, 116, 117, 116, 117, 116, 117, 116, 117, 116, 116, 117, 117,
    115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116,
    116, 117, 117, 117, 116, 117, 116, 117, 116, 117, 116, 117, 116, 117, 117, 117,
    116, 117, 117, 116, 116, 116, 116, 117, 116, 117, 116, 116, 116, 116, 117, 117,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118,
    115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118,
    116, 117, 117, 116, 116, 116, 116, 117, 116, 117, 116, 116, 116, 116, 117, 117,
    116, 117, 117, 117, 116, 117, 116, 117, 116, 117, 116, 117, 116, 117, 117, 117,
    115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116, 115, 116, 116, 116,
    116, 117, 117, 116, 116, 117, 116, 117, 116, 117, 116, 117, 116, 116, 117, 117,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118,
    116, 118, 118, 117, 116, 117, 117, 118, 116, 118, 117, 117, 116, 117, 118, 118
  }

};
UInt g_aiDeQuantCoef256_Chen[6][256] =
{
  {
    81, 82, 82, 81, 81, 81, 81, 82, 81, 82, 81, 81, 81, 81, 82, 82,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    81, 82, 82, 81, 81, 81, 81, 82, 81, 82, 81, 81, 81, 81, 82, 82,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83,
    81, 82, 82, 81, 81, 81, 81, 82, 81, 82, 81, 81, 81, 81, 82, 82,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    81, 82, 82, 81, 81, 81, 81, 82, 81, 82, 81, 81, 81, 81, 82, 82,
    81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82, 81, 82, 82, 82,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83,
    82, 83, 83, 82, 82, 82, 82, 83, 82, 83, 82, 82, 82, 82, 83, 83
  },
  {
    89, 90, 90, 90, 89, 90, 89, 90, 89, 90, 89, 90, 89, 90, 90, 90,
    90, 91, 91, 90, 90, 91, 90, 91, 90, 91, 90, 91, 90, 90, 91, 91,
    90, 91, 91, 90, 90, 91, 90, 91, 90, 91, 90, 91, 90, 90, 91, 91,
    90, 90, 90, 90, 89, 90, 90, 91, 90, 91, 90, 90, 89, 90, 90, 90,
    89, 90, 90, 89, 89, 90, 89, 90, 89, 90, 89, 90, 89, 89, 90, 90,
    90, 91, 91, 90, 90, 90, 90, 91, 90, 91, 90, 90, 90, 90, 91, 91,
    89, 90, 90, 90, 89, 90, 90, 90, 89, 90, 90, 90, 89, 90, 90, 90,
    90, 91, 91, 91, 90, 91, 90, 91, 90, 91, 90, 91, 90, 91, 91, 91,
    89, 90, 90, 90, 89, 90, 89, 90, 89, 90, 89, 90, 89, 90, 90, 90,
    90, 91, 91, 91, 90, 91, 90, 91, 90, 91, 90, 91, 90, 91, 91, 91,
    89, 90, 90, 90, 89, 90, 90, 90, 89, 90, 90, 90, 89, 90, 90, 90,
    90, 91, 91, 90, 90, 90, 90, 91, 90, 91, 90, 90, 90, 90, 91, 91,
    89, 90, 90, 89, 89, 90, 89, 90, 89, 90, 89, 90, 89, 89, 90, 90,
    90, 90, 90, 90, 89, 90, 90, 91, 90, 91, 90, 90, 89, 90, 90, 90,
    90, 91, 91, 90, 90, 91, 90, 91, 90, 91, 90, 91, 90, 90, 91, 91,
    90, 91, 91, 90, 90, 91, 90, 91, 90, 91, 90, 91, 90, 90, 91, 91
  },
  {
    105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106,
    106, 107, 107, 107, 106, 107, 107, 108, 106, 108, 107, 107, 106, 107, 107, 107,
    106, 107, 107, 107, 106, 107, 107, 107, 106, 108, 107, 107, 106, 107, 107, 107,
    106, 107, 107, 106, 106, 107, 106, 107, 106, 107, 106, 107, 106, 106, 107, 107,
    105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106,
    106, 107, 107, 107, 106, 107, 106, 107, 106, 107, 106, 107, 106, 107, 107, 107,
    106, 107, 107, 106, 106, 106, 106, 107, 106, 107, 106, 106, 106, 106, 107, 107,
    106, 108, 107, 107, 106, 107, 107, 108, 106, 108, 107, 107, 106, 107, 107, 108,
    105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106,
    106, 108, 108, 107, 106, 107, 107, 108, 106, 108, 107, 107, 106, 107, 108, 108,
    106, 107, 107, 106, 106, 106, 106, 107, 106, 107, 106, 106, 106, 106, 107, 107,
    106, 107, 107, 107, 106, 107, 106, 107, 106, 107, 106, 107, 106, 107, 107, 107,
    105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106, 105, 106, 106, 106,
    106, 107, 107, 106, 106, 107, 106, 107, 106, 107, 106, 107, 106, 106, 107, 107,
    106, 107, 107, 107, 106, 107, 107, 107, 106, 108, 107, 107, 106, 107, 107, 107,
    106, 107, 107, 107, 106, 107, 107, 108, 106, 108, 107, 107, 106, 107, 107, 107
  },
  {
    113, 115, 114, 114, 113, 114, 114, 115, 113, 115, 114, 114, 113, 114, 114, 114,
    115, 116, 116, 115, 114, 115, 115, 116, 115, 116, 115, 115, 114, 115, 116, 116,
    114, 116, 116, 115, 114, 115, 115, 116, 114, 116, 115, 115, 114, 115, 116, 116,
    114, 115, 115, 115, 114, 115, 114, 115, 114, 115, 114, 115, 114, 115, 115, 115,
    113, 114, 114, 114, 113, 114, 114, 114, 113, 114, 114, 114, 113, 114, 114, 114,
    114, 115, 115, 115, 114, 115, 115, 115, 114, 115, 115, 115, 114, 115, 115, 115,
    114, 115, 115, 114, 114, 115, 114, 115, 114, 115, 114, 115, 114, 114, 115, 115,
    115, 116, 116, 115, 114, 115, 115, 116, 115, 116, 115, 115, 114, 115, 116, 116,
    113, 115, 114, 114, 113, 114, 114, 115, 113, 115, 114, 114, 113, 114, 114, 115,
    115, 116, 116, 115, 114, 115, 115, 116, 115, 116, 115, 115, 114, 115, 116, 116,
    114, 115, 115, 114, 114, 115, 114, 115, 114, 115, 114, 115, 114, 114, 115, 115,
    114, 115, 115, 115, 114, 115, 115, 115, 114, 115, 115, 115, 114, 115, 115, 115,
    113, 114, 114, 114, 113, 114, 114, 114, 113, 114, 114, 114, 113, 114, 114, 114,
    114, 115, 115, 115, 114, 115, 114, 115, 114, 115, 114, 115, 114, 115, 115, 115,
    114, 116, 116, 115, 114, 115, 115, 116, 114, 116, 115, 115, 114, 115, 116, 116,
    114, 116, 116, 115, 114, 115, 115, 116, 115, 116, 115, 115, 114, 115, 116, 116
  },
  {
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    130, 131, 131, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 131, 131,
    131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132, 131, 132, 132, 132,
    130, 131, 131, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    129, 131, 131, 130, 129, 130, 130, 131, 129, 131, 130, 130, 129, 130, 131, 131,
    130, 132, 132, 131, 130, 131, 131, 132, 130, 132, 131, 131, 130, 131, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132,
    131, 132, 132, 132, 131, 132, 131, 132, 131, 132, 131, 132, 131, 132, 132, 132
  },
  {
    146, 147, 147, 146, 146, 147, 146, 147, 146, 147, 146, 147, 145, 146, 147, 147,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149,
    146, 148, 148, 147, 146, 147, 147, 148, 146, 148, 147, 147, 146, 147, 148, 148,
    146, 147, 147, 146, 145, 147, 146, 147, 146, 147, 146, 147, 145, 146, 147, 147,
    147, 148, 148, 147, 147, 148, 147, 148, 147, 148, 147, 148, 147, 147, 148, 148,
    146, 148, 148, 147, 146, 147, 147, 148, 146, 148, 147, 147, 146, 147, 148, 148,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149,
    146, 147, 147, 146, 146, 147, 146, 147, 146, 147, 146, 147, 146, 146, 147, 147,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149,
    146, 148, 148, 147, 146, 147, 147, 148, 146, 148, 147, 147, 146, 147, 148, 148,
    147, 148, 148, 147, 147, 148, 147, 148, 147, 148, 147, 148, 147, 147, 148, 148,
    145, 147, 147, 146, 145, 147, 146, 147, 146, 147, 146, 147, 145, 146, 147, 147,
    146, 148, 148, 147, 146, 147, 147, 148, 146, 148, 147, 147, 146, 147, 148, 148,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149,
    147, 149, 149, 148, 147, 148, 148, 149, 147, 149, 148, 148, 147, 148, 149, 149
  }

};

const Int g_aiSigmapTree64[64] = {
   0,   2,   1,   5,   4,   3,   9,   8,
   7,   6,  14,  13,  12,  11,  10,  20,
  19,  18,  17,  16,  15,  27,  26,  25,
  24,  23,  22,  21,  49,  33,  32,  31,
  30,  29,  28,  49,  49,  38,  37,  36,
  35,  34,  49,  49,  42,  41,  40,  39,
  49,  49,  45,  44,  43,  49,  49,  47,
  46,  49,  49,  48,  49,  49,  49,  49
};
const Int g_aiSigTStartPoint64[50] = {
  0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
 20, 21, 22, 23, 24, 25, 26, 27, 29, 30,
 31, 32, 33, 34, 37, 38, 39, 40, 41, 44,
 45, 46, 47, 50, 51, 52, 55, 56, 59, 35
};
const Int g_aiSigmapTree256[256] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  16,  10,  11,  12,  17,  17,
  17,  13,  14,  16,  16,  16,  16,  16,  15,  17,  17,  17,  17,  17,  17,  17,
  16,  16,  16,  16,  19,  16,  16,  16,  18,  17,  17,  17,  20,  20,  20,  17,
  17,  18,  18,  16,  16,  19,  19,  19,  19,  19,  16,  18,  18,  18,  17,  20,
  20,  20,  20,  20,  20,  20,  18,  18,  18,  18,  19,  19,  19,  19,  19,  19,
  19,  19,  19,  18,  18,  18,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,
  20,  18,  18,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,
  18,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,
  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  21,
  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  21,  21,  19,
  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  21,  21,  21,  20,  20,  20,
  20,  20,  20,  20,  20,  20,  21,  21,  21,  21,  19,  19,  19,  19,  19,  19,
  19,  21,  21,  21,  21,  21,  20,  20,  20,  20,  20,  21,  21,  21,  21,  21,
  21,  19,  19,  19,  21,  21,  21,  21,  21,  21,  21,  20,  21,  21,  21,  21,
  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,
  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21,  21
};
const Int g_aiSigTStartPoint256[22] = { 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 11, 12, 13, 17, 18, 24, 10, 14, 40, 36, 44, 143 };

const Int g_aiSigmapTree1024[1024] = {
   0,  1,  2,  3,  4,  5,  10,  6,  7,  9,  9,  9,  8,  10,  10,  10,  10,  10,  9,  9,  9,  12,  9,  9,  11,  10,  10,  13,  13,  13,  10,  11,
  11,  9,  12,  12,  12,  12,  12,  11,  11,  11,  13,  13,  13,  13,  13,  13,  13,  11,  11,  12,  12,  12,  12,  12,  12,  12,  12,  12,  11,  13,  13,  13,
  13,  13,  13,  13,  13,  13,  13,  13,  12,  12,  12,  12,  12,  12,  15,  12,  12,  12,  12,  12,  14,  13,  13,  13,  13,  13,  16,  16,  16,  13,  13,  13,
  13,  14,  14,  12,  12,  12,  12,  15,  15,  15,  15,  15,  12,  12,  12,  14,  14,  14,  13,  13,  13,  16,  16,  16,  16,  16,  16,  16,  13,  13,  14,  14,
  14,  14,  12,  12,  15,  15,  15,  15,  15,  15,  15,  15,  15,  12,  14,  14,  14,  14,  14,  13,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  14,
  14,  14,  14,  14,  14,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  14,  14,  14,  14,  14,  16,  16,  16,  16,  16,  16,  16,  16,  16,
  16,  16,  16,  16,  16,  16,  14,  14,  14,  14,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  14,  14,  14,  16,  16,
  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  14,  14,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,
  15,  15,  15,  15,  15,  15,  15,  15,  14,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,
  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  18,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  17,  16,  16,  16,  16,  16,  16,  16,
  16,  16,  16,  16,  18,  18,  18,  16,  16,  16,  16,  16,  16,  16,  16,  16,  16,  17,  17,  15,  15,  15,  15,  15,  15,  15,  15,  15,  15,  18,  18,  18,
  18,  18,  15,  15,  15,  15,  15,  15,  15,  15,  15,  17,  17,  17,  16,  16,  16,  16,  16,  16,  16,  16,  16,  18,  18,  18,  18,  18,  18,  18,  16,  16,
  16,  16,  16,  16,  16,  16,  17,  17,  17,  17,  15,  15,  15,  15,  15,  15,  15,  15,  18,  18,  18,  18,  18,  18,  18,  18,  18,  15,  15,  15,  15,  15,
  15,  15,  17,  17,  17,  17,  17,  16,  16,  16,  16,  16,  16,  16,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  16,  16,  16,  16,  16,  16,  17,
  17,  17,  17,  17,  17,  15,  15,  15,  15,  15,  15,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  15,  15,  15,  15,  15,  17,  17,  17,
  17,  17,  17,  17,  16,  16,  16,  16,  16,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  16,  16,  16,  16,  17,  17,  17,  17,
  17,  17,  17,  17,  15,  15,  15,  15,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  15,  15,  15,  17,  17,  17,  17,  17,
  17,  17,  17,  17,  16,  16,  16,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  16,  16,  17,  17,  17,  17,  17,  17,  17,
  17,  17,  17,  15,  15,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  15,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,
  17,  16,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  17,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  17,  17,  17,  17,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,
  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  17,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  17,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  17,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  17,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,
  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18,  18
};
const Int g_aiSigTStartPoint1024[19] = { 0,  1,  2,  3,  4,  5,  7,  8, 12,  9,  6, 24, 21, 27, 84, 78, 90, 312, 300 };
const Int g_aiSigmapTree4096[4096] = {
   0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  16,  10,  11,  12,  17,  17,  17,  13,  14,  16,  16,  16,  16,  16,  15,  17,  17,  17,  17,  17,  17,  17,  16,  16,  16,  16,  19,  16,  16,  16,  18,  17,  17,  17,  20,  20,  20,  17,  17,  18,  18,  16,  16,  19,  19,  19,  19,  19,  16,  18,  18,  18,  17,  20,
  20,  20,  20,  20,  20,  20,  18,  18,  18,  18,  19,  19,  19,  19,  19,  19,  19,  19,  19,  18,  18,  18,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  18,  18,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  19,  18,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,  20,
  19,  19,  19,  19,  19,  19,  19,  19,  22,  19,  19,  19,  19,  19,  19,  19,  21,  20,  20,  20,  20,  20,  20,  20,  23,  23,  23,  20,  20,  20,  20,  20,  20,  21,  21,  19,  19,  19,  19,  19,  19,  22,  22,  22,  22,  22,  19,  19,  19,  19,  19,  21,  21,  21,  20,  20,  20,  20,  20,  23,  23,  23,  23,  23,
  23,  23,  20,  20,  20,  20,  21,  21,  21,  21,  19,  19,  19,  19,  22,  22,  22,  22,  22,  22,  22,  22,  22,  19,  19,  19,  21,  21,  21,  21,  21,  20,  20,  20,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  20,  20,  21,  21,  21,  21,  21,  21,  19,  19,  22,  22,  22,  22,  22,  22,  22,  22,  22,
  22,  22,  22,  22,  19,  21,  21,  21,  21,  21,  21,  21,  20,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  21,  21,  21,  21,  21,  21,  21,  21,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  21,  21,  21,  21,  21,  21,  21,  23,  23,  23,  23,
  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  21,  21,  21,  21,  21,  21,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  21,  21,  21,  21,  21,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,
  23,  23,  23,  23,  23,  23,  21,  21,  21,  21,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  21,  21,  21,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,
  23,  21,  21,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  21,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,
  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  25,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  24,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  26,  26,  26,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,
  23,  24,  24,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  25,  25,  25,  25,  25,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  24,  24,  24,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  26,  26,  26,  26,  26,  26,  26,  23,  23,  23,  23,  23,  23,
  23,  23,  23,  23,  23,  23,  24,  24,  24,  24,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  25,  25,  25,  25,  25,  25,  25,  25,  25,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  24,  24,  24,  24,  24,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  23,  23,  23,  23,  23,  23,  23,  23,  23,  23,  24,  24,  24,  24,  24,  24,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  22,  22,  22,  22,  22,  22,  22,  22,  22,  24,  24,  24,  24,  24,  24,  24,  23,  23,  23,  23,
  23,  23,  23,  23,  23,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  23,  23,  23,  23,  23,  23,  23,  23,  24,  24,  24,  24,  24,  24,  24,  24,  22,  22,  22,  22,  22,  22,  22,  22,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  22,  22,  22,
  22,  22,  22,  22,  24,  24,  24,  24,  24,  24,  24,  24,  24,  23,  23,  23,  23,  23,  23,  23,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  23,  23,  23,  23,  23,  23,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  22,  22,  22,  22,  22,  22,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  22,  22,  22,  22,  22,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  23,  23,  23,  23,  23,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  23,  23,
  23,  23,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  22,  22,  22,  22,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  22,  22,  22,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  23,  23,  23,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  23,  23,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  22,  22,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  22,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  23,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  24,  24,  24,  24,
  24,  24,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  24,  24,  24,  24,
  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  24,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  24,  24,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  24,  24,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  24,  24,  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  24,  24,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  24,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,
  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,
  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,
  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  25,  25,  25,  25,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  26,  26,  26,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  25,  25,  25,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  26,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,
  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27,  27
};
const Int g_aiSigTStartPoint4096[28] = { 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 11, 12, 13, 17, 18, 24, 10, 14, 40, 36, 44, 144, 136, 152, 544, 528, 560, 2111 };

// ADI
// context pixel filtering flag for each CU size and modes
const UChar g_aucIntraFilter[7][40] =
{
  {0,0,0,0,0,  0,0,0,0,0,    // 2x2
   0,0,0,0,0,  0,0,0,0,0,
   0,0,0,0,0,  0,0,0,0,0,
   0,0,0,0,0,  0,0,0,0,0
  },
  {0,0,0,0,1,  1,1,1,1,1,    // 4x4
   1,1,1,1,1,  1,1,1,1,1,
   0,0,0,0,0,  0,0,0,0,0,
   0,0,0,0,0,  0,0,0,0,0
  },
  {1,1,1,1,2,  2,2,2,2,2,    // 8x8
   2,2,2,2,2,  2,2,2,1,1,
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1
  },
  {0,0,0,0,1,  1,1,1,1,1,  // 16x16
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
  },
  {0,0,0,0,1,  1,1,1,1,1,   // 32x32
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
  },
  {0,0,0,0,1,  1,1,1,1,1,   // 64x64
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
  },
  {0,0,0,0,1,  1,1,1,1,1,   // 128x128
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
   1,1,1,1,1,  1,1,1,1,1,
  }

};

const UChar g_aucIntraModeOrder[7][40] =
{
  {0,1,2,0,0,  0,0,0,0,0,    // 2x2
   0,0,0,0,0,  0,0,0,0,0,
   0,0,0,0,0,  0,0,0,0,0,
   0,0,0,0,0,  0,0,0,0,0
  },
  {0, 1, 2, 4, 5,   6, 7, 8, 9,10,    // 4x4
   11,12,13,14,15,  16,17,18,19,20,
   21,22,23,24,25,  26,27,28,29,30,
   31,32,33,34,35,  36,37,38,39,40
  },
  {0, 1, 2, 4, 5,   6, 7, 8, 9,10,    // 8x8
   11,12,13,14,15,  16,17,18,19,20,
   21,22,23,24,25,  26,27,28,29,30,
   31,32,33,34,35,  36,37,38,39,40
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 9,   // 16x16
   10,11,12,13,14,  15,16,17,18,19,
   20,21,22,23,24,  25,26,27,28,29,
   30,31,32,33,34,  35,36,37,38,39
  },
  {0, 1, 2, 3, 4,   5, 6, 7, 8, 9,   // 32x32
   10,11,12,13,14,  15,16,17,18,19,
   20,21,22,23,24,  25,26,27,28,29,
   30,31,32,33,34,  35,36,37,38,39
  },
  {0, 1, 2, 3, 4,   5, 6, 7, 8, 9,   // 64x64
   10,11,12,13,14,  15,16,17,18,19,
   20,21,22,23,24,  25,26,27,28,29,
   30,31,32,33,34,  35,36,37,38,39
  },
  {0, 1, 2, 3, 4,   5, 6, 7, 8, 9,   // 128x128
   10,11,12,13,14,  15,16,17,18,19,
   20,21,22,23,24,  25,26,27,28,29,
   30,31,32,33,34,  35,36,37,38,39
  },
};

const UChar g_aucIntraModeConv[7][40] = // intra mode conversion for most probable
{
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,   // 2x2
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,   // 4x4
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,   // 8x8
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 2, 3,   4, 5, 6, 7, 8,   // 16x16
   0, 3, 3, 6, 5,   0, 3, 8, 0, 7,
   3, 3, 3, 8, 1,   1, 6, 5, 5, 5,
   3, 3, 2, 8, 8,   8, 2, 2, 2, 2
  },
  {0, 1, 2, 2, 3,   4, 5, 6, 7, 8,   // 32x32
   0, 3, 3, 6, 5,   0, 3, 8, 0, 7,
   3, 3, 3, 8, 1,   1, 6, 5, 5, 5,
   3, 3, 2, 8, 8,   8, 2, 2, 2, 2
  },
  {0, 1, 2, 2, 3,   4, 5, 6, 7, 8,   // 64x64
   0, 3, 3, 6, 5,   0, 3, 8, 0, 7,
   3, 3, 3, 8, 1,   1, 6, 5, 5, 5,
   3, 3, 2, 8, 8,   8, 2, 2, 2, 2
  },
  {0, 1, 2, 2, 3,   4, 5, 6, 7, 8,   // 128x128
   0, 3, 3, 6, 5,   0, 3, 8, 0, 7,
   3, 3, 3, 8, 1,   1, 6, 5, 5, 5,
   3, 3, 2, 8, 8,   8, 2, 2, 2, 2
  },
};

const UChar g_aucIntraModeConvInv[7][9] = // intra mode conversion for most probable
{
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //2x2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //4x4
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //8x8
  },
  {0, 1, 2, 4, 5,  6, 7, 8, 9   //16x16
  },
  {0, 1, 2, 4, 5,  6, 7, 8, 9   //32x32
  },
  {0, 1, 2, 4, 5,  6, 7, 8, 9   //64x64
  },
  {0, 1, 2, 4, 5,  6, 7, 8, 9   //128x128
  }
};

const UChar g_aucIntraModeNum[7] =
{
    3,  //   2x2
    9,  //   4x4
    9,  //   8x8
   33,  //  16x16   33
   33,  //  32x32   33
    5,  //  64x64   33
    5   // 128x128  5
};

const UChar g_aucIntraModeBits[7] =
{
   2,  //   2x2   1+1
   4,  //   4x4   3+1
   4,  //   8x8   3+1
   6,  //  16x16   33   5+1
   6,  //  32x32   33   5+1
   3,  //  64x64   33   4+1
   3   // 128x128  33   4+1
};

const UChar g_aucIntraModeNumFast[7] =
{
   3,  //   2x2
   9,  //   4x4
   9,  //   8x8
   4,  //  16x16   33
   4,  //  32x32   33
   5,  //  64x64   33
   4   // 128x128  33
};

const UChar g_aucIntraAvail[40][2] =
{
  {1,0},
  {0,1},
  {0,0},
  {1,1},
  {1,0},
  {1,1},
  {1,1},
  {1,1},
  {1,0},
  {0,1},

  {1,0},
  {1,0},
  {1,0},
  {1,1},
  {1,1},
  {1,1},
  {0,1},
  {0,1},
  {1,0},
  {1,0},

  {1,0},
  {1,0},
  {1,0},
  {1,0},
  {1,1},
  {1,1},
  {1,1},
  {1,1},
  {1,1},
  {1,1},

  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1},
  {0,1}
};

const UChar g_aucXYflg[40]=
{
   0, 0, 0, 0,
   0, 0, 0, 0, 0,
   1, 0, 0, 0, 0,
   0, 0, 1, 1, 0,
   0, 0, 0, 0, 0,
   0, 0, 0, 0, 0,
   0, 1, 1, 1, 1,
   1, 1,
   1, 1, 1, 1
};

const Char g_aucDirDx[32] =
{
   1,  1,  1,  2,  1,
   2,  2,  5, 10, 11,
   4,  1,  1, 12,  1,
   3,  3,  5,  7,  7,
  11,  6,  8,  5,  5,
   2,  5,  4,  5,  5,
   8, 10
};

const Char g_aucDirDy[32] =
{
   -1,   1,   2,   1,  -2,
   -1, -11,  -7,  -7,   3,
    3,  11,  -1,  -3, -11,
  -10,  -5,  -6,  -6,  -4,
    1,   1,   3,   3,   7,
    7,  -7,  -3,  -3,  -1,
   -3,  -1
};
//--ADI


// ADI_CHROMA--
// chroma
const UChar g_aucIntraModeConvC[7][40] = // intra mode conversion for most probable
{
  {0, 1, 2, 3, 4,  2, 2, 2, 2, 2,   // 2x2        4 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  2, 2, 2, 2, 2,    // 4x4       4 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 8x8       9 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 16x16     9 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 32x32     9 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  /*
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 32x32     33 mode
   0, 4, 4, 7, 6,  0, 4, 2, 0, 8,
   4, 4, 4, 2, 1,  1, 7, 6, 6, 6,
   4, 4, 3, 2, 2,  2, 2, 2, 2, 2
  },
  */
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 64x64     9 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8, 2,    // 128x128   9 mode
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2,
   2, 2, 2, 2, 2,  2, 2, 2, 2, 2
  },
};

const UChar g_aucIntraModeConvInvC[7][9] = // intra mode conversion for most probable
{
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //2x2
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //4x4
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8    //8x8
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8   //16x16
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8   //32x32
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8   //64x64
  },
  {0, 1, 2, 3, 4,  5, 6, 7, 8   //128x128
  }
};

const UChar g_aucIntraModeNumC[7] =
{

    4,  //   2x2
    4,  //   4x4
    9,  //   8x8     9
    9,  //  16x16    9
    9,  //  32x32    9
    9,  //  64x64    9
    9   // 128x128   9


};

const UChar g_aucIntraModeBitsC[7] =
{
   3,  //   2x2   2+1
   3,  //   4x4   2+1
   5,  //   8x8   2+1
   5,  //  16x16      4+1
   5,  //  32x32      4+1
   5,  //  64x64      4+1
   5   // 128x128     4+1
};

// -- ADI_CHROMA



const UChar g_aucConvertTxtTypeToIdx[4] = { 0, 1, 1, 2 };

UInt g_auiConvertRelToAbsIdx[4096] = { 0, };
UInt g_auiConvertAbsToRelIdx[4096] = { 0, };

Void   ConvertRelToAbsIdx ( Int iMaxDepth, Int iDepth, UInt uiStartVal, UInt*& rpuiCurrIdx )
{
  Int iStride = 1 << ( iMaxDepth - 1 );

  if ( iDepth == iMaxDepth )
  {
    rpuiCurrIdx[0] = uiStartVal;
    rpuiCurrIdx++;
  }
  else
  {
    Int iStep = iStride >> iDepth;
    ConvertRelToAbsIdx( iMaxDepth, iDepth+1, uiStartVal,                     rpuiCurrIdx );
    ConvertRelToAbsIdx( iMaxDepth, iDepth+1, uiStartVal+iStep,               rpuiCurrIdx );
    ConvertRelToAbsIdx( iMaxDepth, iDepth+1, uiStartVal+iStep*iStride,       rpuiCurrIdx );
    ConvertRelToAbsIdx( iMaxDepth, iDepth+1, uiStartVal+iStep*iStride+iStep, rpuiCurrIdx );
  }
}

Void   ConvertAbsToRelIdx ( UInt uiMaxCUWidth, UInt uiMaxCUHeight, UInt uiMaxDepth )
{
  UInt  uiMinCUWidth  = uiMaxCUWidth  >> ( uiMaxDepth - 1 );
  UInt  uiMinCUHeight = uiMaxCUHeight >> ( uiMaxDepth - 1 );

  UInt  uiNumPartInWidth  = (UInt)uiMaxCUWidth  / uiMinCUWidth;
  UInt  uiNumPartInHeight = (UInt)uiMaxCUHeight / uiMinCUHeight;

  for ( UInt i = 0; i < uiNumPartInWidth*uiNumPartInHeight; i++ )
  {
    g_auiConvertAbsToRelIdx[ g_auiConvertRelToAbsIdx[i] ] = i;
  }
}

UInt g_auiConvertPartIdxToPelX[4096] = {0, };
UInt g_auiConvertPartIdxToPelY[4096] = {0, };

Void   ConvertPartIdxToPel ( UInt uiMaxCUWidth, UInt uiMaxCUHeight, UInt uiMaxDepth )
{
  UInt    i;

  UInt* uiTempX = &g_auiConvertPartIdxToPelX[0];
  UInt* uiTempY = &g_auiConvertPartIdxToPelY[0];

  UInt  uiMinCUWidth  = uiMaxCUWidth  >> ( uiMaxDepth - 1 );
  UInt  uiMinCUHeight = uiMaxCUHeight >> ( uiMaxDepth - 1 );

  UInt  uiNumPartInWidth  = uiMaxCUWidth  / uiMinCUWidth;
  UInt  uiNumPartInHeight = uiMaxCUHeight / uiMinCUHeight;

  uiTempX[0] = 0; uiTempX++;
  for ( i = 1; i < uiNumPartInWidth; i++ )
  {
    uiTempX[0] = uiTempX[-1] + uiMinCUWidth; uiTempX++;
  }
  for ( i = 1; i < uiNumPartInHeight; i++ )
  {
    memcpy(uiTempX, uiTempX-uiNumPartInWidth, sizeof(UInt)*uiNumPartInWidth);
    uiTempX += uiNumPartInWidth;
  }

  for ( i = 1; i < uiNumPartInWidth*uiNumPartInHeight; i++ )
  {
    uiTempY[i] = ( i / uiNumPartInWidth ) * uiMinCUWidth;
  }
};

const Long ROT_MATRIX[27][96]=
  {
 // ----   67108864-----left and  right ROT ----------
	 { 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,},
	 { 65120396,-14404904, -7445279,        0,-15665940,-63842560,-13502035,        0, -4184694, 14839991,-65313576,        0,        0,        0,        0, 67108864, 65506236,-12826692, -6928823,        0,-14278223,-62887816,-18570264,        0, -2943623, 19600982,-64115024,        0,        0,        0,        0, 67108864, 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864, 61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864, 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864, 61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864,},
	 { 66831944,  5754485, -1994019,        0, -5873764, 66709580, -4350896,        0,  1609069,  4507468, 66937976,        0,        0,        0,        0, 67108864, 63201476,-22056710, -4761730,        0, 18044674, 57905836,-28721138,        0, 13548521, 25768502, 60465048,        0,        0,        0,        0, 67108864, 63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864, 66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864, 63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864, 66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864,},
	 { 62199072,-25090676, -2309256,        0, 23461976, 59915692,-19058998,        0,  9187519, 16857270, 64304136,        0,        0,        0,        0, 67108864, 63729196,-20973680, -1514647,        0, 18469970, 58141160,-27965432,        0, 10052344, 26140198, 60985568,        0,        0,        0,        0, 67108864, 63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864, 66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864, 63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864, 66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864,},
	 { 65431788, -4615802,-14176580,        0, -3372992, 57574192,-34313776,        0, 14522533, 34168800, 55903388,        0,        0,        0,        0, 67108864, 55613788,-37487684,  2319558,        0, 35095856, 50390636,-27067784,        0, 13378622, 23644398, 61364116,        0,        0,        0,        0, 67108864, 45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864, 61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864, 45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864, 61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864,},
	 { 51378192,-43133444, -1840461,        0,-42239572,-49630448,-16007397,        0,  8927452, 13413599,-65145804,        0,        0,        0,        0, 67108864, 65243284, -2292379,-15545379,        0, -6444672, 56646392,-35401296,        0, 14331084, 35910032, 54851516,        0,        0,        0,        0, 67108864, 52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864, 66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864, 52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864, 66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864,},
	 { 65720852,  1288088,-13517027,        0, -6829985, 60875392,-27406890,        0, 11735434, 28215728, 59747404,        0,        0,        0,        0, 67108864, 64298316,-19217858,   -17052,        0, 19166084, 64129424, -4865997,        0,  1409763,  4657335, 66932212,        0,        0,        0,        0, 67108864, 59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864, 64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864, 59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864, 64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864,},
	 { 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864, 61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864, 66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864, 64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864, 66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864, 64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864,},
	 { 66771012,  -383782, -6714517,        0, -6650005,-13776510,-65342064,        0, -1004720, 65678460,-13745184,        0,        0,        0,        0, 67108864, 64284528,-19211034, -1426714,        0, -4829302,-11259854,-65981004,        0, 18648788, 63306804,-12168441,        0,        0,        0,        0, 67108864, 62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864, 46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864, 62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864, 46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864,},
///-----------FOR VERT
	 { 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,},
	 { 65120396,-14404904, -7445279,        0,-15665940,-63842560,-13502035,        0, -4184694, 14839991,-65313576,        0,        0,        0,        0, 67108864, 65506236,-12826692, -6928823,        0,-14278223,-62887816,-18570264,        0, -2943623, 19600982,-64115024,        0,        0,        0,        0, 67108864, 62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864,  46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864,  62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864,  46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864,},
	 { 66831944,  5754485, -1994019,        0, -5873764, 66709580, -4350896,        0,  1609069,  4507468, 66937976,        0,        0,        0,        0, 67108864, 63201476,-22056710, -4761730,        0, 18044674, 57905836,-28721138,        0, 13548521, 25768502, 60465048,        0,        0,        0,        0, 67108864, 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864,  61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864,  65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864,  61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864,},
	 { 62199072,-25090676, -2309256,        0, 23461976, 59915692,-19058998,        0,  9187519, 16857270, 64304136,        0,        0,        0,        0, 67108864, 63729196,-20973680, -1514647,        0, 18469970, 58141160,-27965432,        0, 10052344, 26140198, 60985568,        0,        0,        0,        0, 67108864, 45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864,  61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864,  45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864,  61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864,},
	 { 65431788, -4615802,-14176580,        0, -3372992, 57574192,-34313776,        0, 14522533, 34168800, 55903388,        0,        0,        0,        0, 67108864, 55613788,-37487684,  2319558,        0, 35095856, 50390636,-27067784,        0, 13378622, 23644398, 61364116,        0,        0,        0,        0, 67108864, 52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864,  66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864,  52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864,  66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864,},
	 { 51378192,-43133444, -1840461,        0,-42239572,-49630448,-16007397,        0,  8927452, 13413599,-65145804,        0,        0,        0,        0, 67108864, 65243284, -2292379,-15545379,        0, -6444672, 56646392,-35401296,        0, 14331084, 35910032, 54851516,        0,        0,        0,        0, 67108864, 63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864,  66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864,  63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864,  66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864,},
	 { 65720852,  1288088,-13517027,        0, -6829985, 60875392,-27406890,        0, 11735434, 28215728, 59747404,        0,        0,        0,        0, 67108864, 64298316,-19217858,   -17052,        0, 19166084, 64129424, -4865997,        0,  1409763,  4657335, 66932212,        0,        0,        0,        0, 67108864, 63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864,  66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864,  63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864,  66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864,},
	 { 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864, 61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864, 59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864,  64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864,  59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864,  64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864,},
	 { 66771012,  -383782, -6714517,        0, -6650005,-13776510,-65342064,        0, -1004720, 65678460,-13745184,        0,        0,        0,        0, 67108864, 64284528,-19211034, -1426714,        0, -4829302,-11259854,-65981004,        0, 18648788, 63306804,-12168441,        0,        0,        0,        0, 67108864, 66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864,  64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864,  66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864,  64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864,},
/////////////---FOR HOR
	 { 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,  67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,        0,        0,        0,        0, 67108864,},
	 { 65120396,-14404904, -7445279,        0,-15665940,-63842560,-13502035,        0, -4184694, 14839991,-65313576,        0,        0,        0,        0, 67108864, 65506236,-12826692, -6928823,        0,-14278223,-62887816,-18570264,        0, -2943623, 19600982,-64115024,        0,        0,        0,        0, 67108864, 45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864,  61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864,  45570584,-48885096,  6096666,        0, 47569452, 41505612,-22760294,        0, 12808941, 19777042, 62836292,        0,        0,        0,        0, 67108864,  61851816,-25916550, -2507040,        0,-26036388,-61501260, -6580380,        0,   243706,  7037559,-66738388,        0,        0,        0,        0, 67108864,},
	 { 66831944,  5754485, -1994019,        0, -5873764, 66709580, -4350896,        0,  1609069,  4507468, 66937976,        0,        0,        0,        0, 67108864, 63201476,-22056710, -4761730,        0, 18044674, 57905836,-28721138,        0, 13548521, 25768502, 60465048,        0,        0,        0,        0, 67108864, 52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864,  66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864,  52843068,-40677204, -7521622,        0,-41099764,-50242636,-17031934,        0,  4692453, 18017830,-64474324,        0,        0,        0,        0, 67108864,  66279352, -4471283, -9521305,        0, -2246443, 53326104,-40680212,        0, 10276239, 40496100, 52517276,        0,        0,        0,        0, 67108864,},
	 { 62199072,-25090676, -2309256,        0, 23461976, 59915692,-19058998,        0,  9187519, 16857270, 64304136,        0,        0,        0,        0, 67108864, 63729196,-20973680, -1514647,        0, 18469970, 58141160,-27965432,        0, 10052344, 26140198, 60985568,        0,        0,        0,        0, 67108864, 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864,  61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864,  65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864,  61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864,},
	 { 65431788, -4615802,-14176580,        0, -3372992, 57574192,-34313776,        0, 14522533, 34168800, 55903388,        0,        0,        0,        0, 67108864, 55613788,-37487684,  2319558,        0, 35095856, 50390636,-27067784,        0, 13378622, 23644398, 61364116,        0,        0,        0,        0, 67108864, 63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864,  66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864,  63818416,-16095363,-13105287,        0,-19695116,-60332156,-21811300,        0, -6550691, 24587996,-62097652,        0,        0,        0,        0, 67108864,  66558580, -7075509, -4846891,        0,  5865315, 65222876,-14668817,        0,  6257258, 14124913, 65306456,        0,        0,        0,        0, 67108864,},
	 { 51378192,-43133444, -1840461,        0,-42239572,-49630448,-16007397,        0,  8927452, 13413599,-65145804,        0,        0,        0,        0, 67108864, 65243284, -2292379,-15545379,        0, -6444672, 56646392,-35401296,        0, 14331084, 35910032, 54851516,        0,        0,        0,        0, 67108864, 63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864,  66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864,  63587200,-21209656,  3227769,        0, 21257612, 60925876,-18432336,        0,  2895130, 18487498, 64447112,        0,        0,        0,        0, 67108864,  66191064,-10810889, -2338348,        0,-11003290,-65804080, -7235369,        0, -1127301,  7519817,-66676688,        0,        0,        0,        0, 67108864,},
	 { 65720852,  1288088,-13517027,        0, -6829985, 60875392,-27406890,        0, 11735434, 28215728, 59747404,        0,        0,        0,        0, 67108864, 64298316,-19217858,   -17052,        0, 19166084, 64129424, -4865997,        0,  1409763,  4657335, 66932212,        0,        0,        0,        0, 67108864, 59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864,  64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864,  59493932,-31030864,  1075675,        0, 30659556, 58344184,-12631197,        0,  4905423, 11689351, 65900648,        0,        0,        0,        0, 67108864,  64908572,-16971322,  1565583,        0, 17029116, 64832208, -3223917,        0,  -697167,  3515485, 67013092,        0,        0,        0,        0, 67108864,},
	 { 65921888, -9549591, -8167585,        0,  5928666, 62095368,-24751072,        0, 11079492, 23591738, 61840716,        0,        0,        0,        0, 67108864, 61782724,-26185832,   892897,        0, 26144218, 61462352, -6516063,        0,  1724799,  6346767, 66785800,        0,        0,        0,        0, 67108864, 66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864,  64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864,  66716760, -6717631, -2710581,        0,  6320628, 66252112, -8620160,        0,  3538859,  8314500, 66497712,        0,        0,        0,        0, 67108864,  64057860,-19928158, -1748803,        0,-19992818,-63977156, -3287985,        0,  -690812,  3659500,-67005448,        0,        0,        0,        0, 67108864,},
	 { 66771012,  -383782, -6714517,        0, -6650005,-13776510,-65342064,        0, -1004720, 65678460,-13745184,        0,        0,        0,        0, 67108864, 64284528,-19211034, -1426714,        0, -4829302,-11259854,-65981004,        0, 18648788, 63306804,-12168441,        0,        0,        0,        0, 67108864, 62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864,  46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864,  62126876,-25259528, -2409859,        0,  4918208, 18240076,-64394960,        0, 24893020, 59437832, 18737170,        0,        0,        0,        0, 67108864,  46965964, 47935016,   178067,        0, 47934404,-46963144,  -596155,        0,  -301218,   544407,-67105980,        0,        0,        0,        0, 67108864,},

};

// ROT_SHIFT_ONLY
const Long ROT_MATRIX_S[2][27][96]=
{
  {
    {  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,},
    {  7949, -1758,  -908,     0, -1912, -7793, -1648,     0,  -510,  1811, -7972,     0,     0,     0,     0,  8192,  7996, -1565,  -845,     0, -1742, -7676, -2266,     0,  -359,  2392, -7826,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,},
    {  8158,   702,  -243,     0,  -717,  8143,  -531,     0,   196,   550,  8171,     0,     0,     0,     0,  8192,  7715, -2692,  -581,     0,  2202,  7068, -3505,     0,  1653,  3145,  7380,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,},
    {  7592, -3062,  -281,     0,  2864,  7313, -2326,     0,  1121,  2057,  7849,     0,     0,     0,     0,  8192,  7779, -2560,  -184,     0,  2254,  7097, -3413,     0,  1227,  3190,  7444,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,},
    {  7987,  -563, -1730,     0,  -411,  7028, -4188,     0,  1772,  4170,  6824,     0,     0,     0,     0,  8192,  6788, -4576,   283,     0,  4284,  6151, -3304,     0,  1633,  2886,  7490,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,},
    {  6271, -5265,  -224,     0, -5156, -6058, -1954,     0,  1089,  1637, -7952,     0,     0,     0,     0,  8192,  7964,  -279, -1897,     0,  -786,  6914, -4321,     0,  1749,  4383,  6695,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,},
    {  8022,   157, -1650,     0,  -833,  7431, -3345,     0,  1432,  3444,  7293,     0,     0,     0,     0,  8192,  7848, -2345,    -2,     0,  2339,  7828,  -593,     0,   172,   568,  8170,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,},
    {  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,},
    {  8150,   -46,  -819,     0,  -811, -1681, -7976,     0,  -122,  8017, -1677,     0,     0,     0,     0,  8192,  7847, -2345,  -174,     0,  -589, -1374, -8054,     0,  2276,  7727, -1485,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,},
    {  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,},
    {  7949, -1758,  -908,     0, -1912, -7793, -1648,     0,  -510,  1811, -7972,     0,     0,     0,     0,  8192,  7996, -1565,  -845,     0, -1742, -7676, -2266,     0,  -359,  2392, -7826,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,},
    {  8158,   702,  -243,     0,  -717,  8143,  -531,     0,   196,   550,  8171,     0,     0,     0,     0,  8192,  7715, -2692,  -581,     0,  2202,  7068, -3505,     0,  1653,  3145,  7380,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,},
    {  7592, -3062,  -281,     0,  2864,  7313, -2326,     0,  1121,  2057,  7849,     0,     0,     0,     0,  8192,  7779, -2560,  -184,     0,  2254,  7097, -3413,     0,  1227,  3190,  7444,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,},
    {  7987,  -563, -1730,     0,  -411,  7028, -4188,     0,  1772,  4170,  6824,     0,     0,     0,     0,  8192,  6788, -4576,   283,     0,  4284,  6151, -3304,     0,  1633,  2886,  7490,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,},
    {  6271, -5265,  -224,     0, -5156, -6058, -1954,     0,  1089,  1637, -7952,     0,     0,     0,     0,  8192,  7964,  -279, -1897,     0,  -786,  6914, -4321,     0,  1749,  4383,  6695,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,},
    {  8022,   157, -1650,     0,  -833,  7431, -3345,     0,  1432,  3444,  7293,     0,     0,     0,     0,  8192,  7848, -2345,    -2,     0,  2339,  7828,  -593,     0,   172,   568,  8170,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,},
    {  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,},
    {  8150,   -46,  -819,     0,  -811, -1681, -7976,     0,  -122,  8017, -1677,     0,     0,     0,     0,  8192,  7847, -2345,  -174,     0,  -589, -1374, -8054,     0,  2276,  7727, -1485,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,},
    {  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,     0,     0,     0,     0,  8192,},
    {  7949, -1758,  -908,     0, -1912, -7793, -1648,     0,  -510,  1811, -7972,     0,     0,     0,     0,  8192,  7996, -1565,  -845,     0, -1742, -7676, -2266,     0,  -359,  2392, -7826,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,  5562, -5967,   744,     0,  5806,  5066, -2778,     0,  1563,  2414,  7670,     0,     0,     0,     0,  8192,  7550, -3163,  -306,     0, -3178, -7507,  -803,     0,    29,   859, -8146,     0,     0,     0,     0,  8192,},
    {  8158,   702,  -243,     0,  -717,  8143,  -531,     0,   196,   550,  8171,     0,     0,     0,     0,  8192,  7715, -2692,  -581,     0,  2202,  7068, -3505,     0,  1653,  3145,  7380,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,  6450, -4965,  -918,     0, -5017, -6133, -2079,     0,   572,  2199, -7870,     0,     0,     0,     0,  8192,  8090,  -545, -1162,     0,  -274,  6509, -4965,     0,  1254,  4943,  6410,     0,     0,     0,     0,  8192,},
    {  7592, -3062,  -281,     0,  2864,  7313, -2326,     0,  1121,  2057,  7849,     0,     0,     0,     0,  8192,  7779, -2560,  -184,     0,  2254,  7097, -3413,     0,  1227,  3190,  7444,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,},
    {  7987,  -563, -1730,     0,  -411,  7028, -4188,     0,  1772,  4170,  6824,     0,     0,     0,     0,  8192,  6788, -4576,   283,     0,  4284,  6151, -3304,     0,  1633,  2886,  7490,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,  7790, -1964, -1599,     0, -2404, -7364, -2662,     0,  -799,  3001, -7580,     0,     0,     0,     0,  8192,  8124,  -863,  -591,     0,   715,  7961, -1790,     0,   763,  1724,  7971,     0,     0,     0,     0,  8192,},
    {  6271, -5265,  -224,     0, -5156, -6058, -1954,     0,  1089,  1637, -7952,     0,     0,     0,     0,  8192,  7964,  -279, -1897,     0,  -786,  6914, -4321,     0,  1749,  4383,  6695,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,  7762, -2589,   394,     0,  2594,  7437, -2250,     0,   353,  2256,  7867,     0,     0,     0,     0,  8192,  8079, -1319,  -285,     0, -1343, -8032,  -883,     0,  -137,   917, -8139,     0,     0,     0,     0,  8192,},
    {  8022,   157, -1650,     0,  -833,  7431, -3345,     0,  1432,  3444,  7293,     0,     0,     0,     0,  8192,  7848, -2345,    -2,     0,  2339,  7828,  -593,     0,   172,   568,  8170,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,  7262, -3787,   131,     0,  3742,  7122, -1541,     0,   598,  1426,  8044,     0,     0,     0,     0,  8192,  7923, -2071,   191,     0,  2078,  7914,  -393,     0,   -85,   429,  8180,     0,     0,     0,     0,  8192,},
    {  8047, -1165,  -997,     0,   723,  7580, -3021,     0,  1352,  2879,  7548,     0,     0,     0,     0,  8192,  7541, -3196,   108,     0,  3191,  7502,  -795,     0,   210,   774,  8152,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,  8144,  -820,  -330,     0,   771,  8087, -1052,     0,   431,  1014,  8117,     0,     0,     0,     0,  8192,  7819, -2432,  -213,     0, -2440, -7809,  -401,     0,   -84,   446, -8179,     0,     0,     0,     0,  8192,},
    {  8150,   -46,  -819,     0,  -811, -1681, -7976,     0,  -122,  8017, -1677,     0,     0,     0,     0,  8192,  7847, -2345,  -174,     0,  -589, -1374, -8054,     0,  2276,  7727, -1485,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,  7583, -3083,  -294,     0,   600,  2226, -7860,     0,  3038,  7255,  2287,     0,     0,     0,     0,  8192,  5733,  5851,    21,     0,  5851, -5732,   -72,     0,   -36,    66, -8191,     0,     0,     0,     0,  8192,},
  },
  {
    {     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,},
    {  2188, -3368, -6943,     0, -2836, -2304, -1619,     0, -6774,  4279, -6952,     0,     0,     0,     0,     0,  3004, -6212, -6583,     0, -7759, -6024, -7192,     0, -2695,  5718, -4432,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,},
    {  1608,  3701, -3363,     0,  -100,  2124,  -944,     0,  3437,  1868,  1144,     0,     0,     0,     0,     0,   196, -3846, -2178,     0,  5890,  4780, -8178,     0,  7145,  4662,  8088,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,},
    {  5408, -6772, -7304,     0,    88,  7596, -4406,     0,  4287,  6326,  5128,     0,     0,     0,     0,     0,  3628, -2160, -7319,     0,  5202,  2536, -6136,     0,   760,  7718,  4320,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,},
    {  2284, -3706, -4420,     0, -6080,   816, -5680,     0,  6309,  8160,  1180,     0,     0,     0,     0,     0,  6492, -1092,  1222,     0,  1328,  1644, -1416,     0,  1086,  2286,  6036,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,},
    {  6160, -2564, -5453,     0, -1620, -3312,  -229,     0,  6364,  3295, -3020,     0,     0,     0,     0,     0,  2196, -6811, -5155,     0, -5760,  6904, -3664,     0,  3276,  4496,  6076,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,},
    {  4628,  1944,  -227,     0, -6049,   640, -4650,     0,  4490,  2480,  3148,     0,     0,     0,     0,     0,  7500, -7618,  -668,     0,  4996,  2448, -8141,     0,   739,  4279,  3572,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,},
    {   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,},
    {  6212, -6950, -5269,     0, -6293, -5758, -2672,     0, -5296,  3196, -7200,     0,     0,     0,     0,     0,  1904,  -794, -1306,     0, -4214, -4046, -2636,     0,  3796,  7220, -3321,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,},
    {     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,},
    {  2188, -3368, -6943,     0, -2836, -2304, -1619,     0, -6774,  4279, -6952,     0,     0,     0,     0,     0,  3004, -6212, -6583,     0, -7759, -6024, -7192,     0, -2695,  5718, -4432,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,},
    {  1608,  3701, -3363,     0,  -100,  2124,  -944,     0,  3437,  1868,  1144,     0,     0,     0,     0,     0,   196, -3846, -2178,     0,  5890,  4780, -8178,     0,  7145,  4662,  8088,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,},
    {  5408, -6772, -7304,     0,    88,  7596, -4406,     0,  4287,  6326,  5128,     0,     0,     0,     0,     0,  3628, -2160, -7319,     0,  5202,  2536, -6136,     0,   760,  7718,  4320,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,},
    {  2284, -3706, -4420,     0, -6080,   816, -5680,     0,  6309,  8160,  1180,     0,     0,     0,     0,     0,  6492, -1092,  1222,     0,  1328,  1644, -1416,     0,  1086,  2286,  6036,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,},
    {  6160, -2564, -5453,     0, -1620, -3312,  -229,     0,  6364,  3295, -3020,     0,     0,     0,     0,     0,  2196, -6811, -5155,     0, -5760,  6904, -3664,     0,  3276,  4496,  6076,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,},
    {  4628,  1944,  -227,     0, -6049,   640, -4650,     0,  4490,  2480,  3148,     0,     0,     0,     0,     0,  7500, -7618,  -668,     0,  4996,  2448, -8141,     0,   739,  4279,  3572,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,},
    {   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,},
    {  6212, -6950, -5269,     0, -6293, -5758, -2672,     0, -5296,  3196, -7200,     0,     0,     0,     0,     0,  1904,  -794, -1306,     0, -4214, -4046, -2636,     0,  3796,  7220, -3321,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,},
    {     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,},
    {  2188, -3368, -6943,     0, -2836, -2304, -1619,     0, -6774,  4279, -6952,     0,     0,     0,     0,     0,  3004, -6212, -6583,     0, -7759, -6024, -7192,     0, -2695,  5718, -4432,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,  6680, -3432,  1818,     0,  6700,  4940, -2918,     0,  4845,  1554,  3652,     0,     0,     0,     0,     0,  2216, -5254,  -288,     0, -2212, -3916, -2204,     0,  6138,   631, -6356,     0,     0,     0,     0,     0,},
    {  1608,  3701, -3363,     0,  -100,  2124,  -944,     0,  3437,  1868,  1144,     0,     0,     0,     0,     0,   196, -3846, -2178,     0,  5890,  4780, -8178,     0,  7145,  4662,  8088,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,  4668, -3924, -1366,     0,  -500, -1100,  -766,     0,  6629,  3622, -3284,     0,     0,     0,     0,     0,  6072, -6643, -2201,     0, -1835,  4376, -6932,     0,  3471,  3044,  6556,     0,     0,     0,     0,     0,},
    {  5408, -6772, -7304,     0,    88,  7596, -4406,     0,  4287,  6326,  5128,     0,     0,     0,     0,     0,  3628, -2160, -7319,     0,  5202,  2536, -6136,     0,   760,  7718,  4320,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,},
    {  2284, -3706, -4420,     0, -6080,   816, -5680,     0,  6309,  8160,  1180,     0,     0,     0,     0,     0,  6492, -1092,  1222,     0,  1328,  1644, -1416,     0,  1086,  2286,  6036,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,  2736, -6275, -6279,     0, -1548, -6268, -4196,     0, -5283,  3804, -2292,     0,     0,     0,     0,     0,  6772, -5813, -5419,     0,  8035,  6364, -5137,     0,  6762,  1905,  8024,     0,     0,     0,     0,     0,},
    {  6160, -2564, -5453,     0, -1620, -3312,  -229,     0,  6364,  3295, -3020,     0,     0,     0,     0,     0,  2196, -6811, -5155,     0, -5760,  6904, -3664,     0,  3276,  4496,  6076,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,   896,  -568,   121,     0,  7564,  1972,  -336,     0,  3354,  6346,   648,     0,     0,     0,     0,     0,  7896, -5641, -3628,     0, -1434, -5936, -1833,     0, -4997,  7753, -2000,     0,     0,     0,     0,     0,},
    {  4628,  1944,  -227,     0, -6049,   640, -4650,     0,  4490,  2480,  3148,     0,     0,     0,     0,     0,  7500, -7618,  -668,     0,  4996,  2448, -8141,     0,   739,  4279,  3572,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,  3628, -7760,  2523,     0,  5092,   760, -7325,     0,  6607,  7559,  4200,     0,     0,     0,     0,     0,  3356, -5690,   911,     0,  6140,   720, -4461,     0,  -847,  1117,  2532,     0,     0,     0,     0,     0,},
    {   864, -5911,  -161,     0,  5850,     8, -3040,     0,  3908,  6970,  7500,     0,     0,     0,     0,     0,  6852, -4200,  8161,     0,  3546,  5968, -3423,     0,  4479,  6159,  4616,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,  1112,  -191, -7221,     0,  4596,  3408, -2176,     0,  8107,  7812,  3248,     0,     0,     0,     0,     0,  4612, -5214, -3907,     0, -4338, -5828, -2993,     0, -2684,  5868, -3080,     0,     0,     0,     0,     0,},
    {  6212, -6950, -5269,     0, -6293, -5758, -2672,     0, -5296,  3196, -7200,     0,     0,     0,     0,     0,  1904,  -794, -1306,     0, -4214, -4046, -2636,     0,  3796,  7220, -3321,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,  6940, -3592, -1411,     0,  3008,  4684, -5840,     0,  5724,  4872,  2066,     0,     0,     0,     0,     0,  1228,  3624,  6035,     0,  3012, -6600, -6331,     0, -6306,  3735, -5308,     0,     0,     0,     0,     0,},

    }
};


UInt g_uiMaxCUWidth  = MAX_CU_SIZE;
UInt g_uiMaxCUHeight = MAX_CU_SIZE;
UInt g_uiMaxCUDepth  = MAX_CU_DEPTH;
UInt g_uiAddCUDepth  = 0;

UInt g_uiBitDepth     = 8;   // base bit-depth
UInt g_uiBitIncrement = 0;   // increments
UInt g_uiIBDI_MAX           = 255; //((1<<(BASE_BIT_DEPTH+BIT_DEPTH_INCREASE))-1)
UInt g_uiBASE_MAX = 255;

Bool g_bUseCADR				= false;
Int	 g_iMinCADR				= 0;
Int	 g_iMaxCADR				= 255;
Int	 g_iRangeCADR			= 256;

Char  g_aucConvertToBit  [ MAX_CU_SIZE+1 ];
UChar g_aucConvertToBitXY[ MAX_CU_SIZE+1 ];

// changed for ACS (Vadim)
UInt* g_auiFrameScanXY[ MAX_COEFF_SCAN ][ MAX_CU_DEPTH  ];
UInt* g_auiFrameScanX [ MAX_COEFF_SCAN ][ MAX_CU_DEPTH  ];
UInt* g_auiFrameScanY [ MAX_COEFF_SCAN ][ MAX_CU_DEPTH  ];

UInt g_auiPartIdxMap    [ 4096 ];

// LTR_2D_CONTEXT_MAPPING
UInt g_auiAntiScan8[64];

// initialize g_auiFrameScanXY
Void initFrameScanXY( UInt* pBuff, UInt* pBuffX, UInt* pBuffY, Int iWidth, Int iHeight )
{
  Int x, y, c = 0;

  // starting point
  pBuffX[ c ] = 0;
  pBuffY[ c ] = 0;
  pBuff[ c++ ] = 0;

  // loop
  x=1; y=0;
  while (1)
  {
    // decrease loop
    while ( x>=0 )
    {
      if ( x >= 0 && x < iWidth && y >= 0 && y < iHeight )
      {
        pBuffX[ c ] = x;
        pBuffY[ c ] = y;
        pBuff[ c++ ] = x+y*iWidth;
      }
      x--; y++;
    }
    x=0;

    // increase loop
    while ( y>=0 )
    {
      if ( x >= 0 && x < iWidth && y >= 0 && y < iHeight )
      {
        pBuffX[ c ] = x;
        pBuffY[ c ] = y;
        pBuff[ c++ ] = x+y*iWidth;
      }
      x++; y--;
    }
    y=0;

    // termination condition
    if ( c >= iWidth*iHeight ) break;
  }

  // LTR_2D_CONTEXT_MAPPING
  if (iWidth == 8 && iHeight == 8)
  {
    for( c = 0; c < iWidth*iHeight; c++)
    {
      g_auiAntiScan8[pBuff[c]] = c;
    }
  }

}

const Int cgaiCorrDiapason[5][2]=
{
	{0,0},
	{0,0},
	{0,15},
	{0,31},
	{3,60},
};

// ACS
Void initFrameScanHorOrVer( UInt* puiScanXY, UInt* puiScanX, UInt* puiScanY, UInt uiSlope, UInt uiWidth, UInt uiHeight, Bool bVer )
{
  Int x, y, k;

  if (uiSlope == 0)
  {
    k = 0;
    for( y = 0; y < uiHeight; y++ )
    {
      for( x = 0; x < uiWidth; x++ )
      {
        if( bVer ) // vertical
        {
          puiScanXY[ k ] = x * uiHeight + y;
          puiScanX [ k ] = y;
          puiScanY [ k ] = x;
        }
        else // horizontal
        {
          puiScanXY[ k ] = k;
          puiScanX [ k ] = x;
          puiScanY [ k ] = y;
        }

        k++;
      }
    }

    return;
  }

  Int *iTempOrdinate = new Int[uiHeight*uiWidth];

  k = 0;

  for( y = 32768; y < uiHeight*65536; y+=65536 )
  {
    for( x = 32768; x < uiWidth*65536; x+=65536  )
    {
      Int iTemp;
      if (bVer)
      {
        iTemp = uiSlope*y+(x<<8);
      }
      else
      {
        iTemp = uiSlope*x+(y<<8);
      }

      iTempOrdinate[k++] = iTemp;
    }
  }

  std::vector<std::pair<int, int> > vec;
  for( k = 0; k < uiWidth*uiHeight; k++ )
  {
    vec.push_back( std::pair<int, int>( iTempOrdinate[k], k) );
  }

  std::sort( vec.begin(), vec.end(), compare() );

  Int i = 0;
  for( std::vector<std::pair<int, int> >::const_iterator itr = vec.begin(); itr != vec.end(); itr++ )
  {
    puiScanX[i] = (*itr).second%uiWidth;
    puiScanY[i] = (*itr).second/uiWidth;
    puiScanXY[i++] = (*itr).second;
  }

  delete [] iTempOrdinate;
}



UInt* g_puiSqrBase;
// initialize ROM variables
Void initROM()
{
	Int i, c;

	// g_aucConvertToBit[ x ]: log2(x/4), if x=4 -> 0, x=8 -> 1, x=16 -> 2, ...
	::memset( g_aucConvertToBit,   -1, sizeof( g_aucConvertToBit ) );
  ::memset( g_aucConvertToBitXY,  0, sizeof( g_aucConvertToBit ) );
	c=0;
	for ( i=4; i<MAX_CU_SIZE; i*=2 )
  {
    g_aucConvertToBit[ i ] = c;
    memset(g_aucConvertToBitXY + i, i, sizeof(UChar)*i);
    c++;
  }
  g_aucConvertToBit[ i ] = c;

	// g_auiFrameScanXY[SCAN_ZIGZAG][ g_aucConvertToBit[ transformSize ] ]: zigzag scan array for transformSize
	c=4;
	for ( i=0; i<MAX_CU_DEPTH; i++ )
	{
		g_auiFrameScanXY[SCAN_ZIGZAG][ i ] = new UInt[ c*c ];
    g_auiFrameScanX [SCAN_ZIGZAG][ i ] = new UInt[ c*c ];
    g_auiFrameScanY [SCAN_ZIGZAG][ i ] = new UInt[ c*c ];
		initFrameScanXY( g_auiFrameScanXY[SCAN_ZIGZAG][i], g_auiFrameScanX[SCAN_ZIGZAG][i], g_auiFrameScanY[SCAN_ZIGZAG][i], c, c );

    // ACS
    g_auiFrameScanXY[SCAN_HOR][ i ] = new UInt[ c*c ];
    g_auiFrameScanX [SCAN_HOR][ i ] = new UInt[ c*c ];
    g_auiFrameScanY [SCAN_HOR][ i ] = new UInt[ c*c ];

    g_auiFrameScanXY[SCAN_VER][ i ] = new UInt[ c*c ];
    g_auiFrameScanX [SCAN_VER][ i ] = new UInt[ c*c ];
    g_auiFrameScanY [SCAN_VER][ i ] = new UInt[ c*c ];

    initFrameScanHorOrVer( g_auiFrameScanXY[SCAN_HOR][ i ], g_auiFrameScanX[SCAN_HOR][ i ], g_auiFrameScanY[SCAN_HOR][ i ], 0, c, c, 0 );
    initFrameScanHorOrVer( g_auiFrameScanXY[SCAN_VER][ i ], g_auiFrameScanX[SCAN_VER][ i ], g_auiFrameScanY[SCAN_VER][ i ], 0, c, c, 1 );
    //--

		c <<= 1;
	}

  Int iDiffRange = (g_uiBASE_MAX + 1)*2;

  g_puiSqrBase = new UInt[iDiffRange];

  g_puiSqrBase += iDiffRange/2;
  for (i=0; i < iDiffRange/2; i++)
  {
    g_puiSqrBase[i]=g_puiSqrBase[-i]=i*i;
  }
}

Void destroyROM()
{
	Int i;

  // destroy g_auiFrameScanXY[SCAN_ZIGZAG]
	for ( i=0; i<MAX_CU_DEPTH; i++ )
	{
    delete[] g_auiFrameScanXY[SCAN_ZIGZAG][i];
    delete[] g_auiFrameScanX [SCAN_ZIGZAG][i];
    delete[] g_auiFrameScanY [SCAN_ZIGZAG][i];

    // ACS
    delete[] g_auiFrameScanXY[SCAN_HOR][i];
    delete[] g_auiFrameScanX [SCAN_HOR][i];
    delete[] g_auiFrameScanY [SCAN_HOR][i];
    delete[] g_auiFrameScanXY[SCAN_VER][i];
    delete[] g_auiFrameScanX [SCAN_VER][i];
    delete[] g_auiFrameScanY [SCAN_VER][i];
    //--

	}

  g_puiSqrBase -= g_uiBASE_MAX+1;
  delete[] g_puiSqrBase;
}

Void initCADRQPTable( Bool bUseLCT )
{
	Int i, j;

	// modify scaling values in forward quantizer
	for ( i=0; i<6; i++ )										  		 g_aiQuantCoef4   [i]    = CADR_DESCALE( g_aiQuantCoef4   [i],    g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<16;   j++ ) g_aiQuantCoef    [i][j] = CADR_DESCALE( g_aiQuantCoef    [i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<64;   j++ ) g_aiQuantCoef64  [i][j] = CADR_DESCALE( g_aiQuantCoef64  [i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<256;  j++ ) g_aiQuantCoef256_Chen [i][j] = CADR_DESCALE( g_aiQuantCoef256_Chen [i][j], g_iRangeCADR );
  for ( i=0; i<6; i++ ) for ( j=0; j<1024; j++ ) g_aiQuantCoef1024_Chen[i][j] = CADR_DESCALE( g_aiQuantCoef1024_Chen[i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ )                          g_aiQuantCoef4096[i]    = CADR_DESCALE( g_aiQuantCoef4096[i],    g_iRangeCADR );

	// modify scaling values in inverse quantizer
	for ( i=0; i<6; i++ )										  		 g_aiDequantCoef4   [i]    = CADR_SCALE( g_aiDequantCoef4   [i],    g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<16;   j++ ) g_aiDequantCoef    [i][j] = CADR_SCALE( g_aiDequantCoef    [i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<64;   j++ ) g_aiDequantCoef64  [i][j] = CADR_SCALE( g_aiDequantCoef64  [i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ ) for ( j=0; j<256;  j++ ) g_aiDeQuantCoef256_Chen [i][j] = CADR_SCALE( g_aiDeQuantCoef256_Chen [i][j], g_iRangeCADR );
  for ( i=0; i<6; i++ ) for ( j=0; j<1024; j++ ) g_aiDeQuantCoef1024_Chen[i][j] = CADR_SCALE( g_aiDeQuantCoef1024_Chen[i][j], g_iRangeCADR );
	for ( i=0; i<6; i++ )                          g_aiDeQuantCoef4096[i]    = CADR_SCALE( g_aiDeQuantCoef4096[i],    g_iRangeCADR );

	// update quantization scaning matrix for LCT
	if ( bUseLCT ) 
	{
		for ( i=0; i<6; i++ ) for ( j=0; j<256;  j++ ) 
			g_aiQuantCoef256_Loeffler_Lifting [i][j] = CADR_DESCALE( g_aiQuantCoef256_Loeffler_Lifting [i][j], g_iRangeCADR );

		for ( i=0; i<6; i++ ) for ( j=0; j<1024; j++ ) 
			g_aiQuantCoef1024_Loeffler_Lifting[i][j] = CADR_DESCALE( g_aiQuantCoef1024_Loeffler_Lifting[i][j], g_iRangeCADR );
	}
  
	if ( bUseLCT )
	{
		for ( i=0; i<6; i++ ) for ( j=0; j<256;  j++ ) 
			g_aiDeQuantCoef256_Loeffler_Lifting [i][j] = CADR_SCALE( g_aiDeQuantCoef256_Loeffler_Lifting [i][j], g_iRangeCADR );

		for ( i=0; i<6; i++ ) for ( j=0; j<1024; j++ ) 
			g_aiDeQuantCoef1024_Loeffler_Lifting[i][j] = CADR_SCALE( g_aiDeQuantCoef1024_Loeffler_Lifting[i][j], g_iRangeCADR );
	}
}
